
/*
###################################################
###### HR3D: A_t * Sparse lightfield times 
 */

//Mex
#include "mex.h" 
#include <string>

//Eigen
#include "eigen/Eigen/Sparse"
#include "eigen/Eigen/Dense"

using namespace Eigen;

//Matrix typedefs
typedef Matrix<double, Dynamic, Dynamic> MatrixXdyn;
typedef Eigen::Triplet<double> T;

//####################################
//#### Main mex entry point 
//####################################

void mexFunction(int nlhs, mxArray *plhs[], /* Output variables */
		 int nrhs, const mxArray *prhs[]) /* Input variables */
{
    //AtL = AtmultL(samples, int32(ch), A, int32(layerInt))
    
    //###################################################
    //####### Get inputs
    
    //#### All in *row-vec format* ####
    if( nlhs != 1 || nrhs != 4 )
	    mexErrMsgTxt("Wrong input/output parameters.");
    
    //Get pointers and data
    //Samples
    double * samples = (double*)mxGetData(prhs[0]);
    const int *samples_dims = mxGetDimensions(prhs[0]);
    int num_samples = samples_dims[1];
    
    //Channel
    int ch = *((int*)mxGetData(prhs[1]));
    if( ch < 1 || ch > samples_dims[0] )
	    mexErrMsgTxt("Channel out of range.\n");
    
    //A
    double * A = (double*)mxGetData(prhs[2]);
    const int  *A_dims = mxGetDimensions(prhs[2]);
    int A_sz_layer = A_dims[0];
    int A_sz_t = A_dims[1];
    
    //Layer int
    int * layerInt = (int*)mxGetData(prhs[3]);
    const int *layerInt_dims = mxGetDimensions(prhs[3]);
    
    //Sanity
    if( layerInt_dims[0] != 6 ||
        layerInt_dims[1] != num_samples )
	    mexErrMsgTxt("Inputs have wrong size.\n");
    
    //Debug
    //printf("num_samples %d, layerdim %d, Ch %d, Aszt %d Aszlayer %d \n", num_samples, layerInt_dims[0],ch,
    //                                                           A_sz_t, A_sz_layer);

    //###################################################
    //####### Allocate result
    
    int ndim = 2;
    int dims[2];
    dims[0] = A_sz_t;
    dims[1] = A_sz_layer;
    
    plhs[0] = mxCreateNumericArray(ndim, dims, mxDOUBLE_CLASS, mxREAL);
    if (plhs[0] == NULL)
	mexErrMsgTxt("Could not create output mxArray.\n");
    
    //output
    double *res = (double *)mxGetPr(plhs[0]);
    
    //###################################################
    //####### Compute result
    
    
    //Initialize result with zeros
    for( int i = 0; i < A_sz_layer * A_sz_t; i++ )
        res[i] = 0.0;
    
    // #### Regular version
	
    /* 
    int backLayerIdx, frontLayerIdx;
    int sampleIdx, AIdx, outIdx;
	for( int s = 0; s < num_samples; s++ )
    {
        //Get indices
        frontLayerIdx = layerInt[ s * 6 + 2] - 1; //C-idx
        backLayerIdx = layerInt[ s * 6 + 5] - 1; //C-idx
        
        //compute sample indices
        sampleIdx = s * samples_dims[0] + (ch - 1); //(ch, s)
        
        //Multiply and add for all time steps
        for( int t = 0; t < A_sz_t; t++ )
        {
            //Compute indices
            AIdx = t * A_sz_layer + frontLayerIdx; //(frontLayer, t )
            outIdx = backLayerIdx * A_sz_t + t; //(backLayer, t)
                                
            //Compute
            res[outIdx] += samples[sampleIdx] * A[AIdx];
        } 
    }
   */
    
    // #### EIGEN Version
	
    //Build up sparse matrix 
    std::vector<T> L_tripletList;
    L_tripletList.reserve(num_samples);
    
    int backLayerIdx, frontLayerIdx;
    int sampleIdx, BIdx, outIdx;
    for( int s = 0; s < num_samples; s++ )
    {
        //Get indices
        frontLayerIdx = layerInt[ s * 6 + 2] - 1; //C-idx
        backLayerIdx = layerInt[ s * 6 + 5] - 1; //C-idx
        
        //compute sample indices
        sampleIdx = s * samples_dims[0] + (ch - 1); //(ch, s)
        
	L_tripletList.push_back( T( frontLayerIdx, backLayerIdx, samples[sampleIdx] ) );
    }
    SparseMatrix<double> L_mat(A_sz_layer, A_sz_layer);
    L_mat.setFromTriplets(L_tripletList.begin(), L_tripletList.end());
    //L_mat.makeCompressed(); //Optional
    
    //Build up dense A
    MatrixXdyn A_mat(A_sz_layer, A_sz_t);
    for( int t = 0; t < A_sz_t; t++ )
      for( int x = 0; x < A_sz_layer; x++ )
	A_mat(x,t) =  A[t * A_sz_layer + x];
      
    //Multiply
    MatrixXdyn res_mat(A_sz_layer, A_sz_t);
    res_mat = L_mat.transpose() * A_mat;
      
    //Transfer back
    for( int x = 0; x < A_sz_layer; x++ )
      for( int t = 0; t < A_sz_t; t++ )
	res[x * A_sz_t + t] = res_mat(x,t);
    
    return;
}



