
/*
###################################################
###### HR3D: Sparse lightfield times B_t
 */

//Mex
#include "mex.h" 
#include <string>

//Eigen
#include "eigen/Eigen/Sparse"
#include "eigen/Eigen/Dense"

using namespace Eigen;

//Matrix typedefs
typedef Matrix<double, Dynamic, Dynamic> MatrixXdyn;
typedef Eigen::Triplet<double> T;

//####################################
//#### Main mex entry point 
//####################################

void mexFunction(int nlhs, mxArray *plhs[], /* Output variables */
		 int nrhs, const mxArray *prhs[]) /* Input variables */
{
    //LBt = LmultBt(samples, int32(ch), B, int32(layerInt))
    
    //###################################################
    //####### Get inputs
    
    //#### All in *row-vec format* ####
    if( nlhs != 1 || nrhs != 4 )
	    mexErrMsgTxt("Wrong input/output parameters.");
    
    //Get pointers and data
    //Samples
    double * samples = (double*)mxGetData(prhs[0]);
    const int *samples_dims = mxGetDimensions(prhs[0]);
    int num_samples = samples_dims[1];
    
    //Channel
    int ch = *((int*)mxGetData(prhs[1]));
    if( ch < 1 || ch > samples_dims[0] )
	    mexErrMsgTxt("Channel out of range.\n");
    
    //B
    double * B = (double*)mxGetData(prhs[2]);
    const int  *B_dims = mxGetDimensions(prhs[2]);
    int B_sz_t = B_dims[0];
    int B_sz_layer = B_dims[1];
    
    //Layer int
    int * layerInt = (int*)mxGetData(prhs[3]);
    const int *layerInt_dims = mxGetDimensions(prhs[3]);
    
    //Sanity
    if( layerInt_dims[0] != 6 ||
        layerInt_dims[1] != num_samples )
	    mexErrMsgTxt("Inputs have wrong size.\n");
    
    //Debug
    //printf("num_samples %d, layerdim %d, Ch %d, Bszt %d Bszlayer %d \n", num_samples, layerInt_dims[0], ch,
    //                                                                     B_sz_t, B_sz_layer);
    
    //###################################################
    //####### Allocate result
    
    int ndim = 2;
    int dims[2];
    dims[0] = B_sz_layer;
    dims[1] = B_sz_t;
    
    plhs[0] = mxCreateNumericArray(ndim, dims, mxDOUBLE_CLASS, mxREAL);
    if (plhs[0] == NULL)
	mexErrMsgTxt("Could not create output mxArray.\n");
    
    //output
    double *res = (double *)mxGetPr(plhs[0]);
    
    //###################################################
    //####### Compute result
    
    //Initialize result with zeros
    for( int i = 0; i < B_sz_layer * B_sz_t; i++ )
        res[i] = 0.0;
   
      
    // #### Regular version
	
    /*
    int backLayerIdx, frontLayerIdx;
    int sampleIdx, BIdx, outIdx;
    for( int s = 0; s < num_samples; s++ )
    {
        //Get indices
        frontLayerIdx = layerInt[ s * 6 + 2] - 1; //C-idx
        backLayerIdx = layerInt[ s * 6 + 5] - 1; //C-idx
        
        //compute sample indices
        sampleIdx = s * samples_dims[0] + (ch - 1); //(ch, s)
        
        //Multiply and add for all time steps
        for( int t = 0; t < B_sz_t; t++ )
        {
            //Compute indices
            BIdx = backLayerIdx * B_sz_t + t; //(frontLayer, t)
            outIdx = t * B_sz_layer + frontLayerIdx; //(frontLayer, t)
                                
            //Compute
            res[outIdx] += samples[sampleIdx] * B[BIdx];
        } 
    }
    */
    
    //#### EIGEN Version
	
    //Build up sparse matrix 
    std::vector<T> L_tripletList;
    L_tripletList.reserve(num_samples);
    
    int backLayerIdx, frontLayerIdx;
    int sampleIdx, BIdx, outIdx;
    for( int s = 0; s < num_samples; s++ )
    {
        //Get indices
        frontLayerIdx = layerInt[ s * 6 + 2] - 1; //C-idx
        backLayerIdx = layerInt[ s * 6 + 5] - 1; //C-idx
        
        //compute sample indices
        sampleIdx = s * samples_dims[0] + (ch - 1); //(ch, s)
        
	L_tripletList.push_back( T( frontLayerIdx, backLayerIdx, samples[sampleIdx] ) );
    }
    SparseMatrix<double> L_mat(B_sz_layer, B_sz_layer);
    L_mat.setFromTriplets(L_tripletList.begin(), L_tripletList.end());
    //L_mat.makeCompressed(); //Optional
    
    //Build up dense B
    MatrixXdyn B_mat(B_sz_t, B_sz_layer);
    for( int x = 0; x < B_sz_layer; x++ )
      for( int t = 0; t < B_sz_t; t++ )
	B_mat(t,x) =  B[x * B_sz_t + t];
      
    //Multiply
    MatrixXdyn res_mat(B_sz_layer, B_sz_t);
    res_mat = L_mat * B_mat.transpose();
    
    //Transfer back
    for( int t = 0; t < B_sz_t; t++ )
      for( int x = 0; x < B_sz_layer; x++ )
	res[t * B_sz_layer + x] = res_mat(x,t);
    
    return;
}



