#include "matrix_factorization.h"

// C includes (no C++, allowed) //
#include <stdlib.h>
#include <stdio.h>

/* Using updated (v2) interfaces to cublas and cusparse */
#include <cuda.h>
#include <cuda_runtime.h>
#include <cusparse_v2.h>
#include <cublas_v2.h>

//Architecture specific cuda constants
#define numBlockThreads 1024
// Aim to launch around ten or more times as many blocks as there
// are multiprocessors on the target device.
#define blocksPerSM 20 

//#### Configuration
void setupConfig( dim3* block, dim3* grid, int numSims)
{
    // Get device properties
    cudaDeviceProp deviceProps;
    if (cudaGetDeviceProperties(&deviceProps, 0) != cudaSuccess)
    {
        printf("Could not get device properties.");
        //exit(1);
    }
    
    // Aim to launch around blockPerSM times as many blocks as there
    // are multiprocessors on the target device.
    int numSMs = 48; //Assumed value (device dependent)
    
    // Determine how to divide the work between cores
    block->x = numBlockThreads;
    grid->x  = (numSims + numBlockThreads) / numBlockThreads;

    //printf("Num SM on card is %d. So we should schedule %d blocks or less. That is %d threads.\n", numSMs, blocksPerSM * numSMs, blocksPerSM * numSMs * numBlockThreads);
    if( grid->x > blocksPerSM * numSMs )
        grid->x = blocksPerSM * numSMs;

    // Check the dimensions are valid
    if (grid->x > deviceProps.maxGridSize[0])
    {
	grid->x = deviceProps.maxGridSize[0] - 128;
    }
}

//#### Kernel helpers
__device__ inline void atomicFloatAdd(float *address, float val)
{
    int tmp0 = *address;
    int i_val = __float_as_int(val + __int_as_float(tmp0));
    int tmp1;
    
    // compare and swap v = (old == tmp0) ? i_val : old; (returns old)
    while( (tmp1 = atomicCAS((int *)address, tmp0, i_val)) != tmp0 )
    {
	tmp0 = tmp1;
	i_val = __float_as_int(val + __int_as_float(tmp1));
    }
}

//#### Kernels

__global__ void AmultB_kernel(float *A, float *B,  int* layerInt, float *res,
			      int num_samples,
			      int A_sz_layer, int B_sz_t)
{
	
    //Sample idx
    int s = blockIdx.x * blockDim.x + threadIdx.x;
    
    //Vars
    int AIdx, BIdx;
    float res_curr;
    for (; s < num_samples; s+= gridDim.x * blockDim.x) 
    {    
	    //Get indices
	    int frontLayerIdx = layerInt[ s * 6 + 2] - 1; //C-idx
	    int backLayerIdx = layerInt[ s * 6 + 5] - 1; //C-idx
	    
	    //Multiply and add
	    res_curr = 0.0;
	    for( int t = 0; t < B_sz_t; t++ )
	    {
		    //Compute indices
		    AIdx = t * A_sz_layer + frontLayerIdx; //(frontLayer, t )
		    BIdx = backLayerIdx * B_sz_t + t; //(backLayer, t)
							    
		    //Compute
		    res_curr += A[AIdx] * B[BIdx];
	    } 
	    
	    //Write back
	    res[s] = res_curr;
    }
}
    
__global__ void AtmultL_kernel(float *A, float *samples,  int* layerInt, float *res,
				int ch,
				int num_samples, int sample_components,
				int A_sz_layer, int A_sz_t)
{
	
    //Sample idx
    int s = blockIdx.x * blockDim.x + threadIdx.x;
    
    //Vars
    int AIdx, outIdx;
    float addComp;
    for (; s < num_samples; s+= gridDim.x * blockDim.x) 
    {    
	    //Get indices
	    int frontLayerIdx = layerInt[ s * 6 + 2] - 1; //C-idx
	    int backLayerIdx = layerInt[ s * 6 + 5] - 1; //C-idx
	    
	    //compute sample indices
	    int sampleIdx = s * sample_components + (ch - 1); //(ch, s)
	    
	    //Multiply and add for all time steps
	    for( int t = 0; t < A_sz_t; t++ )
	    {
		    //Compute indices
		    AIdx = t * A_sz_layer + frontLayerIdx; //(frontLayer, t )
		    outIdx = backLayerIdx * A_sz_t + t; //(backLayer, t)
							    
		    //Compute
		    addComp = samples[sampleIdx] * A[AIdx];
		    //atomicFloatAdd( &(res[outIdx]), addComp );
		    atomicAdd( &(res[outIdx]), addComp );
	    } 
    }
}

__global__ void LmultBt_kernel(float *B, float *samples,  int* layerInt, float *res,
				int ch,
				int num_samples, int sample_components,
				int B_sz_layer, int B_sz_t)
{
	
    //Sample idx
    int s = blockIdx.x * blockDim.x + threadIdx.x;
    
    //Vars
    int BIdx, outIdx;
    float addComp;
    for (; s < num_samples; s+= gridDim.x * blockDim.x) 
    {    
	    //Get indices
	    int frontLayerIdx = layerInt[ s * 6 + 2] - 1; //C-idx
	    int backLayerIdx = layerInt[ s * 6 + 5] - 1; //C-idx
	    
	    //compute sample indices
	    int sampleIdx = s * sample_components + (ch - 1); //(ch, s)
	    
	    //Multiply and add for all time steps
	    for( int t = 0; t < B_sz_t; t++ )
	    {
		    //Compute indices
		    BIdx = backLayerIdx * B_sz_t + t; //(frontLayer, t)
		    outIdx = t * B_sz_layer + frontLayerIdx; //(frontLayer, t)
		    				    
		    //Compute
		    addComp = samples[sampleIdx] * B[BIdx];
		    //atomicFloatAdd( &(res[outIdx]), addComp );
		    atomicAdd( &(res[outIdx]), addComp );
	    } 
    }
}

__global__ void compute_upper_kernel(float *d_upper, float *d_reg, int numElements,
									 float epsilon, float lambdaA, int regularizer)
{
	
    //Sample idx
    int s = blockIdx.x * blockDim.x + threadIdx.x;
    
    //Vars
    float currVal;
    for (; s < numElements; s+= gridDim.x * blockDim.x) 
    {    
	    currVal = d_upper[s];
	    
	    //Regularizer
	    if( regularizer == 1 ) 
	    {
		    //Tikhonov
		    currVal = currVal + lambdaA * d_reg[s];
	    }
	    
	    //Thresh
	    currVal = (currVal <= epsilon) ? epsilon : currVal;
	    
	    //Write back
	    d_upper[s]= currVal;
    }
}

__global__ void compute_update_kernel(float *d_old, float *d_upper, float *d_lower,
									 int numElements,
									 float epsilon, float epsilonLower)
{
				
    //Sample idx
    int s = blockIdx.x * blockDim.x + threadIdx.x;
    
    //Vars
    float currVal;
    for (; s < numElements; s+= gridDim.x * blockDim.x) 
    {    
	    //Update
	    currVal = d_old[s] * d_upper[s] / ( d_lower[s] + epsilonLower );
	    
	    //Thresh
	    currVal = (currVal <= 0.f) ? epsilon : currVal;
	    currVal = (currVal >= 1.f) ? 1.f : currVal;
	    
	    //Write back
	    d_old[s]= currVal;
    }
}

//Matlab column major format as input assumed
__global__ void transpose_kernel(float *M, float *M_t, int m, int n)
{
	
    //Sample idx
    int t = blockIdx.x * blockDim.x + threadIdx.x;
    
    //Vars
    for (; t < m * n; t += gridDim.x * blockDim.x) 
    {    
	    //Get indices
	    int i = t % m;
	    int j = (t - i)  / m;
	    
	    //Copy
	    M_t[i*n + j] = M[t];
    }
}



//#### Launchers
void AmultB_cuda(float *A, float *B, int* layerInt, float *res,  
		  int num_samples,
		  int A_sz_layer, int B_sz_t)
{
	float *d_A = NULL;
	float *d_B = NULL;
	int *d_layerInt = NULL;
	float *d_res = NULL;

	// Allocate on device
	cudaMalloc((void**)&d_A, (A_sz_layer * B_sz_t) * sizeof(float));
	cudaMalloc((void**)&d_B, (B_sz_t * A_sz_layer) * sizeof(float));
	cudaMalloc((void**)&d_layerInt, ( 6 * num_samples) * sizeof(int));
	cudaMalloc((void**)&d_res, (num_samples) * sizeof(float));
	
	// Transfer src to device
	cudaMemcpy(d_A, A, (A_sz_layer * B_sz_t) * sizeof(float), cudaMemcpyHostToDevice);
	cudaMemcpy(d_B, B, (B_sz_t * A_sz_layer) * sizeof(float), cudaMemcpyHostToDevice);
	cudaMemcpy(d_layerInt, layerInt, ( 6 * num_samples) * sizeof(int), cudaMemcpyHostToDevice);
	
	//Reset result
	cudaMemset(d_res, 0, (num_samples) * sizeof(float));

	//Compute configuration 
	dim3 kernelBlockSize, kernelGridSize;
	setupConfig( &kernelBlockSize, &kernelGridSize, num_samples);
	
	AmultB_kernel<<<kernelBlockSize, kernelGridSize>>>(d_A, d_B, d_layerInt, d_res,
							num_samples,
							A_sz_layer, B_sz_t);
	
	// Fetch data back
	cudaMemcpy(res, d_res, (num_samples) * sizeof(float), cudaMemcpyDeviceToHost);

	// Release memory
	cudaFree(d_A);
	cudaFree(d_B);
	cudaFree(d_layerInt);
	cudaFree(d_res);
}

void AtmultL_cuda(float *A, float *samples, int* layerInt, float *res,  
		  int ch,
		  int num_samples, int sample_components,
		  int A_sz_layer, int A_sz_t)
{
	float *d_A = NULL;
	float *d_samples = NULL;
	int *d_layerInt = NULL;
	float *d_res = NULL;

	// Allocate on device
	cudaMalloc((void**)&d_A, (A_sz_layer * A_sz_t) * sizeof(float));
	cudaMalloc((void**)&d_samples, (num_samples * sample_components) * sizeof(float));
	cudaMalloc((void**)&d_layerInt, ( 6 * num_samples) * sizeof(int));
	cudaMalloc((void**)&d_res, ( A_sz_t * A_sz_layer) * sizeof(float));
	
	// Transfer src to device
	cudaMemcpy(d_A, A, (A_sz_layer * A_sz_t) * sizeof(float), cudaMemcpyHostToDevice);
	cudaMemcpy(d_samples, samples, (num_samples * sample_components) * sizeof(float), cudaMemcpyHostToDevice);
	cudaMemcpy(d_layerInt, layerInt, ( 6 * num_samples) * sizeof(int), cudaMemcpyHostToDevice);
	
	//Reset result
	cudaMemset(d_res, 0, ( A_sz_t * A_sz_layer) * sizeof(float));

	//Compute configuration 
	dim3 kernelBlockSize, kernelGridSize;
	setupConfig( &kernelBlockSize, &kernelGridSize, num_samples);
	
	AtmultL_kernel<<<kernelBlockSize, kernelGridSize>>>(d_A, d_samples, d_layerInt, d_res,
							ch,
							num_samples, sample_components,
							A_sz_layer, A_sz_t);
	// Fetch data back
	cudaMemcpy(res, d_res, ( A_sz_t * A_sz_layer) * sizeof(float), cudaMemcpyDeviceToHost);

	// Release memory
	cudaFree(d_A);
	cudaFree(d_samples);
	cudaFree(d_layerInt);
	cudaFree(d_res);
}

/*
void AtmultL_cuda(float *A, float *samples, int* layerInt, float *res,  
		  int ch,
		  int num_samples, int sample_components,
		  int A_sz_layer, int A_sz_t)
{
	//CUSPARSE version
	
	//#### initialize cusparse library
	cusparseStatus_t status;
    cusparseHandle_t handle=0;
    cusparseMatDescr_t descr=0;

    status = cusparseCreate(&handle);
    if (status != CUSPARSE_STATUS_SUCCESS) {
        printf("CUSPARSE Library initialization failed");
        return;
    }

    // create and setup matrix descriptor 
    status= cusparseCreateMatDescr(&descr); 
    if (status != CUSPARSE_STATUS_SUCCESS) {
        printf("Matrix descriptor initialization failed");
        return;
    }       
    cusparseSetMatType(descr,CUSPARSE_MATRIX_TYPE_GENERAL);
    cusparseSetMatIndexBase(descr,CUSPARSE_INDEX_BASE_ZERO);  

	//#### create the n x n lightfield-matrix for cusparse
    int n = A_sz_layer; 
	int nnz = num_samples; 
    int* cooRowIndexHostPtr = new int[nnz]; 
    int* cooColIndexHostPtr = new int[nnz]; 
    float* cooValHostPtr    = new float[nnz];
	
	//Tranfer
	for( int s = 0; s < num_samples; s++ )
	{
		cooRowIndexHostPtr[s] = layerInt[ s * 6 + 2] - 1; //Frontlayer
	    cooColIndexHostPtr[s] = layerInt[ s * 6 + 5] - 1; //Backlayer
	    cooValHostPtr[s] = samples[ s * sample_components + (ch - 1) ]; //Samples
	}
	
	//Allocate and tranfer matrix to cuda
	int *    cooRowIndex=0;
    int *    cooColIndex=0;    
    float *  cooVal=0;
	
	cudaMalloc((void**)&cooRowIndex, nnz*sizeof(int));  
	cudaMalloc((void**)&cooColIndex, nnz*sizeof(int)); 
	cudaMalloc((void**)&cooVal, nnz*sizeof(float)); 
	
    cudaMemcpy(cooRowIndex, cooRowIndexHostPtr, nnz*sizeof(int), cudaMemcpyHostToDevice);
    cudaMemcpy(cooColIndex, cooColIndexHostPtr, nnz*sizeof(int), cudaMemcpyHostToDevice);
    cudaMemcpy(cooVal,      cooValHostPtr, nnz*sizeof(float), cudaMemcpyHostToDevice);
	
	//#### Create dense matrix
    
	//Alloc cuda array
	float *d_A = NULL;
	cudaMalloc((void**)&d_A, (A_sz_layer *  A_sz_t) * sizeof(float));
	cudaMemcpy(d_A, A, (A_sz_layer *  A_sz_t) * sizeof(float), cudaMemcpyHostToDevice);

	//#### Allocate dense result
	float *d_res = NULL;
	cudaMalloc((void**)&d_res, ( (A_sz_layer + 1) * A_sz_t) * sizeof(float));
	
	//Reset result
	cudaMemset(d_res, 0, ( (A_sz_layer + 1) * A_sz_t) * sizeof(float));
	
	//#### Do the sparse matrix multiplication

	//Compress the rows (convert matrix from COO 2 CSR format) 
	int * csrRowPtr=0;
    cudaMalloc((void**)&csrRowPtr,(n+1)*sizeof(int));
    cusparseXcoo2csr(handle,cooRowIndex,nnz,n, csrRowPtr,CUSPARSE_INDEX_BASE_ZERO); 

	//Do the computation
	float alpha =1.f;
	float beta =0.f;
    status= cusparseScsrmm(handle, CUSPARSE_OPERATION_TRANSPOSE, n, A_sz_t, n, 
                           nnz, &alpha, descr, cooVal, csrRowPtr, cooColIndex, 
                           d_A, n, &beta, d_res, n+1);
	
    if (status != CUSPARSE_STATUS_SUCCESS) {
        printf("Matrix-matrix multiplication failed");
        return;
    }  

	//#### Fetch data back
	float* tmp_res = new float[( (A_sz_layer + 1) * A_sz_t)];
	cudaMemcpy(tmp_res, d_res, ( (A_sz_layer + 1) * A_sz_t) * sizeof(float), cudaMemcpyDeviceToHost);

	for( int x = 0; x < A_sz_layer; x++ )
		for( int t = 0; t < A_sz_t; t++ )
			res[x * A_sz_t + t] = tmp_res[ t * (A_sz_layer + 1) + x ];

	// Release memory
	cudaFree(d_A);
	cudaFree(d_res);

	cudaFree(cooRowIndex);
	cudaFree(cooColIndex);
	cudaFree(cooVal);
	
	cudaFree(csrRowPtr);
	
	// destroy matrix descriptor
    status = cusparseDestroyMatDescr(descr); 
    descr = 0;
    if (status != CUSPARSE_STATUS_SUCCESS) {
        printf("Matrix descriptor destruction failed");
        return;
    }    

    // destroy handle
    status = cusparseDestroy(handle);
    handle = 0;
    if (status != CUSPARSE_STATUS_SUCCESS) {
        printf("CUSPARSE Library release of resources failed");
        return;
    }   
	
	//Tmp
	delete[] cooRowIndexHostPtr;
	delete[] cooColIndexHostPtr;
	delete[] cooValHostPtr;
	delete[] tmp_res;
}
*/

void LmultBt_cuda(float *B, float *samples, int* layerInt, float *res,  
		  int ch,
		  int num_samples, int sample_components,
		  int B_sz_layer, int B_sz_t)
{
		
	float *d_B = NULL;
	float *d_samples = NULL;
	int *d_layerInt = NULL;
	float *d_res = NULL;

	// Allocate on device
	cudaMalloc((void**)&d_B, (B_sz_t * B_sz_layer) * sizeof(float));
	cudaMalloc((void**)&d_samples, (num_samples * sample_components) * sizeof(float));
	cudaMalloc((void**)&d_layerInt, ( 6 * num_samples) * sizeof(int));
	cudaMalloc((void**)&d_res, ( B_sz_layer * B_sz_t) * sizeof(float));
	
	// Transfer src to device
	cudaMemcpy(d_B, B, (B_sz_t * B_sz_layer) * sizeof(float), cudaMemcpyHostToDevice);
	cudaMemcpy(d_samples, samples, (num_samples * sample_components) * sizeof(float), cudaMemcpyHostToDevice);
	cudaMemcpy(d_layerInt, layerInt, ( 6 * num_samples) * sizeof(int), cudaMemcpyHostToDevice);
	
	//Reset result
	cudaMemset(d_res, 0, ( B_sz_layer * B_sz_t) * sizeof(float));

	//Compute configuration 
	dim3 kernelBlockSize, kernelGridSize;
	setupConfig( &kernelBlockSize, &kernelGridSize, num_samples);
	
	LmultBt_kernel<<<kernelBlockSize, kernelGridSize>>>(d_B, d_samples, d_layerInt, d_res,
							ch,
							num_samples, sample_components,
							B_sz_layer, B_sz_t);
	
	//Transpose test
	//dim3 kernelBlockSizeLayer, kernelGridSizeLayer;
	//setupConfig( &kernelBlockSizeLayer, &kernelGridSizeLayer, B_sz_layer * B_sz_t);
	//transpose_kernel<<<kernelBlockSizeLayer, kernelGridSizeLayer>>>(d_B, d_res, B_sz_t, B_sz_layer);
	
	
	// Fetch data back
	cudaMemcpy(res, d_res, ( B_sz_layer * B_sz_t) * sizeof(float), cudaMemcpyDeviceToHost);

	// Release memory
	cudaFree(d_B);
	cudaFree(d_samples);
	cudaFree(d_layerInt);
	cudaFree(d_res); 
}

/*
void LmultBt_cuda(float *B, float *samples, int* layerInt, float *res,  
		  int ch,
		  int num_samples, int sample_components,
		  int B_sz_layer, int B_sz_t)
{
	//CUSPARSE version
	
	//#### initialize cusparse library
	cusparseStatus_t status;
	cusparseHandle_t handle=0;
	cusparseMatDescr_t descr=0;

	status = cusparseCreate(&handle);
	if (status != CUSPARSE_STATUS_SUCCESS) {
	    printf("CUSPARSE Library initialization failed");
	    return;
	}

	// create and setup matrix descriptor  
	status= cusparseCreateMatDescr(&descr); 
	if (status != CUSPARSE_STATUS_SUCCESS) {
	    printf("Matrix descriptor initialization failed");
	    return;
	}       
	cusparseSetMatType(descr,CUSPARSE_MATRIX_TYPE_GENERAL);
	cusparseSetMatIndexBase(descr,CUSPARSE_INDEX_BASE_ZERO);  

	//#### create the n x n lightfield-matrix for cusparse
	int n = B_sz_layer; 
	int nnz = num_samples; 
	    
	int* cooRowIndexHostPtr = (int *)   malloc(nnz*sizeof(int)); 
	int* cooColIndexHostPtr = (int *)   malloc(nnz*sizeof(int)); 
	float* cooValHostPtr    = (float *) malloc(nnz*sizeof(float)); 
	    
	//Tranfer
	for( int s = 0; s < num_samples; s++ )
	{
		cooRowIndexHostPtr[s] = layerInt[ s * 6 + 2] - 1; //Frontlayer
	    cooColIndexHostPtr[s] = layerInt[ s * 6 + 5] - 1; //Backlayer
	    cooValHostPtr[s] = samples[ s * sample_components + (ch - 1) ]; //Samples
	}
	
	//Allocate and tranfer matrix to cuda
	int *    cooRowIndex=0;
	int *    cooColIndex=0;    
	float *  cooVal=0;
	    
	cudaMalloc((void**)&cooRowIndex, nnz*sizeof(int));  
	cudaMalloc((void**)&cooColIndex, nnz*sizeof(int)); 
	cudaMalloc((void**)&cooVal, nnz*sizeof(float)); 
	    
	cudaMemcpy(cooRowIndex, cooRowIndexHostPtr, nnz*sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(cooColIndex, cooColIndexHostPtr, nnz*sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(cooVal, cooValHostPtr, nnz*sizeof(float), cudaMemcpyHostToDevice);
	    
	//#### Create dense matrix

	//Transpose
	float* BtPtr = (float *)malloc( (B_sz_layer * B_sz_t) *sizeof( float ) ); 
	
	//Transfer
	for( int t = 0; t < B_sz_t; t++ )
		for( int x = 0; x < B_sz_layer; x++ )
			BtPtr[t * B_sz_layer + x] = B[ x * B_sz_t + t ];

	//Alloc cuda array
	float *d_Bt = NULL;
	cudaMalloc((void**)&d_Bt, (B_sz_layer *  B_sz_t) * sizeof(float));
	cudaMemcpy(d_Bt, BtPtr, (B_sz_layer *  B_sz_t) * sizeof(float), cudaMemcpyHostToDevice);
	
	//#### Allocate dense result
	float *d_res = NULL;
	cudaMalloc((void**)&d_res, ( B_sz_layer * B_sz_t) * sizeof(float));
	
	//Reset result
	cudaMemset(d_res, 0, ( B_sz_layer * B_sz_t) * sizeof(float));
	
	//#### Do the sparse matrix multiplication

	//Compress the rows (convert matrix from COO 2 CSR format) 
	int * csrRowPtr=0;
	cudaMalloc((void**)&csrRowPtr,(n+1)*sizeof(int));
	cusparseXcoo2csr(handle,cooRowIndex,nnz,n, csrRowPtr,CUSPARSE_INDEX_BASE_ZERO); 
	    
	//Do the computation
	float alpha =1.f;
	float beta =1.f;
	status= cusparseScsrmm(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, n, B_sz_t, n, 
			      nnz, &alpha, descr, cooVal, csrRowPtr, cooColIndex, 
			      d_Bt, n, &beta, d_res, n );
	    
	if (status != CUSPARSE_STATUS_SUCCESS) {
	    printf("Matrix-matrix multiplication failed");
	    return;
	}  

	//#### Fetch data back
	cudaMemcpy(res, d_res, ( B_sz_layer * B_sz_t) * sizeof(float), cudaMemcpyDeviceToHost);

	
	// Release memory
	cudaFree(d_Bt);
	cudaFree(d_res);

	cudaFree(cooRowIndex);
	cudaFree(cooColIndex);
	cudaFree(cooVal);
	
	cudaFree(csrRowPtr);
	    
	    // destroy matrix descriptor
	status = cusparseDestroyMatDescr(descr); 
	descr = 0;
	if (status != CUSPARSE_STATUS_SUCCESS) {
	    printf("Matrix descriptor destruction failed");
	    return;
	}    

	// destroy handle
	status = cusparseDestroy(handle);
	handle = 0;
	if (status != CUSPARSE_STATUS_SUCCESS) {
	    printf("CUSPARSE Library release of resources failed");
	    return;
	}   
	    
	//Tmp
	free(cooRowIndexHostPtr);
	free(cooColIndexHostPtr);
	free(cooValHostPtr);
	free(BtPtr);
}
*/

void matrix_factorization_cuda(float *A, float *B, float *samples, int* layerInt, 
		  float *resA, float *resB,     
		  int ch,
		  int num_samples, int sample_components,
		  int A_sz_layer, int A_sz_t,
		  int B_sz_layer, int B_sz_t)
{
	float *d_A = NULL;
	float *d_B = NULL;
	float *d_samples = NULL;
	float *d_AB = NULL;
	int *d_layerInt = NULL;
	
	float *d_upper = NULL;
	float *d_lower = NULL;
	
	// Allocate on device
	cudaMalloc((void**)&d_A, (A_sz_layer * A_sz_t) * sizeof(float));
	cudaMalloc((void**)&d_B, (B_sz_t * B_sz_layer) * sizeof(float));
	cudaMalloc((void**)&d_samples, (num_samples * sample_components) * sizeof(float));
	cudaMalloc((void**)&d_AB, (num_samples) * sizeof(float));
	cudaMalloc((void**)&d_layerInt, ( 6 * num_samples) * sizeof(int));
	
	//Temporary vars
	cudaMalloc((void**)&d_upper, (A_sz_layer * A_sz_t) * sizeof(float));
	cudaMalloc((void**)&d_lower, (A_sz_layer * A_sz_t) * sizeof(float));

	// Transfer src to device
	cudaMemcpy(d_A, A, (A_sz_layer * A_sz_t) * sizeof(float), cudaMemcpyHostToDevice);
	cudaMemcpy(d_B, B, (B_sz_t * B_sz_layer) * sizeof(float), cudaMemcpyHostToDevice);
	cudaMemcpy(d_samples, samples, (num_samples * sample_components) * sizeof(float), cudaMemcpyHostToDevice);
	cudaMemcpy(d_layerInt, layerInt, ( 6 * num_samples) * sizeof(int), cudaMemcpyHostToDevice);

	//Compute configurations
	dim3 kernelBlockSize, kernelGridSize;
	setupConfig( &kernelBlockSize, &kernelGridSize, num_samples); //Parallel for all threads in numsamples
	
	int num_unknowns = ( A_sz_layer * A_sz_t);
	dim3 kernelBlockSizeUnknowns, kernelGridSizeUnknowns;
	setupConfig( &kernelBlockSizeUnknowns, &kernelGridSizeUnknowns, num_unknowns); //Parallel for numunknowns
	
	//Setup iterations:
	//small value thresh
	float epsilon         = 1e-9;
	float epsilonLower    = epsilon;

	//Regularizer weights
	float lambdaA = 0.005;
	float lambdaB = 0.005;
	
	//Tikhonov
	int regularizer = 1;
		
	//Iteration config
	for( int iter = 0; iter < 150; iter++ )
	{
		//#### COMPUTE NEW SAMPLES
		cudaMemset(d_AB, 0, ( num_samples) * sizeof(float));
		AmultB_kernel<<<kernelBlockSize, kernelGridSize>>>(d_A, d_B, d_layerInt, d_AB,
							num_samples,
							A_sz_layer, B_sz_t);
		
		//#### UPDATE A
		//Compute A_upper
		cudaMemset(d_upper, 0, ( B_sz_layer * B_sz_t) * sizeof(float));
		LmultBt_kernel<<<kernelBlockSize, kernelGridSize>>>(d_B, d_samples, d_layerInt, d_upper,
								ch,
								num_samples, sample_components,
								B_sz_layer, B_sz_t);
		
		compute_upper_kernel<<<kernelBlockSizeUnknowns, kernelGridSizeUnknowns>>>(d_upper, d_A, num_unknowns,
									 epsilon, lambdaA, regularizer);
		
		//Compute A_lower 
		cudaMemset(d_lower, 0, ( B_sz_layer * B_sz_t) * sizeof(float));
		LmultBt_kernel<<<kernelBlockSize, kernelGridSize>>>(d_B, d_AB, d_layerInt, d_lower,
								1,
								num_samples, 1,
								B_sz_layer, B_sz_t);
		
		//Do the update on A
		compute_update_kernel<<<kernelBlockSizeUnknowns, kernelGridSizeUnknowns>>>(d_A, d_upper, d_lower,
									 num_unknowns,
									 epsilon, epsilonLower);
		
		
		//#### UPDATE B
		//Compute B_upper
		cudaMemset(d_upper, 0, ( A_sz_t * A_sz_layer) * sizeof(float));
		AtmultL_kernel<<<kernelBlockSize, kernelGridSize>>>(d_A, d_samples, d_layerInt, d_upper,
							ch,
							num_samples, sample_components,
							A_sz_layer, A_sz_t);
		
		compute_upper_kernel<<<kernelBlockSizeUnknowns, kernelGridSizeUnknowns>>>(d_upper, d_B, num_unknowns,
									 epsilon, lambdaB, regularizer);
		
		//Compute B_lower 
		cudaMemset(d_lower, 0, ( A_sz_t * A_sz_layer) * sizeof(float));
		AtmultL_kernel<<<kernelBlockSize, kernelGridSize>>>(d_A, d_AB, d_layerInt, d_lower,
							1,
							num_samples, 1,
							A_sz_layer, A_sz_t);
		
		//Do the update on B
		compute_update_kernel<<<kernelBlockSizeUnknowns, kernelGridSizeUnknowns>>>(d_B, d_upper, d_lower,
									 num_unknowns,
									 epsilon, epsilonLower);
	
	}
	
	// Fetch data back
	cudaMemcpy(resA, d_A, ( A_sz_layer * A_sz_t) * sizeof(float), cudaMemcpyDeviceToHost);
	cudaMemcpy(resB, d_B, ( B_sz_layer * B_sz_t) * sizeof(float), cudaMemcpyDeviceToHost);

	// Release memory
	cudaFree(d_A);
	cudaFree(d_B);
	cudaFree(d_samples);
	cudaFree(d_AB);
	cudaFree(d_layerInt);
	
	//Tmp
	cudaFree(d_upper);
	cudaFree(d_lower);
}

