/*
###################################################
###### HR3D: Sparse matrix factorization
 */

//Includes
#include <iostream>
#include <string>
#include <cassert>
#include "mex.h"
#include "matrix.h"

//Eigen
#include "eigen/Eigen/Sparse"
#include "eigen/Eigen/Dense"

using namespace Eigen;

//Matrix typedefs
typedef Matrix<double, Dynamic, Dynamic> MatrixXdyn;
typedef Eigen::Triplet<double> T;

//####################################
//#### Main mex entry point 
//####################################

void mexFunction(int nlhs, mxArray *plhs[], /* Output variables */
		 int nrhs, const mxArray *prhs[]) /* Input variables */
{
    //A, B = sparse_matrix_factorization(samples, int32(ch), A, B, int32(layerInt), int32(numFixedCols),
    //					lambdaA, lambdaB, tol)
	
    //OLD
    //A, B = sparse_matrix_factorization(samples, int32(ch), A, B, int32(layerInt), Aprev, Bprev, int32(numFixedCols),
    //					lambdaA, lambdaB, tol)
	
    //###################################################
    //####### Get inputs
    
    //#### All in *row-vec format* ####
    if( nlhs != 2 || nrhs != 9 )
	    mexErrMsgTxt("Wrong input/output parameters.");
    
    //Get pointers and data
    //Samples
    double * samples = (double*)mxGetData(prhs[0]);
    const int *samples_dims = mxGetDimensions(prhs[0]);
    int num_samples = samples_dims[1];
    
    //Channel
    int ch = *((int*)mxGetData(prhs[1]));
    if( ch < 1 || ch > samples_dims[0] )
	    mexErrMsgTxt("Channel out of range.\n");
	
    //A
    double * A = (double*)mxGetData(prhs[2]);
    const int  *A_dims = mxGetDimensions(prhs[2]);
    int A_sz_layer = A_dims[0];
    int A_sz_t = A_dims[1];
    
    //B
    double * B = (double*)mxGetData(prhs[3]);
    const int  *B_dims = mxGetDimensions(prhs[3]);
    int B_sz_t = B_dims[0];
    int B_sz_layer = B_dims[1];
    
    //Layer int
    int * layerInt = (int*)mxGetData(prhs[4]);
    const int *layerInt_dims = mxGetDimensions(prhs[4]);
    
    //A prev
    //double * A_prev = (double*)mxGetData(prhs[5]);
    
    //B prev
    //double * B_prev = (double*)mxGetData(prhs[6]);
    
    //Number of previous fixed layers
    int * numFixedCols = (int*)mxGetData(prhs[5]);
    int numRightCols = A_sz_t - *numFixedCols;
    
    //Regularization params
    double * lbd_A = (double*)mxGetData(prhs[6]);
    double * lbd_B = (double*)mxGetData(prhs[7]);
    
    //Tolerance
    double * tol = (double*)mxGetData(prhs[8]);
    
    //Sanity
    if( layerInt_dims[0] != 6 ||
        layerInt_dims[1] != num_samples ||
	numRightCols < 0)
	    mexErrMsgTxt("Inputs have wrong size.\n");
	    

    
    //###################################################
    //####### Allocate result
		
    int ndim = 2;
    int dims[2];
    dims[0] = A_sz_layer;
    dims[1] = A_sz_t;
    
    plhs[0] = mxCreateNumericArray(ndim, dims, mxDOUBLE_CLASS, mxREAL);
    if (plhs[0] == NULL)
		mexErrMsgTxt("Could not create output mxArray.\n");
    
    //output
    double *resA = mxGetPr(plhs[0]);
    
    dims[0] = B_sz_t;
    dims[1] = B_sz_layer;
    
    plhs[1] = mxCreateNumericArray(ndim, dims, mxDOUBLE_CLASS, mxREAL);
    if (plhs[1] == NULL)
		mexErrMsgTxt("Could not create output mxArray.\n");
    
    //output
    double *resB = mxGetPr(plhs[1]);
    
    //###################################################
    //####### Compute result

    //###################################################
    //####### Build up sparse lightfield matrix
    //####### and preallocate A*B result
    std::vector<T> L_tripletList;
    L_tripletList.reserve(num_samples);
    
    std::vector<T> AB_tripletList;
    AB_tripletList.reserve(num_samples);
    
    int backLayerIdx, frontLayerIdx;
    int sampleIdx, BIdx, outIdx;
    for( int s = 0; s < num_samples; s++ )
    {
        //Get indices
        frontLayerIdx = layerInt[ s * 6 + 2] - 1; //C-idx
        backLayerIdx = layerInt[ s * 6 + 5] - 1; //C-idx
        
        //compute sample indices
        sampleIdx = s * samples_dims[0] + (ch - 1); //(ch, s)
        
        //Add corresponding triple
	L_tripletList.push_back( T( frontLayerIdx, backLayerIdx, samples[sampleIdx] ) );
	AB_tripletList.push_back(  T( frontLayerIdx, backLayerIdx, samples[sampleIdx] ) );
    }
    SparseMatrix<double> L_mat(A_sz_layer, B_sz_layer);
    L_mat.setFromTriplets(L_tripletList.begin(), L_tripletList.end());
    //L_mat.makeCompressed(); //Optional
    
    SparseMatrix<double> AB_mat(A_sz_layer, B_sz_layer);
    
    //###################################################
    //####### Build up dense A and B matrices
    MatrixXdyn A_mat(A_sz_layer, A_sz_t);
    for( int t = 0; t < A_sz_t; t++ )
      for( int x = 0; x < A_sz_layer; x++ )
	A_mat(x,t) =  A[t * A_sz_layer + x];
      
    //Build up dense B
    MatrixXdyn B_mat(B_sz_t, B_sz_layer);
    for( int x = 0; x < B_sz_layer; x++ )
      for( int t = 0; t < B_sz_t; t++ )
	B_mat(t,x) =  B[x * B_sz_t + t];
      
    /*
    //####### Build up dense A prev and B prev matrices
    MatrixXdyn Aprev_mat(A_sz_layer, A_sz_t);
    for( int t = 0; t < A_sz_t; t++ )
      for( int x = 0; x < A_sz_layer; x++ )
	Aprev_mat(x,t) =  A_prev[t * A_sz_layer + x];
      
    //Build up dense B
    MatrixXdyn Bprev_mat(B_sz_t, B_sz_layer);
    for( int x = 0; x < B_sz_layer; x++ )
      for( int t = 0; t < B_sz_t; t++ )
	Bprev_mat(t,x) =  B_prev[x * B_sz_t + t];
    */
      
    //###################################################
    //####### Do the iteration
	
    //Setup iterations:
    //small value thresh
    double epsilon         = 1e-9;
    double epsilonLower    = epsilon;

    //Regularizer weights
    double lambdaA = *lbd_A;
    double lambdaB = *lbd_B;
    
    //Temp matrices
    MatrixXdyn upper_A(A_sz_layer, numRightCols);
    MatrixXdyn upper_B(numRightCols, B_sz_layer);
    
    //Old solutions
    MatrixXdyn A_mat_old(A_sz_layer, A_sz_t);
    MatrixXdyn B_mat_old(B_sz_t, B_sz_layer);
    
    //Iteration config
    double rel_change;
    int iter;
    
    //printf("Iteraterion started \n\n");
    for(iter = 0; iter < 150; iter++ )
    {
	//##########################################################
	//Store old solutions
	A_mat_old = A_mat;
	B_mat_old = B_mat;
    
	//##########################################################
	//#### COMPUTE NEW SAMPLES
	double res_curr;
	for( int s = 0; s < num_samples; s++ )
	{
	    //Get indices
	    frontLayerIdx = layerInt[ s * 6 + 2] - 1; //C-idx
	    backLayerIdx = layerInt[ s * 6 + 5] - 1; //C-idx
	    
	    //Multiply and add
	    res_curr = 0.0;
	    for( int t = 0; t < B_sz_t; t++ )
		res_curr += A_mat(frontLayerIdx,t) * B_mat(t,backLayerIdx);
	    
	    //Write back
	    AB_tripletList.at(s) = T( frontLayerIdx, 
				      backLayerIdx, 
				      res_curr );
	}
	AB_mat.setFromTriplets(AB_tripletList.begin(), AB_tripletList.end());
	AB_mat.makeCompressed(); //Optional
	
	//##########################################################
	//#### UPDATE A

	// update A
	upper_A = L_mat * (B_mat.bottomRows(numRightCols)).transpose() - 
			   lambdaA * A_mat.rightCols(numRightCols);
			 //lambdaA * ( A_mat.rightCols(numRightCols) - Aprev_mat.rightCols(numRightCols) );
	upper_A = upper_A.cwiseMax( epsilon ); //Threshold	

	//Multiply
	A_mat.rightCols(numRightCols) = (A_mat.rightCols(numRightCols)).array() * upper_A.array();
	
	//Divide
	upper_A = AB_mat * (B_mat.bottomRows(numRightCols)).transpose();
	A_mat.rightCols(numRightCols) = (A_mat.rightCols(numRightCols)).
					cwiseQuotient( upper_A.cwiseMax( epsilonLower ) );
	
	// clamp to feasible region
	A_mat = A_mat.cwiseMax( 0.0 ); 
	A_mat = A_mat.cwiseMin( 1.0 ); 

	//##########################################################
	// update B
	upper_B = (L_mat.transpose() * A_mat.rightCols(numRightCols)).transpose() - 
		  lambdaB * B_mat.bottomRows(numRightCols);
		  //lambdaB * ( B_mat.bottomRows(numRightCols) - Bprev_mat.bottomRows(numRightCols) );
	upper_B = upper_B.cwiseMax( epsilon ); //Threshold	
	
	//Multiply
	B_mat.bottomRows(numRightCols) = (B_mat.bottomRows(numRightCols)).array() * upper_B.array();
	
	//Divide
	upper_B = (AB_mat.transpose() * A_mat.rightCols(numRightCols)).transpose();
	B_mat.bottomRows(numRightCols) = (B_mat.bottomRows(numRightCols)).
					  cwiseQuotient( upper_B.cwiseMax( epsilonLower ) );
	
	// clamp to feasible region
	B_mat = B_mat.cwiseMax( 0.0 ); 
	B_mat = B_mat.cwiseMin( 1.0 ); 
	
	//Compute relative change
	rel_change = (A_mat - A_mat_old).norm() / A_mat_old.norm() +
		     (B_mat - B_mat_old).norm() / B_mat_old.norm();
	  
	if( rel_change < *tol )
	  break;
	  
	//Debug
	//printf("Iter %d, Change %3.3f \n", iter, rel_change);
    }

    //Debug
    printf("Final iter %d, Change %3.3f \n", iter, rel_change);
	
    //###################################################
    //####### Return the results
	
    //A
    for( int t = 0; t < A_sz_t; t++ )
	for( int x = 0; x < A_sz_layer; x++ )
	    resA[t * A_sz_layer + x] = A_mat(x,t);
      
    //B
    for( int x = 0; x < B_sz_layer; x++ )
	for( int t = 0; t < B_sz_t; t++ )
	    resB[x * B_sz_t + t] = B_mat(t,x);
    
    return;
}
