/*
    Lightfield for PBRT
 */

#include "stdafx.h"
#include "scene.h"
#include "geometry.h"
#include "lightfield.h"


// Lightfield Method Definitions

Lightfield::Lightfield( const AnimatedTransform &cam2world, Scene *scene, LightfieldFilm* lfilm,
			float sX, float sY,
			float sopen, float sclose)
    : Camera(cam2world, sopen, sclose, lfilm),
      sizeX(sX), sizeY(sY)
{
    if (lfilm->xResolution < 1 || lfilm->yResolution < 1 ||
	lfilm->resAngleX < 1 || lfilm->resAngleX < 1)
        Severe("Lightfield resolution smaller than one !\n");
      
    if (CameraToWorld.HasScale())
        Warning("Unexpected scaling detected in lightfield-to-world transformation!\n");
    
    //Set scene
    extWorldBound = NULL;
    setScene(scene);
}

Lightfield::~Lightfield() {
    if(extWorldBound != NULL)
      delete extWorldBound;
}

void Lightfield::setScene(Scene *sc)
{
    if( sc == NULL )
      return;
    
    //Set scene
    scene = sc;
    
    //Set extended scene bound
    extWorldBound = new BBox;
}

float Lightfield::GenerateRay(const CameraSample &s, Ray *ray) const
{
    //Sanity check
    if (extWorldBound == NULL)
    {
        Severe("Scene not set for lightfield!\n");
	return 1.f;
    }
    
    //Cast
    LightfieldFilm *lf = (LightfieldFilm*)film;
    
    //Compute according ray from lightfield sample (s is all continuous between [0,size])
    //Origin
    Point origin(s.imageX * (sizeX / (float)(lf->xResolution)), -s.imageY * (sizeY / (float)(lf->yResolution)), 0.f);
    
    //Dir (no sampling here, but just nearest neighbor)
    //Convert to v-units
    float angleMinXV = tan( Radians(lf->angleMinX) );
    float angleMaxXV = tan( Radians(lf->angleMaxX) );
    float angleMinYV = tan( Radians(lf->angleMinY) );
    float angleMaxYV = tan( Radians(lf->angleMaxY) );
    
    //Actual v-samples
    float vx = ( floor(s.lensU) + 0.5f ) / ((float)(lf->resAngleX)) * ( angleMaxXV - angleMinXV ) + angleMinXV;
    float vy = ( floor(s.lensV) + 0.5f ) / ((float)(lf->resAngleY)) * ( angleMaxYV - angleMinYV ) + angleMinYV;
    
    //Compute direction
    Vector dir( vx, -vy, 1.0f); //From lightfield out to front
    //Vector dir(0.f, 0.f, -1.0f); //Straight out to front
    
    *ray = Ray(origin, Normalize(dir), 0.f, INFINITY);

    //Compute first with zero ray time
    ray->time = 0.f;
    
    //Get rayOffset if set
    float rayOffset = lf->rayOffset;
    
    if( rayOffset < 0.f )
    {   
	//If not defined transfor and hit with extended bbox
	
	//Convert to world
	CameraToWorld(*ray, ray);
      
	//Compute hit point with bbox
	//Generate bbox
	Transform currCameraToWorld;
	CameraToWorld.Interpolate(s.time, &currCameraToWorld);
	*extWorldBound = scene->WorldBound();
	
	float pXS = (sizeX / (float)(film->xResolution));
	float pYS = (sizeY / (float)(film->yResolution));
	*extWorldBound = Union(*extWorldBound, currCameraToWorld( Point(0.f - pXS, -(0.f - pYS), 0.f) ));
	*extWorldBound = Union(*extWorldBound, currCameraToWorld( Point(0.f - pXS, -(sizeY + pYS), 0.f) ));
	*extWorldBound = Union(*extWorldBound, currCameraToWorld( Point(sizeX + pXS, -(0.f - pYS), 0.f) ));
	*extWorldBound = Union(*extWorldBound, currCameraToWorld( Point(sizeX + pXS, -(sizeY + pYS), 0.f) ));
	
	//Compute hitpoint on bbox
	float hitt0, hitt1;
	extWorldBound->IntersectP(*ray, &hitt0, &hitt1);
	
	//Check for sanity
	if( hitt1 >= INFINITY)
	  Severe("Lightfield out of global boundingbox!\n");
	
	//Compute the intersection now
	Point boundIntersection = (*ray)(hitt1);
	
	//Store final bound ray position and direction
	ray->o = boundIntersection;
	ray->d = -ray->d;
    }
    else
    {
	//If just offset, then offset ray siimply
	
	//Offset the origin
	ray->o = ray->o + dir * rayOffset; //Rectangular box of width rayOffset around lightfield
	
	//Convert to world
	CameraToWorld(*ray, ray);
	
	//Invert
	ray->d = -ray->d;
    }
    
    //Assign now the actual ray time
    ray->time = s.time;
  
    return 1.f;
}