
/*
    pbrt source code Copyright(c) 1998-2012 Matt Pharr and Greg Humphreys.

    This file is part of pbrt.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

    - Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    - Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
    IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
    TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */


// core/api.cpp*
#include "api.h"

// API Additional Headers
#include "accelerators/bvh.h"
#include "accelerators/grid.h"
#include "accelerators/kdtreeaccel.h"
#include "cameras/environment.h"
#include "cameras/orthographic.h"
#include "cameras/perspective.h"
#include "cameras/lightfield.h"
#include "film/image.h"
#include "film/lightfield_film.h"
#include "film/lightfield_reconstruction_data.h"
#include "film/lightfield_reconstruction_layered_attenuator.h"
#include "film/lightfield_reconstruction_timemultiplexed_layered_attenuator.h"
#include "filters/box.h"
#include "filters/gaussian.h"
#include "filters/mitchell.h"
#include "filters/sinc.h"
#include "filters/triangle.h"
#include "integrators/ambientocclusion.h"
#include "integrators/diffuseprt.h"
#include "integrators/dipolesubsurface.h"
#include "integrators/directlighting.h"
#include "integrators/emission.h"
#include "integrators/glossyprt.h"
#include "integrators/igi.h"
#include "integrators/irradiancecache.h"
#include "integrators/path.h"
#include "integrators/photonmap.h"
#include "integrators/single.h"
#include "integrators/useprobes.h"
#include "integrators/whitted.h"
#include "lights/diffuse.h"
#include "lights/distant.h"
#include "lights/goniometric.h"
#include "lights/infinite.h"
#include "lights/point.h"
#include "lights/projection.h"
#include "lights/spot.h"
#include "materials/glass.h"
#include "materials/kdsubsurface.h"
#include "materials/matte.h"
#include "materials/measured.h"
#include "materials/metal.h"
#include "materials/mirror.h"
#include "materials/mixmat.h"
#include "materials/plastic.h"
#include "materials/substrate.h"
#include "materials/subsurface.h"
#include "materials/shinymetal.h"
#include "materials/translucent.h"
#include "materials/uber.h"
#include "renderers/aggregatetest.h"
#include "renderers/createprobes.h"
#include "renderers/metropolis.h"
#include "renderers/samplerrenderer.h"
#include "renderers/adaptiveLFsamplerrenderer.h"
#include "renderers/surfacepoints.h"
#include "samplers/adaptive.h"
#include "samplers/bestcandidate.h"
#include "samplers/halton.h"
#include "samplers/lowdiscrepancy.h"
#include "samplers/lowdiscrepancy_lightfield.h"
#include "samplers/random.h"
#include "samplers/stratified.h"
#include "shapes/cone.h"
#include "shapes/cylinder.h"
#include "shapes/disk.h"
#include "shapes/heightfield.h"
#include "shapes/hyperboloid.h"
#include "shapes/loopsubdiv.h"
#include "shapes/nurbs.h"
#include "shapes/paraboloid.h"
#include "shapes/sphere.h"
#include "shapes/trianglemesh.h"
#include "textures/bilerp.h"
#include "textures/checkerboard.h"
#include "textures/constant.h"
#include "textures/dots.h"
#include "textures/fbm.h"
#include "textures/imagemap.h"
#include "textures/marble.h"
#include "textures/mix.h"
#include "textures/scale.h"
#include "textures/uv.h"
#include "textures/windy.h"
#include "textures/wrinkled.h"
#include "volumes/exponential.h"
#include "volumes/homogeneous.h"
#include "volumes/volumegrid.h"
#include <map>
 #if (_MSC_VER >= 1400)
 #include <stdio.h>
 #define snprintf _snprintf
 #endif
 
#include <sys/time.h>

using std::map;

//###################################
//######### API Global state
//###################################

// API Global Variables
Options PbrtOptions;

// API Static Data
static int currentApiState = STATE_UNINITIALIZED;
static TransformSet curTransform;
static Point curLookAtPoint;
static bool lookAtDone = false;
static int activeTransformBits = ALL_TRANSFORMS_BITS;
static map<string, TransformSet> namedCoordinateSystems;
static RenderOptions *renderOptions = NULL;
static GraphicsState graphicsState;
static vector<GraphicsState> pushedGraphicsStates;
static vector<TransformSet> pushedTransforms;
static vector<uint32_t> pushedActiveTransformBits;
static TransformCache transformCache;

static Scene *curScene = NULL;
static Camera *curCamera = NULL;
static Renderer *curRenderer = NULL;
static Sampler* curSampler = NULL;
static Lightfield *curLightfield = NULL;
static int curReconstructionType = 0;
static int curDuplicateCheck = 0;
static LightfieldReconstruction *curLightfieldRec = NULL;
static SamplingParams samplingParams;
static LayerRecParams *layerRecParams = NULL;

// API Macros
#define VERIFY_INITIALIZED(func) \
if (currentApiState == STATE_UNINITIALIZED) { \
    Error("pbrtInit() must be before calling \"%s()\". " \
          "Ignoring.", func); \
    return; \
} else /* swallow trailing semicolon */
#define VERIFY_OPTIONS(func) \
VERIFY_INITIALIZED(func); \
if (currentApiState == STATE_WORLD_BLOCK) { \
    Error("Options cannot be set inside world block; " \
          "\"%s\" not allowed.  Ignoring.", func); \
    return; \
} else /* swallow trailing semicolon */
#define VERIFY_WORLD(func) \
VERIFY_INITIALIZED(func); \
if (currentApiState == STATE_OPTIONS_BLOCK) { \
    Error("Scene description must be inside world block; " \
          "\"%s\" not allowed. Ignoring.", func); \
    return; \
} else /* swallow trailing semicolon */
#define FOR_ACTIVE_TRANSFORMS(expr) \
    for (int i = 0; i < MAX_TRANSFORMS; ++i) \
        if (activeTransformBits & (1 << i)) { expr }
#define WARN_IF_ANIMATED_TRANSFORM(func) \
do { if (curTransform.IsAnimated()) \
         Warning("Animated transformations set; ignoring for \"%s\"" \
                 "and using the start transform only", func); \
} while (false)

//###################################
//######### API Class implementations
//###################################

RenderOptions::RenderOptions() {
    // RenderOptions Constructor Implementation
    transformStartTime = 0.f;
    transformEndTime = 1.f;
    FilterName = "box";
    FilmName = "image";
    SamplerName = "lowdiscrepancy";
    AcceleratorName = "bvh";
    RendererName = "sampler";
    SurfIntegratorName = "directlighting";
    VolIntegratorName = "emission";
    CameraName = "perspective";
    currentInstance = NULL;
    CameraLookAtPoint = Point(0.f, 0.f, 0.f);
}


GraphicsState::GraphicsState() {
    // GraphicsState Constructor Implementation
    material = "matte";
    reverseOrientation = false;
}

SamplingParams::SamplingParams()
{
    //Default sampling params
    numChains = 1000;
    chainLength = 10;
    maxAcceptedSamplesPerChains = 10;
    
    mutation_sigma_spatial = 1.f;
    mutation_sigma_angular = 1.f;
    thresh_residual = 0.01;
    
    numChainDuplicateChecks = 0;
    replaceSamples = false;
    
    sampleTime = false;	
    sampleFull = false;
}

LayerRecParams::LayerRecParams()
{
    //Default params
    //Lightfield stuff
    OX = 0.f;
    OY = 0.f; 
    OZ = 0.f;
    
    lightfieldSzX = 0.f;
    lightfieldSzY = 0.f;
		
    //Layer stuff
    layerResX = 0; 
    layerResY = 0; 
    layerResZ = 0;
    
    layerOX = 0.f;
    layerOY = 0.f;
    layerOZ = 0.f;
    
    layerSzX = 0.f;
    layerSzY = 0.f;
    
    layerDst = 0.f;
    
    scaleRec = 1.f;
    
    basisFunction = 0;
    
    timeStps = 0;
    layerSubSet.clear();
}

//###################################
//######### API Local functions
//###################################

// Object Creation Function Definitions
Reference<Shape> MakeShape(const string &name,
        const Transform *object2world, const Transform *world2object,
        bool reverseOrientation, const ParamSet &paramSet) {
    Shape *s = NULL;

    if (name == "sphere")
        s = CreateSphereShape(object2world, world2object,
                              reverseOrientation, paramSet);
    // Create remaining _Shape_ types
    else if (name == "cylinder")
        s = CreateCylinderShape(object2world, world2object, reverseOrientation,
                                paramSet);
    else if (name == "disk")
        s = CreateDiskShape(object2world, world2object, reverseOrientation,
                            paramSet);
    else if (name == "cone")
        s = CreateConeShape(object2world, world2object, reverseOrientation,
                            paramSet);
    else if (name == "paraboloid")
        s = CreateParaboloidShape(object2world, world2object, reverseOrientation,
                                  paramSet);
    else if (name == "hyperboloid")
        s = CreateHyperboloidShape(object2world, world2object, reverseOrientation,
                                   paramSet);
    else if (name == "trianglemesh")
        s = CreateTriangleMeshShape(object2world, world2object, reverseOrientation,
                                    paramSet, &graphicsState.floatTextures);
    else if (name == "heightfield")
        s = CreateHeightfieldShape(object2world, world2object, reverseOrientation,
                                   paramSet);
    else if (name == "loopsubdiv")
        s = CreateLoopSubdivShape(object2world, world2object, reverseOrientation,
                                  paramSet);
    else if (name == "nurbs")
        s = CreateNURBSShape(object2world, world2object, reverseOrientation,
                             paramSet);
    else
        Warning("Shape \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return s;
}


Reference<Material> MakeMaterial(const string &name,
        const Transform &mtl2world,
        const TextureParams &mp) {
    Material *material = NULL;
    if (name == "matte")
        material = CreateMatteMaterial(mtl2world, mp);
    else if (name == "plastic")
        material = CreatePlasticMaterial(mtl2world, mp);
    else if (name == "translucent")
        material = CreateTranslucentMaterial(mtl2world, mp);
    else if (name == "glass")
        material = CreateGlassMaterial(mtl2world, mp);
    else if (name == "mirror")
        material = CreateMirrorMaterial(mtl2world, mp);
    else if (name == "mix") {
        string m1 = mp.FindString("namedmaterial1", "");
        string m2 = mp.FindString("namedmaterial2", "");
        Reference<Material> mat1 = graphicsState.namedMaterials[m1];
        Reference<Material> mat2 = graphicsState.namedMaterials[m2];
        if (!mat1) {
            Error("Named material \"%s\" undefined.  Using \"matte\"",
                  m1.c_str());
            mat1 = MakeMaterial("matte", curTransform[0], mp);
        }
        if (!mat2) {
            Error("Named material \"%s\" undefined.  Using \"matte\"",
                  m2.c_str());
            mat2 = MakeMaterial("matte", curTransform[0], mp);
        }

        material = CreateMixMaterial(mtl2world, mp, mat1, mat2);
    }
    else if (name == "metal")
        material = CreateMetalMaterial(mtl2world, mp);
    else if (name == "substrate")
        material = CreateSubstrateMaterial(mtl2world, mp);
    else if (name == "uber")
        material = CreateUberMaterial(mtl2world, mp);
    else if (name == "subsurface")
        material = CreateSubsurfaceMaterial(mtl2world, mp);
    else if (name == "kdsubsurface")
        material = CreateKdSubsurfaceMaterial(mtl2world, mp);
    else if (name == "measured")
        material = CreateMeasuredMaterial(mtl2world, mp);
    else if (name == "shinymetal")
        material = CreateShinyMetalMaterial(mtl2world, mp);
    else
        Warning("Material \"%s\" unknown.", name.c_str());
    mp.ReportUnused();
    if (!material) Error("Unable to create material \"%s\"", name.c_str());
    return material;
}


Reference<Texture<float> > MakeFloatTexture(const string &name,
        const Transform &tex2world, const TextureParams &tp) {
    Texture<float> *tex = NULL;
    if (name == "constant")
        tex = CreateConstantFloatTexture(tex2world, tp);
    else if (name == "scale")
        tex = CreateScaleFloatTexture(tex2world, tp);
    else if (name == "mix")
        tex = CreateMixFloatTexture(tex2world, tp);
    else if (name == "bilerp")
        tex = CreateBilerpFloatTexture(tex2world, tp);
    else if (name == "imagemap")
        tex = CreateImageFloatTexture(tex2world, tp);
    else if (name == "uv")
        tex = CreateUVFloatTexture(tex2world, tp);
    else if (name == "checkerboard")
        tex = CreateCheckerboardFloatTexture(tex2world, tp);
    else if (name == "dots")
        tex = CreateDotsFloatTexture(tex2world, tp);
    else if (name == "fbm")
        tex = CreateFBmFloatTexture(tex2world, tp);
    else if (name == "wrinkled")
        tex = CreateWrinkledFloatTexture(tex2world, tp);
    else if (name == "marble")
        tex = CreateMarbleFloatTexture(tex2world, tp);
    else if (name == "windy")
        tex = CreateWindyFloatTexture(tex2world, tp);
    else
        Warning("Float texture \"%s\" unknown.", name.c_str());
    tp.ReportUnused();
    return tex;
}


Reference<Texture<Spectrum> > MakeSpectrumTexture(const string &name,
        const Transform &tex2world, const TextureParams &tp) {
    Texture<Spectrum> *tex = NULL;
    if (name == "constant")
        tex = CreateConstantSpectrumTexture(tex2world, tp);
    else if (name == "scale")
        tex = CreateScaleSpectrumTexture(tex2world, tp);
    else if (name == "mix")
        tex = CreateMixSpectrumTexture(tex2world, tp);
    else if (name == "bilerp")
        tex = CreateBilerpSpectrumTexture(tex2world, tp);
    else if (name == "imagemap")
        tex = CreateImageSpectrumTexture(tex2world, tp);
    else if (name == "uv")
        tex = CreateUVSpectrumTexture(tex2world, tp);
    else if (name == "checkerboard")
        tex = CreateCheckerboardSpectrumTexture(tex2world, tp);
    else if (name == "dots")
        tex = CreateDotsSpectrumTexture(tex2world, tp);
    else if (name == "fbm")
        tex = CreateFBmSpectrumTexture(tex2world, tp);
    else if (name == "wrinkled")
        tex = CreateWrinkledSpectrumTexture(tex2world, tp);
    else if (name == "marble")
        tex = CreateMarbleSpectrumTexture(tex2world, tp);
    else if (name == "windy")
        tex = CreateWindySpectrumTexture(tex2world, tp);
    else
        Warning("Spectrum texture \"%s\" unknown.", name.c_str());
    tp.ReportUnused();
    return tex;
}


Light *MakeLight(const string &name,
        const Transform &light2world, const ParamSet &paramSet) {
    Light *light = NULL;
    if (name == "point")
        light = CreatePointLight(light2world, paramSet);
    else if (name == "spot")
        light = CreateSpotLight(light2world, paramSet);
    else if (name == "goniometric")
        light = CreateGoniometricLight(light2world, paramSet);
    else if (name == "projection")
        light = CreateProjectionLight(light2world, paramSet);
    else if (name == "distant")
        light = CreateDistantLight(light2world, paramSet);
    else if (name == "infinite" || name == "exinfinite")
        light = CreateInfiniteLight(light2world, paramSet);
    else
        Warning("Light \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return light;
}


AreaLight *MakeAreaLight(const string &name,
        const Transform &light2world, const ParamSet &paramSet,
        const Reference<Shape> &shape) {
    AreaLight *area = NULL;
    if (name == "area" || name == "diffuse")
        area = CreateDiffuseAreaLight(light2world, paramSet, shape);
    else
        Warning("Area light \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return area;
}


VolumeRegion *MakeVolumeRegion(const string &name,
        const Transform &volume2world, const ParamSet &paramSet) {
    VolumeRegion *vr = NULL;
    if (name == "homogeneous")
        vr = CreateHomogeneousVolumeDensityRegion(volume2world, paramSet);
    else if (name == "volumegrid")
        vr = CreateGridVolumeRegion(volume2world, paramSet);
    else if (name == "exponential")
        vr = CreateExponentialVolumeRegion(volume2world, paramSet);
    else
        Warning("Volume region \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return vr;
}


SurfaceIntegrator *MakeSurfaceIntegrator(const string &name,
        const ParamSet &paramSet) {
    SurfaceIntegrator *si = NULL;
    if (name == "whitted")
        si = CreateWhittedSurfaceIntegrator(paramSet);
    else if (name == "directlighting")
        si = CreateDirectLightingIntegrator(paramSet);
    else if (name == "path")
        si = CreatePathSurfaceIntegrator(paramSet);
    else if (name == "photonmap" || name == "exphotonmap")
        si = CreatePhotonMapSurfaceIntegrator(paramSet);
    else if (name == "irradiancecache")
        si = CreateIrradianceCacheIntegrator(paramSet);
    else if (name == "igi")
        si = CreateIGISurfaceIntegrator(paramSet);
    else if (name == "dipolesubsurface")
        si = CreateDipoleSubsurfaceIntegrator(paramSet);
    else if (name == "ambientocclusion")
        si = CreateAmbientOcclusionIntegrator(paramSet);
    else if (name == "useprobes")
        si = CreateRadianceProbesSurfaceIntegrator(paramSet);
    else if (name == "diffuseprt")
        si = CreateDiffusePRTIntegratorSurfaceIntegrator(paramSet);
    else if (name == "glossyprt")
        si = CreateGlossyPRTIntegratorSurfaceIntegrator(paramSet);
    else
        Warning("Surface integrator \"%s\" unknown.", name.c_str());

    paramSet.ReportUnused();
    return si;
}


VolumeIntegrator *MakeVolumeIntegrator(const string &name,
        const ParamSet &paramSet) {
    VolumeIntegrator *vi = NULL;
    if (name == "single")
        vi = CreateSingleScatteringIntegrator(paramSet);
    else if (name == "emission")
        vi = CreateEmissionVolumeIntegrator(paramSet);
    else
        Warning("Volume integrator \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return vi;
}


Primitive *MakeAccelerator(const string &name,
        const vector<Reference<Primitive> > &prims,
        const ParamSet &paramSet) {
    Primitive *accel = NULL;
    if (name == "bvh")
        accel = CreateBVHAccelerator(prims, paramSet);
    else if (name == "grid")
        accel = CreateGridAccelerator(prims, paramSet);
    else if (name == "kdtree")
        accel = CreateKdTreeAccelerator(prims, paramSet);
    else
        Warning("Accelerator \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return accel;
}


Camera *MakeCamera(const string &name,
        const ParamSet &paramSet,
        const TransformSet &cam2worldSet, float transformStart,
        float transformEnd, Film *film) {
    Camera *camera = NULL;
    Assert(MAX_TRANSFORMS == 2);
    
    //Lookup for report
    bool centerScanLine = paramSet.FindOneBool("centerScanline", false);
    
    //Regular
    Transform *cam2world[2];
    transformCache.Lookup(cam2worldSet[0], &cam2world[0], NULL);
    transformCache.Lookup(cam2worldSet[1], &cam2world[1], NULL);
    AnimatedTransform animatedCam2World(cam2world[0], transformStart,
        cam2world[1], transformEnd);
    if (name == "perspective")
        camera = CreatePerspectiveCamera(paramSet, animatedCam2World, film);
    else if (name == "orthographic")
        camera = CreateOrthographicCamera(paramSet, animatedCam2World, film);
    else if (name == "environment")
        camera = CreateEnvironmentCamera(paramSet, animatedCam2World, film);
    else
        Warning("Camera \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return camera;
}


Sampler *MakeSampler(const string &name,
        const ParamSet &paramSet, const Film *film, const Camera *camera, const Lightfield* lf) {
    //Sanity
    if( lf != NULL && name != "lowdiscrepancy")
	Severe("Error: Sampler is not implemented for lightfield sampling!\n");
    
    Sampler *sampler = NULL;
    if (name == "adaptive")
        sampler = CreateAdaptiveSampler(paramSet, film, camera);
    else if (name == "bestcandidate")
        sampler = CreateBestCandidateSampler(paramSet, film, camera);
    else if (name == "halton")
        sampler = CreateHaltonSampler(paramSet, film, camera);
    else if (name == "lowdiscrepancy")
    {
	if( lf != NULL)
	    sampler = CreateLowDiscrepancySamplerLightfield(paramSet, (LightfieldFilm*)film, lf);
	else
	    sampler = CreateLowDiscrepancySampler(paramSet, film, camera);
    }
    else if (name == "random")
        sampler = CreateRandomSampler(paramSet, film, camera);
    else if (name == "stratified")
        sampler = CreateStratifiedSampler(paramSet, film, camera);
    else
        Warning("Sampler \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return sampler;
}


Filter *MakeFilter(const string &name,
    const ParamSet &paramSet) {
    Filter *filter = NULL;
    if (name == "box")
        filter = CreateBoxFilter(paramSet);
    else if (name == "gaussian")
        filter = CreateGaussianFilter(paramSet);
    else if (name == "mitchell")
        filter = CreateMitchellFilter(paramSet);
    else if (name == "sinc")
        filter = CreateSincFilter(paramSet);
    else if (name == "triangle")
        filter = CreateTriangleFilter(paramSet);
    else
        Warning("Filter \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return filter;
}


Film *MakeFilm(const string &name,
    const ParamSet &paramSet, Filter *filter) {
    Film *film = NULL;
    if (name == "image")
    {
        film = CreateImageFilm(paramSet, filter);
    }
    else if (name == "lightfield")
    {
        film = CreateLightfieldFilm( paramSet, !(curReconstructionType > 0) );
	if(!lookAtDone)
	  Severe("Lookat point needed by lightfield film!\n");
    }
    else
        Warning("Film \"%s\" unknown.", name.c_str());
    paramSet.ReportUnused();
    return film;
}

//###################################
//######### API Function definitions
//###################################

void pbrtSaveAPIState(APIState &st)
{
    //Save api state variables
    st.PbrtOptions = PbrtOptions;

    // API Static Data
    st.currentApiState = currentApiState;
    st.curTransform = curTransform;
    st.curLookAtPoint = curLookAtPoint;
    st.lookAtDone = lookAtDone;
    st.activeTransformBits = activeTransformBits;
    st.namedCoordinateSystems = namedCoordinateSystems;
    st.renderOptions = renderOptions;
    st.graphicsState = graphicsState;
    st.pushedGraphicsStates = pushedGraphicsStates;
    st.pushedTransforms = pushedTransforms;
    st.pushedActiveTransformBits = pushedActiveTransformBits;
    st.transformCache = transformCache;

    st.curScene = curScene;
    st.curCamera = curCamera;
    st.curRenderer = curRenderer;
    st.curSampler = curSampler;
    st.curLightfield = curLightfield;
    st.curReconstructionType = curReconstructionType;
    st.curDuplicateCheck = curDuplicateCheck;
    st.curLightfieldRec = curLightfieldRec;
    st.samplingParams = samplingParams;
    st.layerRecParams = layerRecParams;
}

void pbrtRestoreAPIState(const APIState &st)
{
    //Save api state variables
    PbrtOptions = st.PbrtOptions;

    // API Static Data
    currentApiState = st.currentApiState;
    curTransform = st.curTransform;
    curLookAtPoint = st.curLookAtPoint;
    lookAtDone = st.lookAtDone;
    activeTransformBits = st.activeTransformBits;
    namedCoordinateSystems = st.namedCoordinateSystems;
    renderOptions = st.renderOptions;
    graphicsState = st.graphicsState; 
    pushedGraphicsStates = st.pushedGraphicsStates;
    pushedTransforms = st.pushedTransforms;
    pushedActiveTransformBits = st.pushedActiveTransformBits;
    transformCache = st.transformCache;
    
    curScene = st.curScene;
    curCamera = st.curCamera;
    curRenderer = st.curRenderer;
    curSampler = st.curSampler;
    curLightfield = st.curLightfield;
    curReconstructionType = st.curReconstructionType;
    curDuplicateCheck = st.curDuplicateCheck;
    curLightfieldRec = st.curLightfieldRec;
    samplingParams = st.samplingParams;
    layerRecParams = st.layerRecParams;
    
    //Spectrum init
    SampledSpectrum::Init();
}

//Debug function
/*
void pbrtDestroyState()
{
    // API Global Variables
    PbrtOptions = Options();

    // API Static Data
    currentApiState = STATE_UNINITIALIZED;
    curTransform = TransformSet();
    curLookAtPoint = Point();
    lookAtDone = false;
    activeTransformBits = ALL_TRANSFORMS_BITS;
    namedCoordinateSystems =  map<string, TransformSet>();
    renderOptions = NULL;
    graphicsState = GraphicsState();
    pushedGraphicsStates = vector<GraphicsState>();
    pushedTransforms = vector<TransformSet>();
    pushedActiveTransformBits = vector<uint32_t>();
    transformCache = TransformCache();

    curScene = NULL;
    curCamera = NULL;
    curRenderer = NULL;
    curLightfield = NULL;
    curLightfieldRec = NULL;
}
*/

void pbrtInit(const Options &opt) {
    PbrtOptions = opt;
    
    // API Initialization
    if (currentApiState != STATE_UNINITIALIZED)
        Error("pbrtInit() has already been called.");
    currentApiState = STATE_OPTIONS_BLOCK;
    renderOptions = new RenderOptions;
    graphicsState = GraphicsState();
    samplingParams = SamplingParams();
    layerRecParams = new LayerRecParams();
    
    //Initialize random number generators
    struct timeval tv;
    gettimeofday( &tv, NULL );
    long daytime_in_microseconds = long(tv.tv_sec)*1e6 + long(tv.tv_usec);

    //Seed
    srand48( daytime_in_microseconds );
    
    SampledSpectrum::Init();
}


void pbrtCleanup() {
    ProbesCleanup();
    // API Cleanup
    if (currentApiState == STATE_UNINITIALIZED)
        Error("pbrtCleanup() called without pbrtInit().");
    else if (currentApiState == STATE_WORLD_BLOCK)
        Error("pbrtCleanup() called while inside world block.");
    currentApiState = STATE_UNINITIALIZED;
    
    delete renderOptions;
    renderOptions = NULL;
    
    delete layerRecParams;
    layerRecParams = NULL;
}


void pbrtIdentity() {
    VERIFY_INITIALIZED("Identity");
    FOR_ACTIVE_TRANSFORMS(curTransform[i] = Transform();)
}


void pbrtTranslate(float dx, float dy, float dz) {
    VERIFY_INITIALIZED("Translate");
    FOR_ACTIVE_TRANSFORMS(curTransform[i] =
        curTransform[i] * Translate(Vector(dx, dy, dz));)
}


void pbrtTransform(float tr[16]) {
    VERIFY_INITIALIZED("Transform");
    FOR_ACTIVE_TRANSFORMS(curTransform[i] = Transform(Matrix4x4(
        tr[0], tr[4], tr[8], tr[12],
        tr[1], tr[5], tr[9], tr[13],
        tr[2], tr[6], tr[10], tr[14],
        tr[3], tr[7], tr[11], tr[15]));)
}


void pbrtConcatTransform(float tr[16]) {
    VERIFY_INITIALIZED("ConcatTransform");
    FOR_ACTIVE_TRANSFORMS(curTransform[i] = curTransform[i] * Transform(
                Matrix4x4(tr[0], tr[4], tr[8], tr[12],
                          tr[1], tr[5], tr[9], tr[13],
                          tr[2], tr[6], tr[10], tr[14],
                          tr[3], tr[7], tr[11], tr[15]));)
}


void pbrtRotate(float angle, float dx, float dy, float dz) {
    VERIFY_INITIALIZED("Rotate");
    FOR_ACTIVE_TRANSFORMS(curTransform[i] = curTransform[i] * Rotate(angle, Vector(dx, dy, dz));)
}


void pbrtScale(float sx, float sy, float sz) {
    VERIFY_INITIALIZED("Scale");
    FOR_ACTIVE_TRANSFORMS(curTransform[i] = curTransform[i] * Scale(sx, sy, sz);)
}


void pbrtLookAt(float ex, float ey, float ez, float lx, float ly,
        float lz, float ux, float uy, float uz) {
    VERIFY_INITIALIZED("LookAt");
    FOR_ACTIVE_TRANSFORMS({ Warning("This version of pbrt fixes a bug in the LookAt transformation.\n"
                                    "If your rendered images unexpectedly change, add a \"Scale -1 1 1\"\n"
                                    "to the start of your scene file."); break; })
    FOR_ACTIVE_TRANSFORMS(curTransform[i] =
        curTransform[i] * LookAt(Point(ex, ey, ez), Point(lx, ly, lz), Vector(ux, uy, uz));)
    
    //Save curLookAt
    curLookAtPoint = Point(lx, ly, lz);
    lookAtDone = true;
}


void pbrtCoordinateSystem(const string &name) {
    VERIFY_INITIALIZED("CoordinateSystem");
    namedCoordinateSystems[name] = curTransform;
}


void pbrtCoordSysTransform(const string &name) {
    VERIFY_INITIALIZED("CoordSysTransform");
    if (namedCoordinateSystems.find(name) !=
        namedCoordinateSystems.end())
        curTransform = namedCoordinateSystems[name];
    else
        Warning("Couldn't find named coordinate system \"%s\"",
                name.c_str());
}


void pbrtActiveTransformAll() {
    activeTransformBits = ALL_TRANSFORMS_BITS;
}


void pbrtActiveTransformEndTime() {
    activeTransformBits = END_TRANSFORM_BITS;
}


void pbrtActiveTransformStartTime() {
    activeTransformBits = START_TRANSFORM_BITS;
}


void pbrtTransformTimes(float start, float end) {
    VERIFY_OPTIONS("TransformTimes");
    renderOptions->transformStartTime = start;
    renderOptions->transformEndTime = end;
}


void pbrtPixelFilter(const string &name, const ParamSet &params) {
    VERIFY_OPTIONS("PixelFilter");
    renderOptions->FilterName = name;
    renderOptions->FilterParams = params;
}


void pbrtFilm(const string &type, const ParamSet &params) {
    VERIFY_OPTIONS("Film");
    renderOptions->FilmParams = params;
    renderOptions->FilmName = type;
}


void pbrtSampler(const string &name, const ParamSet &params) {
    VERIFY_OPTIONS("Sampler");
    renderOptions->SamplerName = name;
    renderOptions->SamplerParams = params;
}


void pbrtAccelerator(const string &name, const ParamSet &params) {
    VERIFY_OPTIONS("Accelerator");
    renderOptions->AcceleratorName = name;
    renderOptions->AcceleratorParams = params;
}


void pbrtSurfaceIntegrator(const string &name, const ParamSet &params) {
    VERIFY_OPTIONS("SurfaceIntegrator");
    renderOptions->SurfIntegratorName = name;
    renderOptions->SurfIntegratorParams = params;
}


void pbrtVolumeIntegrator(const string &name, const ParamSet &params) {
    VERIFY_OPTIONS("VolumeIntegrator");
    renderOptions->VolIntegratorName = name;
    renderOptions->VolIntegratorParams = params;
}


void pbrtRenderer(const string &name, const ParamSet &params) {
    VERIFY_OPTIONS("Renderer");
    renderOptions->RendererName = name;
    renderOptions->RendererParams = params;
}


void pbrtCamera(const string &name, const ParamSet &params) {
    VERIFY_OPTIONS("Camera");
    renderOptions->CameraName = name;
    renderOptions->CameraParams = params;
    renderOptions->CameraToWorld = Inverse(curTransform);
    namedCoordinateSystems["camera"] = renderOptions->CameraToWorld;
    
    //Additional camera lookat point stored
    if(!lookAtDone)
      Warning("No lookat was done for camera. Lookat point needed by lightfield film\n");
    
    renderOptions->CameraLookAtPoint = curLookAtPoint;
}


void pbrtWorldBegin() {
    VERIFY_OPTIONS("WorldBegin");
    currentApiState = STATE_WORLD_BLOCK;
    for (int i = 0; i < MAX_TRANSFORMS; ++i)
        curTransform[i] = Transform();
    activeTransformBits = ALL_TRANSFORMS_BITS;
    namedCoordinateSystems["world"] = curTransform;
}


void pbrtAttributeBegin() {
    VERIFY_WORLD("AttributeBegin");
    pushedGraphicsStates.push_back(graphicsState);
    pushedTransforms.push_back(curTransform);
    pushedActiveTransformBits.push_back(activeTransformBits);
}


void pbrtAttributeEnd() {
    VERIFY_WORLD("AttributeEnd");
    if (!pushedGraphicsStates.size()) {
        Error("Unmatched pbrtAttributeEnd() encountered. "
              "Ignoring it.");
        return;
    }
    graphicsState = pushedGraphicsStates.back();
    pushedGraphicsStates.pop_back();
    curTransform = pushedTransforms.back();
    pushedTransforms.pop_back();
    activeTransformBits = pushedActiveTransformBits.back();
    pushedActiveTransformBits.pop_back();
}


void pbrtTransformBegin() {
    VERIFY_WORLD("TransformBegin");
    pushedTransforms.push_back(curTransform);
    pushedActiveTransformBits.push_back(activeTransformBits);
}


void pbrtTransformEnd() {
    VERIFY_WORLD("TransformEnd");
    if (!pushedTransforms.size()) {
        Error("Unmatched pbrtTransformEnd() encountered. "
            "Ignoring it.");
        return;
    }
    curTransform = pushedTransforms.back();
    pushedTransforms.pop_back();
    activeTransformBits = pushedActiveTransformBits.back();
    pushedActiveTransformBits.pop_back();
}


void pbrtTexture(const string &name, const string &type,
                 const string &texname, const ParamSet &params) {
    VERIFY_WORLD("Texture");
    TextureParams tp(params, params, graphicsState.floatTextures,
                     graphicsState.spectrumTextures);
    if (type == "float")  {
        // Create _float_ texture and store in _floatTextures_
        if (graphicsState.floatTextures.find(name) !=
            graphicsState.floatTextures.end())
            Info("Texture \"%s\" being redefined", name.c_str());
        WARN_IF_ANIMATED_TRANSFORM("Texture");
        Reference<Texture<float> > ft = MakeFloatTexture(texname,
                                                         curTransform[0], tp);
        if (ft) graphicsState.floatTextures[name] = ft;
    }
    else if (type == "color" || type == "spectrum")  {
        // Create _color_ texture and store in _spectrumTextures_
        if (graphicsState.spectrumTextures.find(name) != graphicsState.spectrumTextures.end())
            Info("Texture \"%s\" being redefined", name.c_str());
        WARN_IF_ANIMATED_TRANSFORM("Texture");
        Reference<Texture<Spectrum> > st = MakeSpectrumTexture(texname,
            curTransform[0], tp);
        if (st) graphicsState.spectrumTextures[name] = st;
    }
    else
        Error("Texture type \"%s\" unknown.", type.c_str());
}


void pbrtMaterial(const string &name, const ParamSet &params) {
    VERIFY_WORLD("Material");
    graphicsState.material = name;
    graphicsState.materialParams = params;
    graphicsState.currentNamedMaterial = "";
}


void pbrtMakeNamedMaterial(const string &name,
        const ParamSet &params) {
    VERIFY_WORLD("MakeNamedMaterial");
    // error checking, warning if replace, what to use for transform?
    TextureParams mp(params, graphicsState.materialParams,
                     graphicsState.floatTextures,
                     graphicsState.spectrumTextures);
    string matName = mp.FindString("type");
    WARN_IF_ANIMATED_TRANSFORM("MakeNamedMaterial");
    if (matName == "") Error("No parameter string \"type\" found in MakeNamedMaterial");
    else {
        Reference<Material> mtl = MakeMaterial(matName, curTransform[0], mp);
        if (mtl) graphicsState.namedMaterials[name] = mtl;
    }
}



void pbrtNamedMaterial(const string &name) {
    VERIFY_WORLD("NamedMaterial");
    graphicsState.currentNamedMaterial = name;
}


void pbrtLightSource(const string &name, const ParamSet &params) {
    VERIFY_WORLD("LightSource");
    WARN_IF_ANIMATED_TRANSFORM("LightSource");
    Light *lt = MakeLight(name, curTransform[0], params);
    if (lt == NULL)
        Error("pbrtLightSource: light type \"%s\" unknown.", name.c_str());
    else
        renderOptions->lights.push_back(lt);
}


void pbrtAreaLightSource(const string &name,
                         const ParamSet &params) {
    VERIFY_WORLD("AreaLightSource");
    graphicsState.areaLight = name;
    graphicsState.areaLightParams = params;
}


void pbrtShape(const string &name, const ParamSet &params) {
    VERIFY_WORLD("Shape");
    Reference<Primitive> prim;
    AreaLight *area = NULL;
    if (!curTransform.IsAnimated()) {
        // Create primitive for static shape
        Transform *obj2world, *world2obj;
        transformCache.Lookup(curTransform[0], &obj2world, &world2obj);
        Reference<Shape> shape = MakeShape(name, obj2world, world2obj,
            graphicsState.reverseOrientation, params);
        if (!shape) return;
        Reference<Material> mtl = graphicsState.CreateMaterial(params);
        params.ReportUnused();

        // Possibly create area light for shape
        if (graphicsState.areaLight != "") {
            area = MakeAreaLight(graphicsState.areaLight, curTransform[0],
                                 graphicsState.areaLightParams, shape);
        }
        prim = new GeometricPrimitive(shape, mtl, area);
    } else {
        // Create primitive for animated shape

        // Create initial _Shape_ for animated shape
        if (graphicsState.areaLight != "")
            Warning("Ignoring currently set area light when creating "
                    "animated shape");
        Transform *identity;
        transformCache.Lookup(Transform(), &identity, NULL);
        Reference<Shape> shape = MakeShape(name, identity, identity,
            graphicsState.reverseOrientation, params);
        if (!shape) return;
        Reference<Material> mtl = graphicsState.CreateMaterial(params);
        params.ReportUnused();

        // Get _animatedWorldToObject_ transform for shape
        Assert(MAX_TRANSFORMS == 2);
        Transform *world2obj[2];
        transformCache.Lookup(curTransform[0], NULL, &world2obj[0]);
        transformCache.Lookup(curTransform[1], NULL, &world2obj[1]);
        AnimatedTransform
             animatedWorldToObject(world2obj[0], renderOptions->transformStartTime,
                                   world2obj[1], renderOptions->transformEndTime);
        Reference<Primitive> baseprim = new GeometricPrimitive(shape, mtl, NULL);
        if (!baseprim->CanIntersect()) {
            // Refine animated shape and create BVH if more than one shape created
            vector<Reference<Primitive> > refinedPrimitives;
            baseprim->FullyRefine(refinedPrimitives);
            if (refinedPrimitives.size() == 0) return;
            if (refinedPrimitives.size() > 1)
                baseprim = new BVHAccel(refinedPrimitives);
            else
                baseprim = refinedPrimitives[0];
        }
        prim = new TransformedPrimitive(baseprim, animatedWorldToObject);
    }
    // Add primitive to scene or current instance
    if (renderOptions->currentInstance) {
        if (area)
            Warning("Area lights not supported with object instancing");
        renderOptions->currentInstance->push_back(prim);
    }
    
    else {
      
        if (area != NULL) 
	{
            renderOptions->lights.push_back(area);
	    if( area->geomVisible )
		renderOptions->primitives.push_back(prim);
        }else
	    renderOptions->primitives.push_back(prim);
	
    }
}


Reference<Material> GraphicsState::CreateMaterial(const ParamSet &params) {
    TextureParams mp(params, materialParams,
                     floatTextures,
                     spectrumTextures);
    Reference<Material> mtl;
    if (currentNamedMaterial != "" &&
        namedMaterials.find(currentNamedMaterial) != namedMaterials.end())
        mtl = namedMaterials[graphicsState.currentNamedMaterial];
    if (!mtl)
        mtl = MakeMaterial(material, curTransform[0], mp);
    if (!mtl)
        mtl = MakeMaterial("matte", curTransform[0], mp);
    if (!mtl)
        Severe("Unable to create \"matte\" material?!");
    return mtl;
}


void pbrtReverseOrientation() {
    VERIFY_WORLD("ReverseOrientation");
    graphicsState.reverseOrientation =
        !graphicsState.reverseOrientation;
}


void pbrtVolume(const string &name, const ParamSet &params) {
    VERIFY_WORLD("Volume");
    WARN_IF_ANIMATED_TRANSFORM("Volume");
    VolumeRegion *vr = MakeVolumeRegion(name, curTransform[0], params);
    if (vr) renderOptions->volumeRegions.push_back(vr);
}


void pbrtObjectBegin(const string &name) {
    VERIFY_WORLD("ObjectBegin");
    pbrtAttributeBegin();
    if (renderOptions->currentInstance)
        Error("ObjectBegin called inside of instance definition");
    renderOptions->instances[name] = vector<Reference<Primitive> >();
    renderOptions->currentInstance = &renderOptions->instances[name];
}


void pbrtObjectEnd() {
    VERIFY_WORLD("ObjectEnd");
    if (!renderOptions->currentInstance)
        Error("ObjectEnd called outside of instance definition");
    renderOptions->currentInstance = NULL;
    pbrtAttributeEnd();
}


void pbrtObjectInstance(const string &name) {
    VERIFY_WORLD("ObjectInstance");
    // Object instance error checking
    if (renderOptions->currentInstance) {
        Error("ObjectInstance can't be called inside instance definition");
        return;
    }
    if (renderOptions->instances.find(name) == renderOptions->instances.end()) {
        Error("Unable to find instance named \"%s\"", name.c_str());
        return;
    }
    vector<Reference<Primitive> > &in = renderOptions->instances[name];
    if (in.size() == 0) return;
    if (in.size() > 1 || !in[0]->CanIntersect()) {
        // Refine instance _Primitive_s and create aggregate
        Reference<Primitive> accel =
             MakeAccelerator(renderOptions->AcceleratorName,
                             in, renderOptions->AcceleratorParams);
        if (!accel) accel = MakeAccelerator("bvh", in, ParamSet());
        if (!accel) Severe("Unable to create \"bvh\" accelerator");
        in.erase(in.begin(), in.end());
        in.push_back(accel);
    }
    Assert(MAX_TRANSFORMS == 2);
    Transform *world2instance[2];
    transformCache.Lookup(curTransform[0], NULL, &world2instance[0]);
    transformCache.Lookup(curTransform[1], NULL, &world2instance[1]);
    AnimatedTransform animatedWorldToInstance(world2instance[0],
        renderOptions->transformStartTime,
        world2instance[1], renderOptions->transformEndTime);
    Reference<Primitive> prim =
        new TransformedPrimitive(in[0], animatedWorldToInstance);
    renderOptions->primitives.push_back(prim);
}


void pbrtWorldEnd() {
    VERIFY_WORLD("WorldEnd");
    // Ensure there are no pushed graphics states
    while (pushedGraphicsStates.size()) {
        Warning("Missing end to pbrtAttributeBegin()");
        pushedGraphicsStates.pop_back();
        pushedTransforms.pop_back();
    }
    while (pushedTransforms.size()) {
        Warning("Missing end to pbrtTransformBegin()");
        pushedTransforms.pop_back();
    }
}

void pbrtSetSamplingParams( int numChains,
			    int chainLength, 
			    int maxAcceptedSamplesPerChains,
			    
			    float mutation_sigma_spatial,
			    float mutation_sigma_angular,
			    float thresh_residual,
			    
			    int numChainDuplicateChecks,
			    int replaceSamples,
			    
			    int smplTime,
			    int smplFull)
{
    //Set the sampling parameters
    //If renderer already initialized, only change subset of parameters
    if( curRenderer != NULL )
    {
	if( numChains >= 0 || maxAcceptedSamplesPerChains >= 0 || smplTime >= 0 || smplFull >= 0)
	    Warning("Change of selected parameters not supported.\n");
	
	//Change allowed/valid ones
	samplingParams.chainLength = chainLength >= 0.f ? 
				     chainLength : samplingParams.chainLength;
				     
	samplingParams.mutation_sigma_spatial = mutation_sigma_spatial >= 0.f ? 
						mutation_sigma_spatial : samplingParams.mutation_sigma_spatial;
						
	samplingParams.mutation_sigma_angular = mutation_sigma_angular >= 0.f ? 
						mutation_sigma_angular : samplingParams.mutation_sigma_angular;
						
	samplingParams.thresh_residual = thresh_residual >= 0.f ? 
					 thresh_residual : samplingParams.thresh_residual;
	
	samplingParams.numChainDuplicateChecks = numChainDuplicateChecks >= 0 ? 
					 numChainDuplicateChecks : samplingParams.numChainDuplicateChecks;
	
	samplingParams.replaceSamples = replaceSamples >= 0 ? 
					 (replaceSamples > 0) : samplingParams.replaceSamples;

	//Fully update if already initialized
	if( (curReconstructionType > 0) && curLightfield != NULL )
	{
	    AdaptiveLFSamplerRenderer* lfrenderer = (AdaptiveLFSamplerRenderer*)curRenderer;
	    lfrenderer->updateSamplingParams( samplingParams.chainLength, 
					      samplingParams.mutation_sigma_spatial, 
					      samplingParams.mutation_sigma_angular, 
					      samplingParams.thresh_residual,
					      samplingParams.numChainDuplicateChecks,
					      samplingParams.replaceSamples );
	}
	
	return;
    }
    
    //Else change all parameters
    samplingParams.numChains = numChains;
    samplingParams.chainLength = chainLength;
    samplingParams.maxAcceptedSamplesPerChains = maxAcceptedSamplesPerChains;
    samplingParams.mutation_sigma_spatial = mutation_sigma_spatial;
    samplingParams.mutation_sigma_angular = mutation_sigma_angular;
    samplingParams.thresh_residual = thresh_residual;
    samplingParams.numChainDuplicateChecks = numChainDuplicateChecks;
    samplingParams.replaceSamples = (replaceSamples > 0);
    samplingParams.sampleTime = (smplTime > 0);
    samplingParams.sampleFull = (smplFull > 0);
    
    return;
}

void pbrtSetLayerRecParams(  float OX, float OY, float OZ,
			     float lightfieldSzX, float lightfieldSzY,
			     
			    int layerResX, int layerResY, int layerResZ,
			    float layerOX, float layerOY, float layerOZ,
			    float layerSzX, float layerSzY, float layerDst,
			     
			    float scaleRec,
			     
			    int timeStps, std::vector<int> *lyrSubSet,
			    
			    int basisFunction)
{
    //Do online change
    if( curLightfieldRec != NULL )
    {

	if( curReconstructionType == 2 )
	{
	    LightfieldReconstructionLayerAtt* lfrec = (LightfieldReconstructionLayerAtt*)curLightfieldRec;
	    lfrec->updateLayerParams(OX, OY, OZ,	  
				     lightfieldSzX, lightfieldSzY,
				    layerResX, layerResY, layerResZ,
				    layerOX, layerOY, layerOZ,
				    layerSzX, layerSzY, layerDst,
				    scaleRec,
				    basisFunction);
	}
	else if(curReconstructionType == 3 )
	{
	    LightfieldReconstructionTimeLayerAtt* lfrec = (LightfieldReconstructionTimeLayerAtt*)curLightfieldRec;
	    lfrec->updateLayerParams(OX, OY, OZ,	  
				     lightfieldSzX, lightfieldSzY,
				    layerResX, layerResY, layerResZ,
				    layerOX, layerOY, layerOZ,
				    layerSzX, layerSzY, layerDst,
				    scaleRec,
				    basisFunction,
				    timeStps, lyrSubSet);
	}
	else
	{
	    Warning("Layer parameters can not be changed for this reconstruction type.\n");
	}
    }
    
    //Else change all parameters
    //Lighfield part
    layerRecParams->OX = OX;
    layerRecParams->OY = OY;
    layerRecParams->OZ = OZ;
    
    layerRecParams->lightfieldSzX = lightfieldSzX;
    layerRecParams->lightfieldSzY = lightfieldSzY;
    
    //Layer part
    layerRecParams->layerResX = layerResX;
    layerRecParams->layerResY = layerResY;
    layerRecParams->layerResZ = layerResZ;
    
    layerRecParams->layerOX = layerOX;
    layerRecParams->layerOY = layerOY;
    layerRecParams->layerOZ = layerOZ;
    
    layerRecParams->layerSzX = layerSzX;
    layerRecParams->layerSzY = layerSzY;
    
    layerRecParams->layerDst = layerDst;
    
    layerRecParams->timeStps= timeStps;
    
    layerRecParams->scaleRec = scaleRec;
    
    layerRecParams->layerSubSet.clear();
    if( lyrSubSet != NULL )
	for( int i = 0; i < lyrSubSet->size(); i++)
	    layerRecParams->layerSubSet.push_back( lyrSubSet->at(i) );
    
    layerRecParams->basisFunction = basisFunction;
    
    return;
}

void pbrtResetChains()
{
    //Reset all chains
    if( (curReconstructionType > 0) && curLightfield != NULL )
    {
	AdaptiveLFSamplerRenderer* lfrenderer = (AdaptiveLFSamplerRenderer*)curRenderer;
	lfrenderer->resetChains();
    }
    
    return;
}

void pbrtGetView(float* view_rgb, int ax, int ay, int dx, int dy)
{
    //Do online change
    if( curLightfieldRec != NULL )
    {

	if( curReconstructionType == 2 )
	{
	    LightfieldReconstructionLayerAtt* lfrec = (LightfieldReconstructionLayerAtt*)curLightfieldRec;
	    lfrec->GetView(view_rgb, ax,  ay,  dx, dy);
	}
	else if( curReconstructionType == 3 )
	{
	    LightfieldReconstructionTimeLayerAtt* lfrec = (LightfieldReconstructionTimeLayerAtt*)curLightfieldRec;
	    lfrec->GetView(view_rgb, ax,  ay,  dx, dy);
	}
	else	
	{
	    Warning("View can not be retrieved for this reconstruction type.\n");
	} 
    }
    
    return;
}

void pbrtSetupRenderer(int reconstructionType, int duplicateCheck, bool sceneOV) {
    VERIFY_WORLD("WorldEnd");
    
    //Set global curReconstructionType
    curReconstructionType = sceneOV ? 0 : reconstructionType; //No reconstruction for scene OV
    
    //Set global curDuplicateCheck
    curDuplicateCheck = duplicateCheck; 
  
    // Create scene and render
    curCamera = renderOptions->MakeCamera();
    curLightfield = renderOptions->MakeLightfield(curCamera, NULL, sceneOV);
    curScene = renderOptions->MakeScene();
    
    //Set lightfield scene
    if( curLightfield != NULL)
	curLightfield->setScene(curScene);
    
    //Setup recontruction
    if( curLightfield != NULL && (curReconstructionType > 0) )
    {
	curLightfieldRec = renderOptions->MakeLightfieldRec( (LightfieldFilm*) curCamera->film );
    }
    
    //Finally make the full render
    curRenderer = renderOptions->MakeRenderer(curCamera, curLightfield);
      
}

void pbrtUpdateFilmSize(int resX, int resY,
			      float angleMinX, float angleMaxX, int resAngleX,
			      float angleMinY, float angleMaxY, int resAngleY)
{
    //Sanity checked        
    if (curLightfield == NULL)
    {
	Severe("Error: Update of filmsize currently only supported for lightfield films.");
	return;
    }
    
    //Change all the dims in film
    LightfieldFilm *film = (LightfieldFilm*)(curCamera->film); 
    film->updateSize(resX, resY,
		    angleMinX, angleMaxX, resAngleX,
		    angleMinY, angleMaxY, resAngleY);
    
    //Change renderer sample dims
    if(curSampler != NULL)
    {
	curSampler->updateSize( resX, resY,
				resAngleX, resAngleY);
    }
    
    //Change all the dims in reconstruction
    if( curLightfieldRec != NULL )
    {
	curLightfieldRec->updateSize(resX, resY,
				     resAngleX, resAngleY);
    }
    
    //Update renderer (only necessary for adaptive lightfield sampler)
    if( (curReconstructionType > 0) )
    {
	AdaptiveLFSamplerRenderer* curAdaptiveLFRenderer = (AdaptiveLFSamplerRenderer*)curRenderer;
	curAdaptiveLFRenderer->updateSize();
    }
  
    //Debug: Change film filename
    film->filename = "changed_res.exr";
    
    return;
}

void pbrtSetReconstruction( float* rec_in)
{
    //Check for lightfieldrec sanity
    if( !(curReconstructionType > 0) || curLightfieldRec == NULL )
	return;
    
    //Set based on reconstruction type
    if( curReconstructionType == 1 || curReconstructionType >= 4)
    {
	//(Default) Set Lightfield Reconstruction Data
	LightfieldReconstructionData * lrec = (LightfieldReconstructionData*)curLightfieldRec;
	lrec->SetData(rec_in);
	
	//Write reconstruction (debug)
	/*
	for( int ax = 0; ax < lrec->resAngleX; ax++ )
	      for( int ay = 0; ay < lrec->resAngleY; ay++ )
		  lrec->WriteImage(ax, ay, "rec.exr");
	*/
	
	//Write reconstruction single slice (debug)
	//lrec->WriteImageScanLine(0, 0, "rec_test.exr");
    }
    else if( curReconstructionType == 2 )
    {
	//Set Lightfield Reconstruction Layered Attenuator
	LightfieldReconstructionLayerAtt * lrec = (LightfieldReconstructionLayerAtt*)curLightfieldRec;
	lrec->SetData(rec_in);
	
	//Debug
	/*
	for(int z = 0; z < layerRecParams->layerResZ; z++)
	  lrec->WriteImage(z, "test.exr");
	*/
    }
    else if( curReconstructionType == 3 )
    {
	//Set Lightfield Reconstruction Time resolved Layered Attenuator
	LightfieldReconstructionTimeLayerAtt * lrec = (LightfieldReconstructionTimeLayerAtt*)curLightfieldRec;
	lrec->SetData(rec_in);
	
	//Debug
	/*
	for(int t = 0; t < layerRecParams->timeStps; t++)
	{
	    lrec->WriteImage(0, t, "front.exr");
	    lrec->WriteImage(1, t, "back.exr");
	}
	*/
    }
    
    return;
}

void pbrtSet1DTestData(float* rgb)
{
    //Sanity
    if( curLightfield == NULL || !(curReconstructionType > 0))
    {
	Warning("Cannot set 1D test data.\n");
	return;
    }
    
    //Set new 1d test data
    AdaptiveLFSamplerRenderer* sampleRenderer = (AdaptiveLFSamplerRenderer*)curRenderer;
    sampleRenderer->set1DtestData(rgb);
}


void pbrtGetFilmSize(int* resX, int* resY,
		float* angleMinX, float* angleMaxX, int* resAngleX,
		float* angleMinY, float* angleMaxY, int* resAngleY)
{
    VERIFY_WORLD("WorldEnd");
    
    //Sanity checks
    if(curCamera == NULL)
    {
	Severe("Camera is not setup.");
	return;
    } 
    
    //Save values
    if(curLightfield != NULL)
    {
	LightfieldFilm *film = (LightfieldFilm*)(curCamera->film);
	*resX = film->xResolution;
	*resY = film->yResolution;
	*angleMinX = film->angleMinX;
	*angleMaxX = film->angleMaxX;
	*resAngleX = film->resAngleX;
	*angleMinY = film->angleMinY;
	*angleMaxY = film->angleMaxY;
	*resAngleY = film->resAngleY;
    }else 
    {
	Film *film = curCamera->film;
	*resX = film->xResolution;
	*resY = film->yResolution;
	*angleMinX = 0.f;
	*angleMaxX = 0.f;
	*resAngleX = 0;
	*angleMinY = 0.f;
	*angleMaxY = 0.f;
	*resAngleY = 0;
    }
    
    return;
}

void pbrtGetTimeRes(int* resT)
{
    VERIFY_WORLD("WorldEnd");
    
    //Sanity checks
    if(curCamera == NULL)
    {
	Severe("Camera is not setup.");
	return;
    } 
    
    //Save values
    if(curLightfield != NULL)
    {
	LightfieldFilm *film = (LightfieldFilm*)(curCamera->film);
	*resT = film->timeRes;
    }
    else 
    {
	*resT = 0;
    }
}

void pbrtGetLightfieldSize(float* sizeX, float* sizeY)
{
      //Save values
    if(curLightfield != NULL)
    {
	*sizeX = curLightfield->sizeX;
	*sizeY = curLightfield->sizeY;
    }
    else
    {
	*sizeX = -1.f;
	*sizeY = -1.f;
    }
  
    return;
}

void pbrtRenderWorld(float time) {
    VERIFY_WORLD("WorldEnd");
    
    //Sanity checks
    if(curCamera == NULL || 
      curRenderer == NULL ||
      curScene == NULL)
    {
	Severe("Renderer, camera or scene is not setup.");
	return;
    }
  
    // Create scene and render
    curRenderer->Render(curScene, time);
    TasksCleanup();
    
    return;
}

void pbrtGetFilmImage(float* rgb)
{
    //Get image
    if( rgb != NULL )
	Warning("Could not get film image.\n");

    //Get image
    curCamera->film->GetPixels(rgb, 1.0);
    
    return;
}

void pbrtGetSamples(int* numRecordedSamples, float* rgb, int* sample_desc, bool getHistory)
{
    //Sanity
    if( curLightfield == NULL || !(curReconstructionType > 0))
    {
	Warning("Cannot get image samples.\n");
	return;
    }
    
    //Early exit
    if( numRecordedSamples == NULL)
	return;
    
    //Get structures
    AdaptiveLFSamplerRenderer* sampleRenderer = (AdaptiveLFSamplerRenderer*)curRenderer;
    SampleSet *samples = sampleRenderer->sampleSet; 
    int newSampleBegin = sampleRenderer->newSampleBegin;
    
    //Get history if selected
    if( getHistory )
      newSampleBegin = 0;
    
    //Compute last(!) recorded samples
    *numRecordedSamples = samples->pts->size() - newSampleBegin; 
    
    //Exit with just numRecorded samples
    if( rgb == NULL || sample_desc == NULL)
	return;
    
    //Iterate over the last gotten samples
    int offset = 0;
    for( int j = newSampleBegin; j < samples->pts->size(); j++)
    {
	//Add the current sample
	LFSamplePixel* s = &( samples->pts->at(j) );
	
	//RGB 
	rgb[offset*3 + 0] = s->rgb[0];
	rgb[offset*3 + 1] = s->rgb[1];
	rgb[offset*3 + 2] = s->rgb[2];
	
	//Sample description
	sample_desc[offset*4 + 0] = s->x;
	sample_desc[offset*4 + 1] = s->y;
	sample_desc[offset*4 + 2] = s->ax;
	sample_desc[offset*4 + 3] = s->ay;
	
	//Increment offset
	offset++;
    }
    
    //Sanity
    if( offset != *numRecordedSamples )
	Warning("Number of recorded samples and actual samples are not equal.\n");
    
    return;
}


void pbrtGetSamplingTimings(std::vector<float>* timings)
{
    //Sanity
    if( curLightfield == NULL || !(curReconstructionType > 0))
    {
	Warning("Cannot get sampling timings.\n");
	return;
    }
    
    //Early exit
    if( timings == NULL)
	return;
    
    //Get timings
    AdaptiveLFSamplerRenderer* sampleRenderer = (AdaptiveLFSamplerRenderer*)curRenderer;
    timings->clear();
    for( int s = 0; s < sampleRenderer->samplingTimings.size(); s++ )
	timings->push_back( sampleRenderer->samplingTimings.at(s) ); 
    
    return;
}

void pbrtWriteFilm()
{
    VERIFY_WORLD("WorldEnd");
    
    //Sanity
    if( curLightfield != NULL )
    {
	//Create new film if sampling
	LightfieldFilm* lfilm = NULL;
	if( (curReconstructionType > 0) )
	{
	    //Create new film
	    LightfieldFilm* emptyLFilm = (LightfieldFilm*)curLightfield->film;
	    lfilm = new LightfieldFilm(emptyLFilm->xResolution, emptyLFilm->yResolution,
					    emptyLFilm->angleMinX, emptyLFilm->angleMaxX, 
					    emptyLFilm->resAngleX,
					    emptyLFilm->angleMinY, emptyLFilm->angleMaxY,
					    emptyLFilm->resAngleY,
					    emptyLFilm->rayOffset,
					    emptyLFilm->timeRes,
					    emptyLFilm->filename, true );
	    
	    //Fill film with samples
	    AdaptiveLFSamplerRenderer* sampleRenderer = (AdaptiveLFSamplerRenderer*)curRenderer;
	    std::vector<LFSamplingChain> *chains = sampleRenderer->chains; 
	    int totalRecordedSamples = sampleRenderer->totalRecordedSamples;
	    int totalTestedSamples = sampleRenderer->totalTestedSamples;
	    
	    //Debug print:
	    printf("Number of recorded/tested samples [%d / %d]\n", totalRecordedSamples, totalTestedSamples );
	    
	    //Iterate over samples of all chains
	    for( int cIdx = 0; cIdx < chains->size(); cIdx++ )
	    {
		//Get current chain
		LFSamplingChain* currChain = &( chains->at(cIdx) );
		for( int i = 0; i < currChain->recorded->size(); i++ )
		    lfilm->AddSample(currChain->recorded->at(i));
	    }

	}else
	    lfilm = (LightfieldFilm*)curLightfield->film; //Just get film
	
	//Exit if film emtpy
	if( lfilm == NULL )
	    return;
	
	//Write out the full lightfield (or only slice if centerScanLine)
	bool fullLF = !renderOptions->CameraParams.FindOneBool("centerScanline", false);
	if( fullLF )
	{
	    for( int ax = 0; ax < lfilm->resAngleX; ax++ )
		for( int ay = 0; ay < lfilm->resAngleY; ay++ )
		      lfilm->WriteImage(1.f,ax, ay);
	}else
	{
	    //Debug slice output
	    lfilm->WriteImageScanLine(1.f, 0, 0);
	}
    }
    else
	curCamera->film->WriteImage(1.f);
}

void pbrtDestroyRenderer()
{
     // Clean up after rendering
    if(curRenderer != NULL)
    {
	delete curRenderer;
	curRenderer = NULL;
    }
    //Handled by renderer
    curLightfield = NULL;
    
    if(curLightfieldRec != NULL)
    {
	delete curLightfieldRec;
	curLightfieldRec = NULL;
    }
    
    if(curScene != NULL)
    {
	delete curScene;
	curScene = NULL;
    }
    
    //Cleanup API state
    graphicsState = GraphicsState();
    transformCache.Clear();
    for (int i = 0; i < MAX_TRANSFORMS; ++i)
        curTransform[i] = Transform();
    activeTransformBits = ALL_TRANSFORMS_BITS;
    namedCoordinateSystems.erase(namedCoordinateSystems.begin(),
                                 namedCoordinateSystems.end());
    
    currentApiState = STATE_OPTIONS_BLOCK;
    ProbesPrint(stdout);
    ImageTexture<float, float>::ClearCache();
    ImageTexture<RGBSpectrum, Spectrum>::ClearCache();
}


Scene *RenderOptions::MakeScene() {
    // Initialize _volumeRegion_ from volume region(s)
    VolumeRegion *volumeRegion;
    if (volumeRegions.size() == 0)
        volumeRegion = NULL;
    else if (volumeRegions.size() == 1)
        volumeRegion = volumeRegions[0];
    else
        volumeRegion = new AggregateVolume(volumeRegions);
    Primitive *accelerator = MakeAccelerator(AcceleratorName,
        primitives, AcceleratorParams);
    if (!accelerator)
        accelerator = MakeAccelerator("bvh", primitives, ParamSet());
    if (!accelerator)
        Severe("Unable to create \"bvh\" accelerator.");
    
    // Warn if no light sources are defined
    if (lights.size() == 0)
	Warning("No light sources defined in scene; "
	"	possibly rendering a black image.");
	    
    Scene *scene = new Scene(accelerator, lights, volumeRegion);
    // Erase primitives, lights, and volume regions from _RenderOptions_
    primitives.erase(primitives.begin(), primitives.end());
    lights.erase(lights.begin(), lights.end());
    volumeRegions.erase(volumeRegions.begin(), volumeRegions.end());
    return scene;
}


Renderer *RenderOptions::MakeRenderer(Camera *camera, Lightfield *lightfield) const {
  
    //Set camera to lightfield if possible
    if( lightfield != NULL )
    {
	//Check for lightfield reconstruction rendering
	if ( (curReconstructionType > 0) && RendererName != "sampler")
	    Severe("Lightfield sampling only supported by sampler renderer.\n");
	else
	    camera = lightfield;
    }
    
    //Generate renderer
    Renderer *renderer = NULL;
    if (RendererName == "metropolis") {
        renderer = CreateMetropolisRenderer(RendererParams, camera);
        RendererParams.ReportUnused();
    }
    // Create remaining _Renderer_ types
    else if (RendererName == "createprobes") {
        // Create surface and volume integrators
        SurfaceIntegrator *surfaceIntegrator = MakeSurfaceIntegrator(SurfIntegratorName,
            SurfIntegratorParams);
        if (!surfaceIntegrator) Severe("Unable to create surface integrator.");
        VolumeIntegrator *volumeIntegrator = MakeVolumeIntegrator(VolIntegratorName,
            VolIntegratorParams);
        if (!volumeIntegrator) Severe("Unable to create volume integrator.");
        renderer = CreateRadianceProbesRenderer(camera, surfaceIntegrator, volumeIntegrator, RendererParams);
        RendererParams.ReportUnused();
    }
    else if (RendererName == "aggregatetest") {
        renderer = CreateAggregateTestRenderer(RendererParams, primitives);
        RendererParams.ReportUnused();
    }
    else if (RendererName == "surfacepoints") {
        Point pCamera = camera->CameraToWorld(camera->shutterOpen, Point(0, 0, 0));
        renderer = CreateSurfacePointsRenderer(RendererParams, pCamera, camera->shutterOpen);
        RendererParams.ReportUnused();
    }
    else {
        if (RendererName != "sampler")
            Warning("Renderer type \"%s\" unknown.  Using \"sampler\".",
                    RendererName.c_str());
        bool visIds = RendererParams.FindOneBool("visualizeobjectids", false);
        RendererParams.ReportUnused();
        curSampler = MakeSampler(SamplerName, SamplerParams, camera->film, camera, lightfield);
        if (!curSampler) 
	  Severe("Unable to create sampler.");
        // Create surface and volume integrators
        SurfaceIntegrator *surfaceIntegrator = MakeSurfaceIntegrator(SurfIntegratorName,
            SurfIntegratorParams);
        if (!surfaceIntegrator) Severe("Unable to create surface integrator.");
        VolumeIntegrator *volumeIntegrator = MakeVolumeIntegrator(VolIntegratorName,
            VolIntegratorParams);
        if (!volumeIntegrator) Severe("Unable to create volume integrator.");
	
	//Check for normal or sampled version
	if(!(curReconstructionType > 0))
	{
	    renderer = new SamplerRenderer(curSampler, camera, surfaceIntegrator,
					  volumeIntegrator, visIds);
	}
	else
	{   
	    renderer = new AdaptiveLFSamplerRenderer( curSampler, camera, surfaceIntegrator,
					  volumeIntegrator, visIds,
					  curLightfieldRec,
					  curDuplicateCheck, 
					  
					  samplingParams.numChains, 
					  samplingParams.chainLength, 
					  samplingParams.maxAcceptedSamplesPerChains,
					  samplingParams.mutation_sigma_spatial,
					  samplingParams.mutation_sigma_angular,
					  samplingParams.thresh_residual,
					  samplingParams.numChainDuplicateChecks,
					  samplingParams.replaceSamples,
					  samplingParams.sampleTime,
					  samplingParams.sampleFull);
	    
	    //Debug 
	    printf("Used sampling params: %d, %d, %d | %3.3f, %3.3f, %3.3f | %d | %d %d \n",samplingParams.numChains, 
									    samplingParams.chainLength, 
									    samplingParams.maxAcceptedSamplesPerChains,
									    samplingParams.mutation_sigma_spatial,
									    samplingParams.mutation_sigma_angular,
									    samplingParams.thresh_residual,
									    samplingParams.numChainDuplicateChecks,
									    samplingParams.replaceSamples,
									    samplingParams.sampleTime);
	}
	
    }
    return renderer;
}


Camera *RenderOptions::MakeCamera() const {
    Filter *filter = MakeFilter(FilterName, FilterParams);
    Film *film = MakeFilm(FilmName, FilmParams, filter);
    if (!film) Severe("Unable to create film.");
    Camera *camera = ::MakeCamera(CameraName, CameraParams,
        CameraToWorld, renderOptions->transformStartTime,
        renderOptions->transformEndTime, film);
    if (!camera) Severe("Unable to create camera.");
    return camera;
}

Lightfield *RenderOptions::MakeLightfield(Camera *camera, Scene *scene, bool sceneOV ) const
{
    Lightfield* lightfield = NULL;
    
    //Check for lightfield film
    if(FilmName != "lightfield")
    {
	//Warning("Unable to create lightfield. No lightfield film specified.");
	return lightfield;
    }
    
    // Extract common camera parameters from _ParamSet_
    float shutteropen = CameraParams.FindOneFloat("shutteropen", 0.0f);
    float shutterclose = CameraParams.FindOneFloat("shutterclose", 0.0f);
    if (shutterclose < shutteropen) {
        Warning("Shutter close time [%f] < shutter open [%f].  Swapping them.",
                shutterclose, shutteropen);
        swap(shutterclose, shutteropen);
    }
    
    //Compute bounds
    
    //Define a camera sample to probe bounds
    CameraSample currCamSample;
    currCamSample.imageX = 0.f;
    currCamSample.imageY = 0.f;
    currCamSample.lensU = 0.5f;
    currCamSample.lensV = 0.5f;
    currCamSample.time = 0.f;
    
    //Render full or only scanline
    bool fullLF = !CameraParams.FindOneBool("centerScanline", false);
    
    //Probe from camera
    LightfieldFilm *film = (LightfieldFilm*)(camera->film);
    currCamSample.imageX = 0.f;
    currCamSample.imageY = fullLF ? 0.f : (0.5f - 0.001)*float(film->yResolution);
    Ray leftUpRay;
    camera->GenerateRay(currCamSample, &leftUpRay);
    
    currCamSample.imageX = 0.f;
    currCamSample.imageY = fullLF ? film->yResolution : (0.5f + 0.001)*float(film->yResolution);
    Ray leftDownRay;
    camera->GenerateRay(currCamSample, &leftDownRay);
    
    currCamSample.imageX = film->xResolution;
    currCamSample.imageY = fullLF ? 0.f : (0.5f - 0.001)*float(film->yResolution);
    Ray rightUpRay;
    camera->GenerateRay(currCamSample, &rightUpRay);
    
    currCamSample.imageX = film->xResolution;
    currCamSample.imageY = fullLF ? film->yResolution : (0.5f + 0.001)*float(film->yResolution);
    Ray rightDownRay;
    camera->GenerateRay(currCamSample, &rightDownRay);
    
    currCamSample.imageX = ((float)(film->xResolution)) / 2.f;
    currCamSample.imageY = ((float)(film->yResolution)) / 2.f;
    Ray centerRay;
    camera->GenerateRay(currCamSample, &centerRay);
    
    //Compute intersection with lookAtPlane
    float t_curr;
    Normal lookAtPlaneNormal = Normal(centerRay.d);
    lookAtPlaneNormal = Normalize(lookAtPlaneNormal);
    t_curr = Dot(Vector( curLookAtPoint - leftUpRay.o ), lookAtPlaneNormal)
	      /	Dot(leftUpRay.d, lookAtPlaneNormal); //Should be safe due to look
    Point leftUp = leftUpRay(t_curr);
    
    t_curr = Dot(Vector( curLookAtPoint - leftDownRay.o ), lookAtPlaneNormal)
	      /	Dot(leftDownRay.d, lookAtPlaneNormal); //Should be safe due to look
    Point leftDown = leftDownRay(t_curr);
    
    t_curr = Dot(Vector( curLookAtPoint - rightUpRay.o ), lookAtPlaneNormal)
	      /	Dot(rightUpRay.d, lookAtPlaneNormal); //Should be safe due to look
    Point rightUp = rightUpRay(t_curr);
    
    t_curr = Dot(Vector( curLookAtPoint - rightDownRay.o ), lookAtPlaneNormal)
	      /	Dot(rightDownRay.d, lookAtPlaneNormal); //Should be safe due to look
    Point rightDown = rightDownRay(t_curr);
    
    //Debug print
    printf("LeftUp [%3.3f %3.3f %3.3f]\n", leftUp.x, leftUp.y, leftUp.z);
    printf("LeftDown [%3.3f %3.3f %3.3f]\n", leftDown.x, leftDown.y, leftDown.z);
    printf("RightUp [%3.3f %3.3f %3.3f]\n", rightUp.x, rightUp.y, rightUp.z);
    printf("RightDown [%3.3f %3.3f %3.3f]\n", rightDown.x, rightDown.y, rightDown.z);
    
    //Finally compute visible x and y size
    float xSize = Distance(leftDown, rightDown);
    float ySize = Distance(leftDown, leftUp);
    
    /*
    printf("XSize [%3.3f]\n", xSize);
    printf("YSize [%3.3f]\n", ySize);
    */
    
    //Create regular lightfield if no overview, else create orthographic cam from top
    if( !sceneOV )
    {
	//Generate transform with plane normal 
	//*opposite to camera*
	Vector planeNormal = -Normalize(centerRay.d);
	Vector up = Cross( planeNormal, Vector(leftDown - rightDown) );
	
	Transform* l2w = new Transform();
	*l2w = Inverse( LookAt( rightUp,
				rightUp + planeNormal, 
				up ) );
	
	//No movement
	AnimatedTransform animatedCam2World(l2w, renderOptions->transformStartTime,
					    l2w, renderOptions->transformEndTime);
	
	//Generate lightfield
	lightfield = new Lightfield( animatedCam2World, scene, film,
				    xSize, ySize,
				    shutteropen, 
				    shutterclose );

	//Return
	if (!lightfield) 
	    Severe("Unable to create lightfield.");
    }else
    {
      
      	//*opposite to camera*
	Vector planeNormal = -Normalize(centerRay.d);
	Vector up = Cross( planeNormal, Vector(leftDown - rightDown) );
	
	Transform* l2w = new Transform();
	*l2w = Inverse( LookAt( rightUp,
				rightUp - up, 
				planeNormal ));
	
	//Compute large depth range
	float lensradius = 0.f;
	float focaldistance = 1e30f;
	
	//Compute screen coords
	float screen[4];
	screen[0] = 0.f;
	screen[1] = xSize;
	screen[2] = -ySize/2;
	screen[3] = ySize/2;
	
	//No movement
	AnimatedTransform animatedCam2World(l2w, renderOptions->transformStartTime,
					    l2w, renderOptions->transformEndTime);

	//Compute new ortho camera
	delete curCamera;
	
	//Create new 2d film
	Filter *filter = MakeFilter(FilterName, FilterParams);
	Film *film = CreateImageFilm(FilmParams, filter);
	if (!film) Severe("Unable to create new ortho film.");
		       
	curCamera = new OrthoCamera(animatedCam2World, screen, shutteropen, shutterclose,
				  lensradius, focaldistance, film);
    }
    
    return lightfield;
}

LightfieldReconstruction *RenderOptions::MakeLightfieldRec(LightfieldFilm* film ) const 
{
    LightfieldReconstruction* lightfieldRec = NULL;
    
    //Early exit
    if( !(curReconstructionType > 0) )
	return lightfieldRec;
    
    //Check for film
    if(film == NULL)
    {
	Warning("Unable to create lightfield reconstruction. Film is unset.");
	return lightfieldRec;
    }
    
    //Create empty Lightfield rec for specific type
    if( curReconstructionType == 1 || curReconstructionType >= 4 )
    {
	//Default is here full data
	lightfieldRec = new LightfieldReconstructionData(film->xResolution, film->yResolution,
							film->resAngleX, film->resAngleY, NULL);
    }
    else if( curReconstructionType == 2 )
    {
	//Layer reconstruciton
	lightfieldRec = new LightfieldReconstructionLayerAtt(
			  film->xResolution, film->yResolution,
			  film->angleMinX, film->angleMaxX, film->resAngleX, 
			  film->angleMinY, film->angleMaxY, film->resAngleY,
			  
			  layerRecParams->OX, layerRecParams->OY, layerRecParams->OZ,
			  layerRecParams->lightfieldSzX, layerRecParams->lightfieldSzY,
			  
			  layerRecParams->layerResX, layerRecParams->layerResY, layerRecParams->layerResZ,
			  layerRecParams->layerOX, layerRecParams->layerOY, layerRecParams->layerOZ,
			  layerRecParams->layerSzX, layerRecParams->layerSzY, 
			  layerRecParams->layerDst,
			  
			  layerRecParams->scaleRec,
			  
			  layerRecParams->basisFunction,
			
			  NULL);
    } 
    else if( curReconstructionType == 3 )
    {
	//Layer reconstruciton
	lightfieldRec = new LightfieldReconstructionTimeLayerAtt(
			  film->xResolution, film->yResolution,
			  film->angleMinX, film->angleMaxX, film->resAngleX, 
			  film->angleMinY, film->angleMaxY, film->resAngleY,
			  
			  layerRecParams->OX, layerRecParams->OY, layerRecParams->OZ,
			  layerRecParams->lightfieldSzX, layerRecParams->lightfieldSzY,
			  
			  layerRecParams->layerResX, layerRecParams->layerResY, layerRecParams->layerResZ,
			  layerRecParams->layerOX, layerRecParams->layerOY, layerRecParams->layerOZ,
			  layerRecParams->layerSzX, layerRecParams->layerSzY, 
			  layerRecParams->layerDst,
			  
			  layerRecParams->scaleRec,
			  
			  layerRecParams->basisFunction,
			
			  layerRecParams->timeStps, &(layerRecParams->layerSubSet));
    }
    
    //Return
    if (!lightfieldRec) 
	Severe("Unable to create lightfield reconstruction.");
	
    return lightfieldRec;
}


