
/*
    pbrt source code Copyright(c) 1998-2012 Matt Pharr and Greg Humphreys.

    This file is part of pbrt.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

    - Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    - Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
    IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
    TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */

#if defined(_MSC_VER)
#pragma once
#endif

#ifndef PBRT_CORE_API_H
#define PBRT_CORE_API_H

// core/api.h*
#include "pbrt.h"
#include "stdafx.h"
#include "parallel.h"
#include "paramset.h"
#include "spectrum.h"
#include "scene.h"
#include "renderer.h"
#include "film.h"
#include "volume.h"
#include "probes.h"
#include "geometry.h"
#include "cameras/lightfield.h"
#include "lightfield_reconstruction.h"

// Include map
#include <map>
using std::map;

//###################################
//######### API Local Classes
//###################################

//API states
#define STATE_UNINITIALIZED  0
#define STATE_OPTIONS_BLOCK  1
#define STATE_WORLD_BLOCK    2

//Transform states
#define MAX_TRANSFORMS 2
#define START_TRANSFORM_BITS (1 << 0)
#define END_TRANSFORM_BITS   (1 << 1)
#define ALL_TRANSFORMS_BITS  ((1 << MAX_TRANSFORMS) - 1)
struct TransformSet {
   // TransformSet Public Methods
   Transform &operator[](int i) {
       Assert(i >= 0 && i < MAX_TRANSFORMS);
       return t[i];
   }
   const Transform &operator[](int i) const { Assert(i >= 0 && i < MAX_TRANSFORMS); return t[i]; }
   friend TransformSet Inverse(const TransformSet &ts) {
       TransformSet t2;
       for (int i = 0; i < MAX_TRANSFORMS; ++i)
           t2.t[i] = Inverse(ts.t[i]);
       return t2;
   }
   bool IsAnimated() const {
       for (int i = 0; i < MAX_TRANSFORMS-1; ++i)
           if (t[i] != t[i+1]) return true;
       return false;
   }
private:
    Transform t[MAX_TRANSFORMS];
};


struct RenderOptions {
    // RenderOptions Public Methods
    RenderOptions();
    Scene *MakeScene();
    Camera *MakeCamera() const;
    Renderer *MakeRenderer(Camera *camera, Lightfield *lightfield) const;
    Lightfield *MakeLightfield(Camera *camera, Scene *scene, bool sceneOV)  const;
    LightfieldReconstruction *MakeLightfieldRec( LightfieldFilm* film ) const;

    // RenderOptions Public Data
    float transformStartTime, transformEndTime;
    string FilterName;
    ParamSet FilterParams;
    string FilmName;
    ParamSet FilmParams;
    string SamplerName;
    ParamSet SamplerParams;
    string AcceleratorName;
    ParamSet AcceleratorParams;
    string RendererName;
    string SurfIntegratorName, VolIntegratorName;
    ParamSet RendererParams;
    ParamSet SurfIntegratorParams, VolIntegratorParams;
    string CameraName;
    ParamSet CameraParams;
    TransformSet CameraToWorld;
    Point CameraLookAtPoint;
    vector<Light *> lights;
    vector<Reference<Primitive> > primitives;
    mutable vector<VolumeRegion *> volumeRegions;
    map<string, vector<Reference<Primitive> > > instances;
    vector<Reference<Primitive> > *currentInstance;
};

struct GraphicsState {
    // Graphics State Methods
    GraphicsState();
    Reference<Material> CreateMaterial(const ParamSet &params);

    // Graphics State
    map<string, Reference<Texture<float> > > floatTextures;
    map<string, Reference<Texture<Spectrum> > > spectrumTextures;
    ParamSet materialParams;
    string material;
    map<string, Reference<Material> > namedMaterials;
    string currentNamedMaterial;
    ParamSet areaLightParams;
    string areaLight;
    bool reverseOrientation;
};

class TransformCache {
public:
    // TransformCache Public Methods
    void Lookup(const Transform &t, Transform **tCached,
            Transform **tCachedInverse) {
        map<Transform, std::pair<Transform *, Transform *> >::iterator iter;
        iter = cache.find(t);
        if (iter == cache.end()) {
            Transform *tr = arena.Alloc<Transform>();
            *tr = t;
            Transform *tinv = arena.Alloc<Transform>();
            *tinv = Transform(Inverse(t));
            cache[t] = std::make_pair(tr, tinv);
            iter = cache.find(t);
            PBRT_ALLOCATED_CACHED_TRANSFORM();
        }
        else
            PBRT_FOUND_CACHED_TRANSFORM();
        if (tCached) *tCached = iter->second.first;
        if (tCachedInverse) *tCachedInverse = iter->second.second;
    }
    void Clear() {
        arena.FreeAll();
        cache.erase(cache.begin(), cache.end());
    }
private:
    // TransformCache Private Data
    map<Transform, std::pair<Transform *, Transform *> > cache;
    MemoryArena arena;
};

struct SamplingParams {
    // Sampling params methods
    SamplingParams();

    // Sampling parameters
    int numChains;
    int chainLength;
    int maxAcceptedSamplesPerChains;
    
    float mutation_sigma_spatial;
    float mutation_sigma_angular;
    float thresh_residual;
    
    int numChainDuplicateChecks;
    bool replaceSamples;
    
    bool sampleTime;
    bool sampleFull;
};

struct LayerRecParams {
    //LayerRec params methods
    LayerRecParams(); 
    
    //Layerrec params
    
    //Lightfield stuff
    float OX;
    float OY; 
    float OZ;
    
    float lightfieldSzX;
    float lightfieldSzY;
		
    //Layer stuff
    int layerResX; 
    int layerResY; 
    int layerResZ;
    
    float layerOX;
    float layerOY;
    float layerOZ;
    
    float layerSzX;
    float layerSzY;
    
    float layerDst;
    
    float scaleRec;
    
    //Time and subsets
    int timeStps;
    vector<int> layerSubSet;
    
    int basisFunction;
};

class APIState {
public:
    // APIState Methods
    APIState() { 
	  //Initialize according to api global inits
	  currentApiState = STATE_UNINITIALIZED;
	  lookAtDone = false;
	  curReconstructionType = 0;
	  curDuplicateCheck = 0;
	  activeTransformBits = ALL_TRANSFORMS_BITS;
	  renderOptions = NULL;
	  
	  layerRecParams = NULL;

	  curScene = NULL;
	  curCamera = NULL;
	  curRenderer = NULL;
	  curSampler = NULL;
	  curLightfield = NULL;
	  curLightfieldRec = NULL;
    }
    
    //API state variables
    Options PbrtOptions;

    // API Static Data
    int currentApiState;
    TransformSet curTransform;
    Point curLookAtPoint;
    bool lookAtDone;
    int activeTransformBits;
    map<string, TransformSet> namedCoordinateSystems;
    RenderOptions *renderOptions;
    GraphicsState graphicsState;
    vector<GraphicsState> pushedGraphicsStates;
    vector<TransformSet> pushedTransforms;
    vector<uint32_t> pushedActiveTransformBits;
    TransformCache transformCache;

    Scene *curScene;
    Camera *curCamera;
    Renderer *curRenderer;
    Sampler* curSampler;
    Lightfield *curLightfield;
    int curReconstructionType;
    int curDuplicateCheck;
    LightfieldReconstruction *curLightfieldRec;
    SamplingParams samplingParams;
    LayerRecParams* layerRecParams;
    
};

//###################################
//######### API Function Declarations
//###################################

void pbrtInit(const Options &opt);
void pbrtCleanup();
void pbrtIdentity();
void pbrtTranslate(float dx, float dy, float dz);
void pbrtRotate(float angle, float ax, float ay, float az);
void pbrtScale(float sx, float sy, float sz);
void pbrtLookAt(float ex, float ey, float ez,
                float lx, float ly, float lz,
                float ux, float uy, float uz);
void pbrtConcatTransform(float transform[16]);
void pbrtTransform(float transform[16]);
void pbrtCoordinateSystem(const string &);
void pbrtCoordSysTransform(const string &);
void pbrtActiveTransformAll();
void pbrtActiveTransformEndTime();
void pbrtActiveTransformStartTime();
void pbrtTransformTimes(float start, float end);
void pbrtPixelFilter(const string &name, const ParamSet &params);
void pbrtFilm(const string &type, const ParamSet &params);
void pbrtSampler(const string &name, const ParamSet &params);
void pbrtAccelerator(const string &name, const ParamSet &params);
void pbrtSurfaceIntegrator(const string &name, const ParamSet &params);
void pbrtVolumeIntegrator(const string &name, const ParamSet &params);
void pbrtRenderer(const string &name, const ParamSet &params);
void pbrtCamera(const string &, const ParamSet &cameraParams);
void pbrtWorldBegin();
void pbrtAttributeBegin();
void pbrtAttributeEnd();
void pbrtTransformBegin();
void pbrtTransformEnd();
void pbrtTexture(const string &name, const string &type,
    const string &texname, const ParamSet &params);
void pbrtMaterial(const string &name, const ParamSet &params);
void pbrtMakeNamedMaterial(const string &name, const ParamSet &params);
void pbrtNamedMaterial(const string &name);
void pbrtLightSource(const string &name, const ParamSet &params);
void pbrtAreaLightSource(const string &name, const ParamSet &params);
void pbrtShape(const string &name, const ParamSet &params);
void pbrtReverseOrientation();
void pbrtVolume(const string &name, const ParamSet &params);
void pbrtObjectBegin(const string &name);
void pbrtObjectEnd();
void pbrtObjectInstance(const string &name);
void pbrtWorldEnd();

//Additional state functionality
void pbrtSaveAPIState(APIState &st);
void pbrtRestoreAPIState(const APIState &st);
//void pbrtDestroyState();

//Additional functionality besides pbrt core
void pbrtSetupRenderer(int reconstructionType, int duplicateCheck, bool sceneOV);
void pbrtDestroyRenderer();
void pbrtGetFilmSize(int* resX, int* resY,
		float* angleMinX, float* angleMaxX, int* resAngleX,
		float* angleMinY, float* angleMaxY, int* resAngleY);
void pbrtGetLightfieldSize(float* sizeX, float* sizeY);
void pbrtGetTimeRes(int* resT);

void pbrtRenderWorld(float time = 0.f);

void pbrtSetSamplingParams( int numChains,
			    int chainLength, 
			    int maxAcceptedSamplesPerChains,
			    
			    float mutation_sigma_spatial,
			    float mutation_sigma_angular,
			    float thresh_residual,
			    
			    int numChainDuplicateChecks,
			    int replaceSamples,
			    
			    int smplTime,
			    int smplFull);

void pbrtSetLayerRecParams(  float OX, float OY, float OZ,
			     float lightfieldSzX, float lightfieldSzY,
			     
			    int layerResX, int layerResY, int layerResZ,
			    float layerOX, float layerOY, float layerOZ,
			    float layerSzX, float layerSzY, 
			    float layerDst,
			     
			    float scaleRec,
			     
			    int timeStps, std::vector<int> *lyrSubSet,
			    
			    int basisFunction);

void pbrtSet1DTestData(float* rgb);

void pbrtResetChains();

void pbrtGetView(float* view_rgb, int ax, int ay, int dx, int dy);

void pbrtGetSamples(int* numRecordedSamples, float* rgb, int* sample_desc, bool getHistory = false);
void pbrtGetSamplingTimings(std::vector<float>* timings);

void pbrtSetReconstruction( float* rec_in );
void pbrtUpdateFilmSize(int resX, int resY,
			    float angleMinX, float angleMaxX, int resAngleX,
			    float angleMinY, float angleMaxY, int resAngleY);

void pbrtWriteFilm();
void pbrtGetFilmImage(float* rgb);

#endif // PBRT_CORE_API_H
