
/*
    Lightfield film contains lightfield angular and spatial resolutions.
 */


// film/image.cpp*
#include "stdafx.h"
#include "film/lightfield_film.h"
#include "spectrum.h"
#include "parallel.h"
#include "imageio.h"

#include <sstream>
#include <string>
#include <iomanip>

// LightfieldFilm Method Definitions
LightfieldFilm::LightfieldFilm(int xres, int yres,
		  float aMinX, float aMaxX, int rAX,
		  float aMinY, float aMaxY, int rAY,
		  float rOff,
		  int tRes,
                  const string &fn, bool allocData)
    : Film(xres, yres),
      angleMinX(aMinX), angleMaxX(aMaxX), resAngleX(rAX),
      angleMinY(aMinY), angleMaxY(aMaxY), resAngleY(rAY),
      rayOffset(rOff),
      timeRes(tRes)
{
    // Size
    filename = fn;

    // Allocate film image storage
    pixels = NULL;
    if( allocData )
    {
	pixels = new std::vector<Pixel>(xres * yres * rAX * rAY); //We dont need  to write the time images...
    }
}

LightfieldFilm::~LightfieldFilm() 
{
    if( pixels != NULL )
    {
	delete pixels;
    } 
}

void LightfieldFilm::updateSize(int xres, int yres,
		  float aMinX, float aMaxX, int rAX,
		  float aMinY, float aMaxY, int rAY)
{
    // Change size
    xResolution = xres;
    yResolution = yres;
    
    angleMinX = aMinX;
    angleMaxX = aMaxX;
    resAngleX = rAX;
    
    angleMinY = aMinY;
    angleMaxY = aMaxY;
    resAngleY = rAY;

    // Allocate new film image storage if neccessary
    if( pixels != NULL )
    {
	delete pixels;
	pixels = new std::vector<Pixel>(xres * yres * rAX * rAY);
    }
}

//Helper function from lightfield to pixel indices (for film)
int LightfieldFilm::l2px(int x, int y, int ax, int ay)
{
    int res_idx = y * (xResolution * resAngleY * resAngleX) + 
		  x  * (resAngleY * resAngleX) + 
		  ay * (resAngleX) + 
		  ax;
		  
    return res_idx;
}

void LightfieldFilm::AddSample(const CameraSample &s, const Spectrum &L) 
{
    if (pixels == NULL) {
        Warning("Trying to manipulate data of Lightfieldfilm without data.");
        return;
    }
    
    if (L.HasNaNs()) {
        Warning("ImageFilm ignoring splatted spectrum with NaN values");
        return;
    }
    
    // Compute sample's raster position
    int x = Floor2Int(s.imageX);
    int y = Floor2Int(s.imageY);
    int ax = Floor2Int(s.lensU);
    int ay = Floor2Int(s.lensV);

    // Loop over filter support and add sample to pixel arrays
    float xyz[3];
    L.ToXYZ(xyz);
    
    //Range sanity
    if (x  < 0 || x  >= xResolution ||
        y  < 0 || y  >= yResolution ||
        ax < 0 || ax >= resAngleX ||
        ay < 0 || ay >= resAngleY )
	return;
    
    Pixel* pixel = &( (*pixels).at( l2px(x, y, ax, ay) ) );
    AtomicAdd(&(pixel->Lxyz[0]), xyz[0]);
    AtomicAdd(&(pixel->Lxyz[1]), xyz[1]);
    AtomicAdd(&(pixel->Lxyz[2]), xyz[2]);
    AtomicAdd(&(pixel->weightSum), 1.0);
}

void LightfieldFilm::AddSample(const LFSamplePixel &s) 
{
    if (pixels == NULL) {
        Warning("Trying to manipulate data of Lightfieldfilm without data.");
        return;
    }
    
    // Compute sample's raster position
    int x = s.x;
    int y = s.y;
    int ax = s.ax;
    int ay = s.ay;

    // Loop over filter support and add sample to pixel arrays
    float xyz[3];
    RGBToXYZ(s.rgb, xyz);
    
    //Range sanity
    if (x  < 0 || x  >= xResolution ||
        y  < 0 || y  >= yResolution ||
        ax < 0 || ax >= resAngleX ||
        ay < 0 || ay >= resAngleY )
	return;
    
    Pixel* pixel = &( (*pixels).at( l2px(x, y, ax, ay) ) );
    AtomicAdd(&(pixel->Lxyz[0]), xyz[0]);
    AtomicAdd(&(pixel->Lxyz[1]), xyz[1]);
    AtomicAdd(&(pixel->Lxyz[2]), xyz[2]);
    AtomicAdd(&(pixel->weightSum), 1.0);
}

void LightfieldFilm::Splat(const CameraSample &s, const Spectrum &L) 
{
    if (pixels == NULL) {
        Warning("Trying to manipulate data of Lightfieldfilm without data.");
        return;
    }
    
    if (L.HasNaNs()) {
        Warning("ImageFilm ignoring splatted spectrum with NaN values");
        return;
    }
    
    // Compute sample's raster position
    int x = Floor2Int(s.imageX);
    int y = Floor2Int(s.imageY);
    int ax = Floor2Int(s.lensU);
    int ay = Floor2Int(s.lensV);
    
    //Splat sample
    float xyz[3];
    L.ToXYZ(xyz);
    
    //Range sanity
    if (x  < 0 || x  >= xResolution ||
        y  < 0 || y  >= yResolution ||
        ax < 0 || ax >= resAngleX ||
        ay < 0 || ay >= resAngleY )
	return;
    
    Pixel* pixel = &( (*pixels).at( l2px(x, y, ax, ay) ) );
    AtomicAdd(&(pixel->splatXYZ[0]), xyz[0]);
    AtomicAdd(&(pixel->splatXYZ[1]), xyz[1]);
    AtomicAdd(&(pixel->splatXYZ[2]), xyz[2]);
    //pixel->splatXYZ[0] = xyz[0];
    //pixel->splatXYZ[1] = xyz[1];
    //pixel->splatXYZ[2] = xyz[2];
}

void LightfieldFilm::WriteImage(float splatScale)
{
    if (pixels == NULL) {
        Warning("Trying to read data of Lightfieldfilm without data.");
        return;
    }
    
    WriteImage(splatScale, 0, 0);

    return;
}

void LightfieldFilm::WriteImage(float splatScale, int ax, int ay) {
  
    if (pixels == NULL) {
        Warning("Trying to read data of Lightfieldfilm without data.");
        return;
    }
    
    //Range sanity
    if (ax < 0 || ax >= resAngleX ||
        ay < 0 || ay >= resAngleY )
	Severe("Out of range ax or ay in filter writeimage.\n");
  
    // Convert image to RGB for y,a_y slice
    int nPix = xResolution * yResolution;
    float *rgb = new float[3*nPix];
    
    // Convert image to RGB and compute final pixel values
    int offset = 0;
    for (int y = 0; y < yResolution; ++y) {
        for (int x = 0; x < xResolution; ++x) {
	  
		//Get pixel
		Pixel* pixel = &( (*pixels).at( l2px(x, y, ax, ay) ) );

	        // Convert pixel XYZ color to RGB
		XYZToRGB(pixel->Lxyz, &rgb[3*offset]);

		// Normalize pixel with weight sum
		float weightSum = pixel->weightSum;
		if (weightSum != 0.f) {
		    float invWt = 1.f / weightSum;
		    rgb[3*offset  ] = max(0.f, rgb[3*offset  ] * invWt);
		    rgb[3*offset+1] = max(0.f, rgb[3*offset+1] * invWt);
		    rgb[3*offset+2] = max(0.f, rgb[3*offset+2] * invWt);
		}
	      
		// Add splat value at pixel
		float splatRGB[3];
		XYZToRGB( pixel->splatXYZ, splatRGB );
		rgb[3*offset  ] += splatScale * splatRGB[0];
		rgb[3*offset+1] += splatScale * splatRGB[1];
		rgb[3*offset+2] += splatScale * splatRGB[2];
		++offset;
	}
    }

    // Write RGB image
    std::string fname;        
    std::ostringstream convert;   
    
    convert << std::setfill('0') << std::setw(3) << ax;
    convert << "_";
    convert << std::setfill('0') << std::setw(3) << ay;
    convert << "_";
    convert << filename;
    fname = convert.str(); 
    
    ::WriteImage(fname, rgb, NULL, xResolution, yResolution,
                 xResolution, yResolution, 0, 0);

    // Release temporary image memory
    delete[] rgb;
}

void LightfieldFilm::WriteImageScanLine(float splatScale, int y, int a_y) {
  
    if (pixels == NULL) {
        Warning("Trying to read data of Lightfieldfilm without data.");
        return;
    }
    
    // Convert image to RGB for y,a_y slice
    int nPix = xResolution * resAngleX;
    float *rgb = new float[3*nPix];
    
    // Convert image to RGB and compute final pixel values
    int offset = 0;
    for (int ax = 0; ax < resAngleX; ++ax) {
        for (int x = 0; x < xResolution; ++x) {
	  
		//Get pixel
		Pixel* pixel = &( (*pixels).at( l2px(x, y, ax, a_y) ) );

	        // Convert pixel XYZ color to RGB
		XYZToRGB(pixel->Lxyz, &rgb[3*offset]);

		// Normalize pixel with weight sum
		float weightSum = pixel->weightSum;
		if (weightSum != 0.f) {
		    float invWt = 1.f / weightSum;
		    rgb[3*offset  ] = max(0.f, rgb[3*offset  ] * invWt);
		    rgb[3*offset+1] = max(0.f, rgb[3*offset+1] * invWt);
		    rgb[3*offset+2] = max(0.f, rgb[3*offset+2] * invWt);
		}
	      
		// Add splat value at pixel
		float splatRGB[3];
		XYZToRGB( pixel->splatXYZ, splatRGB );
		rgb[3*offset  ] += splatScale * splatRGB[0];
		rgb[3*offset+1] += splatScale * splatRGB[1];
		rgb[3*offset+2] += splatScale * splatRGB[2];
		++offset;
	}
    }

    // Write RGB image
    ::WriteImage(filename, rgb, NULL, xResolution, resAngleX,
                 xResolution, resAngleX, 0, 0);

    // Release temporary image memory
    delete[] rgb;
}

void LightfieldFilm::GetPixels(float* rgb, float splatScale)
{
    if (pixels == NULL) {
        Warning("Trying to read data of Lightfieldfilm without data.");
        return;
    }
    
    //Right size in array assumed 
    // (     int nPix = xResolution * yResolution * resAngleX * resAngleY;
    //	     float *rgb = new float[3*nPix];     )
    
    // Convert image to RGB and compute final pixel values
    int offset = 0;
    for (int y = 0; y < yResolution; ++y) {
        for (int x = 0; x < xResolution; ++x) {
	    for (int ay = 0; ay < resAngleY; ++ay) {
		for (int ax = 0; ax < resAngleX; ++ax) {

		    //Get pixel
		    Pixel* pixel = &( (*pixels).at( l2px(x, y, ax, ay) ) );

		    // Convert pixel XYZ color to RGB
		    XYZToRGB(pixel->Lxyz, &rgb[3*offset]);

		    // Normalize pixel with weight sum
		    float weightSum = pixel->weightSum;
		    if (weightSum != 0.f) {
			float invWt = 1.f / weightSum;
			rgb[3*offset  ] = max(0.f, rgb[3*offset  ] * invWt);
			rgb[3*offset+1] = max(0.f, rgb[3*offset+1] * invWt);
			rgb[3*offset+2] = max(0.f, rgb[3*offset+2] * invWt);
		    }
		  
		    // Add splat value at pixel
		    float splatRGB[3];
		    XYZToRGB( pixel->splatXYZ, splatRGB );
		    rgb[3*offset  ] += splatScale * splatRGB[0];
		    rgb[3*offset+1] += splatScale * splatRGB[1];
		    rgb[3*offset+2] += splatScale * splatRGB[2];
		    ++offset;
	    
		}
	    }
        }
    }
}

void LightfieldFilm::GetSampleExtent(int *xstart, int *xend,
                                int *ystart, int *yend) const {
    *xstart = 0;
    *xend   = xResolution;
    *ystart = 0;
    *yend   = yResolution;
}


void LightfieldFilm::GetPixelExtent(int *xstart, int *xend,
                               int *ystart, int *yend) const {
    *xstart = 0;
    *xend   = xResolution;
    *ystart = 0;
    *yend   = yResolution;
}

LightfieldFilm *CreateLightfieldFilm(const ParamSet &params, bool allocData) 
{
    int xres = params.FindOneInt("xresolution", 640);
    int yres = params.FindOneInt("yresolution", 480);
    
    float aMinX = params.FindOneFloat("xMinAngle", -10.0);
    float aMaxX = params.FindOneFloat("xMaxAngle", 10.0);
    int rAX = params.FindOneInt("xAngleResolution", 100);
    
    float aMinY = params.FindOneFloat("yMinAngle", -10.0);
    float aMaxY = params.FindOneFloat("yMaxAngle", 10.0);
    int rAY = params.FindOneInt("yAngleResolution", 100);
    
    float rayOffset = params.FindOneFloat("rayOffset", -1.0);
    
    int tRes = params.FindOneInt("timeResolution", 1);
    
    //Additional image params
    string filename = params.FindOneString("filename", PbrtOptions.imageFile);
    if (filename == "")
#ifdef PBRT_HAS_OPENEXR
        filename = "pbrt.exr";
#else
        filename = "pbrt.tga";
#endif

    return new LightfieldFilm(xres, yres,
			      aMinX, aMaxX, rAX,
			      aMinY, aMaxY, rAY,
			      rayOffset,
			      tRes,
			      filename, allocData );
}


