
/*
    Lightfield reconstruction data.
 */

// film/image.cpp*
#include "stdafx.h"
#include "film/lightfield_reconstruction_data.h"
#include "spectrum.h"
#include "parallel.h"
#include "imageio.h"

#include <sstream>
#include <string>
#include <iomanip>

// LightfieldReconstructionData Method Definitions
LightfieldReconstructionData::LightfieldReconstructionData(int xres, int yres,
						   int rAX, int rAY, float* rgb_data)
    : LightfieldReconstruction(xres, yres, rAX, rAY)
{
    pixels = new std::vector<LFSamplePixel>(xres * yres * rAX * rAY);
    SetData(rgb_data);
    
    return;
}

LightfieldReconstructionData::~LightfieldReconstructionData() 
{
    delete pixels;
}
    
void LightfieldReconstructionData::updateSize(int xres, int yres,
					      int rAX, int rAY)
{
    // Change size
    xResolution = xres;
    yResolution = yres;
    
    resAngleX = rAX;
    resAngleY = rAY;

    // Allocate new film image storage if neccessary
    delete pixels;
    pixels = new std::vector<LFSamplePixel>(xres * yres * rAX * rAY);
}

//Helper function from lightfield to pixel indices
int LightfieldReconstructionData::l2px(int x, int y, int ax, int ay)
{
    int res_idx = y * (xResolution * resAngleY * resAngleX) + 
		  x  * (resAngleY * resAngleX) + 
		  ay * (resAngleX) + 
		  ax;
		  
    return res_idx;
}

void LightfieldReconstructionData::SetData(float* rgb_data)
{
    //Leave if null
    if( rgb_data == NULL )
      return;
    
    //Read interleaved rgb array
    for(int y = 0; y < yResolution; y++)
      for(int x = 0; x < xResolution; x++)
	for(int ay = 0; ay < resAngleY; ay++)
	  for(int ax = 0; ax < resAngleX; ax++)
	  {
	      //Pixel idx
	      int idx = l2px(x, y, ax, ay);
	    
	      pixels->at(idx).x = x;
	      pixels->at(idx).y = y;
	      pixels->at(idx).ax = ax;
	      pixels->at(idx).ay = ay;
	    
	      for(int ch = 0; ch < 3; ch++)
	      {
		  //Input index
		  int idx_in = ax * (resAngleY * 3 * xResolution * yResolution) + 
			      ay * (3 * xResolution * yResolution) + 
			      ch * (xResolution * yResolution) +
			      x * yResolution +
			      y;
		  
		  //Transfer
		  pixels->at(idx).rgb[ch] = rgb_data[idx_in];
	      }
	  }
      
    return;
}

void LightfieldReconstructionData::GetPixel(const LFSamplePixel &s, float* rgb_val)
{   
    //Range sanity
    if (s.x  < 0 || s.x  >= xResolution ||
        s.y  < 0 || s.y  >= yResolution ||
        s.ax < 0 || s.ax >= resAngleX ||
        s.ay < 0 || s.ay >= resAngleY )
	return;
    
    LFSamplePixel* pixel = &( pixels->at( l2px(s.x, s.y, s.ax, s.ay) ) );
    rgb_val[0] = pixel->rgb[0];
    rgb_val[1] = pixel->rgb[1];
    rgb_val[2] = pixel->rgb[2];
}


void LightfieldReconstructionData::WriteImageScanLine(int y, int a_y, string filename) 
{
  
    // Convert image to RGB for y, a_y slice
    int nPix = xResolution * resAngleX;
    float *rgb = new float[3*nPix];
    
    // Convert image to RGB and compute final pixel values
    int offset = 0;
    for (int ax = 0; ax < resAngleX; ++ax) 
    {
        for (int x = 0; x < xResolution; ++x) 
	{
	  
		//Get pixel
		LFSamplePixel* pixel = &( (*pixels).at( l2px(x, y, ax, a_y) ) );

	     	//Output
		rgb[3*offset  ] = pixel->rgb[0];
		rgb[3*offset+1] = pixel->rgb[1];
		rgb[3*offset+2] = pixel->rgb[2];
		++offset;
	}
    }

    // Write RGB image
    ::WriteImage(filename, rgb, NULL, xResolution, resAngleX,
                 xResolution, resAngleX, 0, 0);

    // Release temporary image memory
    delete[] rgb;
}

void LightfieldReconstructionData::WriteImage(int ax, int ay, string filename) 
{
  
    //Range sanity
    if (ax < 0 || ax >= resAngleX ||
        ay < 0 || ay >= resAngleY )
	Severe("Out of range ax or ay in filter writeimage.\n");
  
    // Convert image to RGB for y,a_y slice
    int nPix = xResolution * yResolution;
    float *rgb = new float[3*nPix];
    
    // Convert image to RGB and compute final pixel values
    int offset = 0;
    for (int y = 0; y < yResolution; ++y)
    {
        for (int x = 0; x < xResolution; ++x)
	{
		//Get pixel
		LFSamplePixel* pixel = &( (*pixels).at( l2px(x, y, ax, ay) ) );

		//Output
		rgb[3*offset  ] = pixel->rgb[0];
		rgb[3*offset+1] = pixel->rgb[1];
		rgb[3*offset+2] = pixel->rgb[2];
		++offset;
	}
    }

    // Write RGB image
    std::string fname;        
    std::ostringstream convert;   
    
    convert << std::setfill('0') << std::setw(3) << ax;
    convert << "_";
    convert << std::setfill('0') << std::setw(3) << ay;
    convert << "_";
    convert << filename;
    fname = convert.str(); 
    
    ::WriteImage(fname, rgb, NULL, xResolution, yResolution,
                 xResolution, yResolution, 0, 0);

    // Release temporary image memory
    delete[] rgb;
}
