
/*
    Lightfield reconstruction with layered attenuator.
 */

// film/image.cpp*
#include "stdafx.h"
#include "film/lightfield_reconstruction_layered_attenuator.h"
#include "spectrum.h"
#include "parallel.h"
#include "imageio.h"

#include <sstream>
#include <string>
#include <iomanip>

// LightfieldReconstructionData Method Definitions
LightfieldReconstructionLayerAtt::LightfieldReconstructionLayerAtt(int xres, int yres,
								  float aMinX, float aMaxX, int rAX,
								  float aMinY, float aMaxY, int rAY,

								  //(all parameters below in reconstruction context)
								  float OX, float OY, float OZ,
								  float lightfieldSzX, float lightfieldSzY,
								  
								  int layerResX, int layerResY, int layerResZ,
								  float layerOX, float layerOY, float layerOZ,
								  float layerSzX, float layerSzY, 
								  float layerDst,
								   
								  float scRec,
								   
								  int basisFunc,
								   
								  float* rgb_data)

    : LightfieldReconstruction(xres, yres, rAX, rAY) //Set lighfield resolutions
{  
    //Set addional lightfield resolution params
    angleMinX = aMinX;
    angleMaxX = aMaxX;
    angleMinY = aMinY;
    angleMaxY = aMaxY;
  
    //Set reconstruction parameters (all parameters below in reconstruction context)
    pixels = NULL;
    updateLayerParams(OX, OY, OZ,	
		      lightfieldSzX, lightfieldSzY,
		      layerResX, layerResY, layerResZ,
		      layerOX, layerOY,layerOZ,
		      layerSzX, layerSzY, layerDst,
		      scRec,
		      basisFunc);
    
    //Set data
    SetData(rgb_data);
    
    return;
}

LightfieldReconstructionLayerAtt::~LightfieldReconstructionLayerAtt() 
{
    if( pixels != NULL )
	delete[] pixels;
}
    
void LightfieldReconstructionLayerAtt::updateSize(int xres, int yres,
						  int rAX, int rAY)
{
    // Change size of lightfield
    xResolution = xres;
    yResolution = yres;
    
    resAngleX = rAX;
    resAngleY = rAY;
    
    //Compute derived parameters
    computeDerivedParameters();
}

void LightfieldReconstructionLayerAtt::updateLayerParams(float OX, float OY, float OZ,	  
							float lightfieldSzX, float lightfieldSzY,
							int layerResX, int layerResY, int layerResZ,
							float layerOX, float layerOY, float layerOZ,
							float layerSzX, float layerSzY, 
							float layerDst,
							float scRec,
							int basisFunc)
{
    //(all parameters below in reconstruction context)
  
    //Set lightfield params
    origX = OX; 
    origY = OY;
    origZ = OZ;
    
    lightfieldSizeX = lightfieldSzX;
    lightfieldSizeY = lightfieldSzY;
    
    //Set layer params
    resLayerX = layerResX;
    resLayerY = layerResY;
    resLayerZ = layerResZ;
    
    layerOrigX = layerOX; 
    layerOrigY = layerOY;
    layerOrigZ = layerOZ;
    
    layerSizeX = layerSzX;
    layerSizeY = layerSzY;
    
    layerDist = layerDst;
    
    scaleRec = scRec; 

    //Sanity checked basisfunction
    basisFunction = basisFunc;
    if( basisFunction != 0 ) 
    {
        Warning("Basis function type currently not supported! Falling back to 0\n");
	basisFunction = 0;
    }
    
    //Compute derived parameters
    computeDerivedParameters();
    
    // Allocate new film image storage if neccessary
    if( pixels != NULL )
      delete[] pixels;
    
    pixels = new float[resLayerX * resLayerY * resLayerZ * 3];
}

void LightfieldReconstructionLayerAtt::computeDerivedParameters()
{
    //Computed derived parameters for faster sampling
    //angles in v-units
    angleMinXV = tan( Radians(angleMinX) );
    angleMaxXV = tan( Radians(angleMaxX) );
    angleMinYV = tan( Radians(angleMinY) );
    angleMaxYV = tan( Radians(angleMaxY) );
                                                                  
    // absolute x coordinates for light field pixel centers
    lightFieldPixelSizeX = lightfieldSizeX / float(xResolution);
    lightFieldPixelSizeY = lightfieldSizeY / float(yResolution);
        
    // scale to be applied to the float-positions to get a basis function index
    lookupScaleX = float(resLayerX) / layerSizeX;
    lookupScaleY = float(resLayerY) / layerSizeY;
}

//Helper function from layer to pixel indices (without color)
int LightfieldReconstructionLayerAtt::i2px(int x, int y, int z)
{
    int res_idx = z * (resLayerY  *  resLayerX) + 
		  y * (resLayerX) + 
		  x;

    return res_idx;
}

void LightfieldReconstructionLayerAtt::SetData(float* rgb_data)
{
    //Leave if null
    if( rgb_data == NULL )
      return;
    
    //Read interleaved rgb array
    for(int z = 0; z < resLayerZ; z++)
      for(int y = 0; y < resLayerY; y++)
	for(int x = 0; x < resLayerX; x++)
	{
	    //Pixel idx (without color)
	    int idx = i2px(x, y, z);
	  
	    for(int ch = 0; ch < 3; ch++)
	    {
		//Input index
		int idx_in = ch * (resLayerZ * resLayerX * resLayerY) +
			     z  * (resLayerX * resLayerY) +
			     x  *  resLayerY +
			     y;
		
		//Transfer
		pixels[idx*3 + ch] = rgb_data[idx_in];
	    }
	}
      
    return;
}

void LightfieldReconstructionLayerAtt::GetPixel(const LFSamplePixel &s, float* rgb_val)
{   
    //Range sanity
    if (s.x  < 0 || s.x  >= xResolution ||
        s.y  < 0 || s.y  >= yResolution ||
        s.ax < 0 || s.ax >= resAngleX ||
        s.ay < 0 || s.ay >= resAngleY )
	return;
    
    // Angular direction in radians (v-units)
    // in continuous coordinates (pbrt)
    float vx =  (float(s.ax) + 0.5f) / float(resAngleX) * ( angleMaxXV - angleMinXV ) + angleMinXV;
    float vy =  (float(s.ay) + 0.5f) / float(resAngleY) * ( angleMaxYV - angleMinYV ) + angleMinYV;

    // get current pixel pos
    // in continuous coordinates (pbrt)
    float lightFieldPixelCenterX = (origX + 0.5f*lightFieldPixelSizeX) + float(s.x) * lightFieldPixelSizeX;
    float lightFieldPixelCenterY = (origY + 0.5f*lightFieldPixelSizeY) + float(s.y) * lightFieldPixelSizeY;    

    // shift light field pixels to pixel centers of 1st volume slice
    float rayPositionX = lightFieldPixelCenterX - origZ*vx + layerOrigZ*vx;
    float rayPositionY = lightFieldPixelCenterY - origZ*vy + layerOrigZ*vy;   
    
    //Iterate over layers and add to current result
    rgb_val[0] = 0.f;
    rgb_val[1] = 0.f;
    rgb_val[2] = 0.f;

    for( int layer = 0; layer < resLayerZ; layer++ )
    {
        if( basisFunction == 0 ) 
	{
	    // nearest neighbor interpolation (basisfunction 0)

	    // get currently effected layer pixels (convert to discrete by floor)                  
	    int layerPixelIdxForRayX = Floor2Int( lookupScaleX * (rayPositionX - layerOrigX));                        
	    int layerPixelIdxForRayY = Floor2Int( lookupScaleY * (rayPositionY - layerOrigY));   

	    // kick out stuff that's outside
	    bool addContribution = true;
	    if (layerPixelIdxForRayX  < 0 || layerPixelIdxForRayX >= resLayerX ||
		layerPixelIdxForRayY  < 0 || layerPixelIdxForRayY >= resLayerY )
		addContribution = false;
	    
	    // Add contribution
	    if( addContribution )
	    {
		int idx = i2px(layerPixelIdxForRayX, layerPixelIdxForRayY, layer);
		rgb_val[0] += pixels[idx*3 + 0];
		rgb_val[1] += pixels[idx*3 + 1];
		rgb_val[2] += pixels[idx*3 + 2];
	    }
	}
	
        // shift x position of rays to next slice
        rayPositionX = rayPositionX - layerDist*vx;
        rayPositionY = rayPositionY - layerDist*vy;
    }
    
    //Apply scale
    rgb_val[0] *= scaleRec;
    rgb_val[1] *= scaleRec;
    rgb_val[2] *= scaleRec;
    
    //Apply exponential
    rgb_val[0] = exp( rgb_val[0] );
    rgb_val[1] = exp( rgb_val[1] );
    rgb_val[2] = exp( rgb_val[2] );
    
    return;
}


void LightfieldReconstructionLayerAtt::WriteImageScanLine(int y, int a_y, string filename) 
{
  
    // Convert image to RGB for y, a_y slice
    int nPix = xResolution * resAngleX;
    float *rgb = new float[3*nPix];
    
    // Convert image to RGB
    LFSamplePixel queryPixel;
    float rgb_val[3];
    
    int offset = 0;
    for (int ax = 0; ax < resAngleX; ++ax) 
    {
        for (int x = 0; x < xResolution; ++x) 
	{
		//Compute pixel
		queryPixel.x = x;
		queryPixel.y = y;
		queryPixel.ax = ax;
		queryPixel.ay = a_y;
		
		//Get pixel
		GetPixel(queryPixel, rgb_val);

	     	//Output
		rgb[3*offset  ] = rgb_val[0];
		rgb[3*offset+1] = rgb_val[1];
		rgb[3*offset+2] = rgb_val[2];
		++offset;
	}
    }

    // Write RGB image
    ::WriteImage(filename, rgb, NULL, xResolution, resAngleX,
                 xResolution, resAngleX, 0, 0);

    // Release temporary image memory
    delete[] rgb;
}

void LightfieldReconstructionLayerAtt::WriteImage(int ax, int ay, string filename) 
{
  
    //Range sanity
    if (ax < 0 || ax >= resAngleX ||
        ay < 0 || ay >= resAngleY )
	Severe("Out of range ax or ay in filter writeimage.\n");
  
    // Convert image to RGB for y,a_y slice
    int nPix = xResolution * yResolution;
    float *rgb = new float[3*nPix];
    
    // Convert image to RGB
    LFSamplePixel queryPixel;
    float rgb_val[3];
    
    int offset = 0;
    for (int y = 0; y < yResolution; ++y)
    {
        for (int x = 0; x < xResolution; ++x)
	{
		//Compute pixel
		queryPixel.x = x;
		queryPixel.y = y;
		queryPixel.ax = ax;
		queryPixel.ay = ay;
		
		//Get pixel
		GetPixel(queryPixel, rgb_val);

		//Output
		rgb[3*offset  ] = rgb_val[0];
		rgb[3*offset+1] = rgb_val[1];
		rgb[3*offset+2] = rgb_val[2];
		++offset;
	}
    }

    // Write RGB image
    std::string fname;        
    std::ostringstream convert;   
    
    convert << std::setfill('0') << std::setw(3) << ax;
    convert << "_";
    convert << std::setfill('0') << std::setw(3) << ay;
    convert << "_";
    convert << filename;
    fname = convert.str(); 
    
    ::WriteImage(fname, rgb, NULL, xResolution, yResolution,
                 xResolution, yResolution, 0, 0);

    // Release temporary image memory
    delete[] rgb;
}

void LightfieldReconstructionLayerAtt::WriteImage(int z, string filename)
{
    //Range sanity
    if (z < 0 || z >= resLayerZ  )
	Severe("Out of range layer z in writeimage.\n");

    // Write RGB image
    std::string fname;        
    std::ostringstream convert;   
    
    convert << std::setfill('0') << std::setw(3) << z;
    convert << "_layer_";
    convert << filename;
    fname = convert.str(); 
    
    float* layerRgb = &(pixels[z * resLayerX * resLayerY * 3]);
    ::WriteImage(fname, layerRgb, NULL, resLayerX, resLayerY,
                 resLayerX, resLayerY, 0, 0);  
}

void LightfieldReconstructionLayerAtt::GetView(float* view_rgb, int ax, int ay, int dx, int dy)
{
  //Range sanity
    if (ax < 0  || ax >= resAngleX ||
        ay < 0  || ay >= resAngleY ||
        dx <= 0 || dy <= 0)
    {
	Warning("Out of range ax, ay, dx or dy in getView.\n");
	return;
    }
  
    // Convert image to RGB
    LFSamplePixel queryPixel;
    float rgb_val[3];
    
    //Result dimensions
    int yResResult = Floor2Int( float(yResolution - 1) / float(dy) ) + 1;
    int xResResult = Floor2Int( float(xResolution - 1) / float(dx) ) + 1 ;
    
    //Iterate over view
    for (int y = 0; y < yResolution; y += dy)
    {
        for (int x = 0; x < xResolution; x += dx)
	{
		//Compute pixel
		queryPixel.x = x;
		queryPixel.y = y;
		queryPixel.ax = ax;
		queryPixel.ay = ay;
		
		//Get pixel
		GetPixel(queryPixel, rgb_val);
		
		//Iterate over channels
		for( int ch = 0; ch < 3; ch++ )
		{
		      //Compute output index
		      int idx_out = ch      * (xResResult * yResResult) +
				    (x/dx)  * yResResult +
				    (y/dy);

		      //Output
		      view_rgb[idx_out] = rgb_val[ch];
		}
	}
    }
}
