
/*
    Lightfield reconstruction with a time-multiplexed layered attenuator.
 */

// film/image.cpp*
#include "stdafx.h"
#include "film/lightfield_reconstruction_timemultiplexed_layered_attenuator.h"
#include "spectrum.h"
#include "parallel.h"
#include "imageio.h"

#include <sstream>
#include <string>
#include <iomanip>

// LightfieldReconstructionTimeLayerAtt Method Definitions
LightfieldReconstructionTimeLayerAtt::LightfieldReconstructionTimeLayerAtt(int xres, int yres,
								  float aMinX, float aMaxX, int rAX,
								  float aMinY, float aMaxY, int rAY,

								  //(all parameters below in reconstruction context)
								  float OX, float OY, float OZ,
								  float lightfieldSzX, float lightfieldSzY,
								  
								  int layerResX, int layerResY, int layerResZ,
								  float layerOX, float layerOY, float layerOZ,
								  float layerSzX, float layerSzY, 
								  float layerDst,
								  
								  float scRec,
								   
								  int basisFunc,
								   
								  int timeStps, 
								  std::vector<int> *lyrSubSet)
    : LightfieldReconstruction(xres, yres, rAX, rAY) //Set lighfield resolutions
{  
    //Set addional lightfield resolution params
    angleMinX = aMinX;
    angleMaxX = aMaxX;
    angleMinY = aMinY;
    angleMaxY = aMaxY;
  
    //Set reconstruction parameters (all parameters below in reconstruction context)
    pixels = NULL;
    layerSubSet.clear();
    updateLayerParams(OX, OY, OZ,	
		      lightfieldSzX, lightfieldSzY,
		      layerResX, layerResY, layerResZ,
		      layerOX, layerOY,layerOZ,
		      layerSzX, layerSzY, layerDst,
		      scRec,
		      basisFunc,
		      timeStps, lyrSubSet );
    
    return;
}

LightfieldReconstructionTimeLayerAtt::~LightfieldReconstructionTimeLayerAtt() 
{
    if( pixels != NULL )
	delete[] pixels;
}
    
void LightfieldReconstructionTimeLayerAtt::updateSize(int xres, int yres,
						  int rAX, int rAY)
{
    // Change size of lightfield
    xResolution = xres;
    yResolution = yres;
    
    resAngleX = rAX;
    resAngleY = rAY;
    
    //Compute derived parameters
    computeDerivedParameters();
}

void LightfieldReconstructionTimeLayerAtt::updateLayerParams(float OX, float OY, float OZ,	  
							float lightfieldSzX, float lightfieldSzY,
							int layerResX, int layerResY, int layerResZ,
							float layerOX, float layerOY, float layerOZ,
							float layerSzX, float layerSzY, 
							float layerDst,
							float scRec,
							int basisFunc,
							int timeStps, 
							std::vector<int> *lyrSubSet)
{
    //(all parameters below in reconstruction context)
  
    //Set lightfield params
    origX = OX; 
    origY = OY;
    origZ = OZ;
    
    lightfieldSizeX = lightfieldSzX;
    lightfieldSizeY = lightfieldSzY;
    
    //Set layer params
    resLayerX = layerResX;
    resLayerY = layerResY;
    resLayerZ = layerResZ;
    
    layerOrigX = layerOX; 
    layerOrigY = layerOY;
    layerOrigZ = layerOZ;
    
    layerSizeX = layerSzX;
    layerSizeY = layerSzY;
    
    layerDist = layerDst;
    
    timeSteps = timeStps;
    
    scaleRec = scRec;
    
    //Fill layers
    layerSubSet.clear();
    if( lyrSubSet != NULL )
    {
	//Fill with predefined layers
	for(int j = 0; j < lyrSubSet->size(); j++)
	    layerSubSet.push_back( lyrSubSet->at(j) );
    }
    else
    {
	//Use all the layers
	for(int z = 0; z < resLayerZ; z++)
	    layerSubSet.push_back( z );
    }

    //Sanity checked basisfunction
    basisFunction = basisFunc;
    if( basisFunction != 0 ) 
    {
        Warning("Basis function type currently not supported! Falling back to 0\n");
	basisFunction = 0;
    }
    
    //Compute derived parameters
    computeDerivedParameters();
    
    // Allocate new film image storage if neccessary
    if( pixels != NULL )
      delete[] pixels;
    
    pixels = new float[resLayerX * resLayerY * timeSteps * layerSubSet.size() * 3];
}

void LightfieldReconstructionTimeLayerAtt::computeDerivedParameters()
{
    //Computed derived parameters for faster sampling
    //angles in v-units
    angleMinXV = tan( Radians(angleMinX) );
    angleMaxXV = tan( Radians(angleMaxX) );
    angleMinYV = tan( Radians(angleMinY) );
    angleMaxYV = tan( Radians(angleMaxY) );
                                                                  
    // absolute x coordinates for light field pixel centers
    lightFieldPixelSizeX = lightfieldSizeX / float(xResolution);
    lightFieldPixelSizeY = lightfieldSizeY / float(yResolution);
        
    // scale to be applied to the float-positions to get a basis function index
    lookupScaleX = float(resLayerX) / layerSizeX;
    lookupScaleY = float(resLayerY) / layerSizeY;
}

//Helper function from layer, time to pixel indices (without color)
int LightfieldReconstructionTimeLayerAtt::i2px(int x, int y, int t, int layerNum)
{
    int res_idx = layerNum * (resLayerY  *  resLayerX * timeSteps) +
		  t * (resLayerY  *  resLayerX) + 
		  y * (resLayerX) + 
		  x;

    return res_idx;
}

void LightfieldReconstructionTimeLayerAtt::SetData(float* rgb_data)
{
    //Leave if null
    if( rgb_data == NULL )
      return;
    
    //Read interleaved rgb array
    for(int layerNum = 0; layerNum < layerSubSet.size(); layerNum++) //The compacted layer number
      for(int t = 0; t < timeSteps; t++)
	for(int y = 0; y < resLayerY; y++)
	  for(int x = 0; x < resLayerX; x++)
	  {
	      //Pixel idx (without color)
	      int idx = i2px(x, y, t, layerNum);
	    
	      for(int ch = 0; ch < 3; ch++)
	      {
		  //Input index (y, x, time, color, layerNum) matlab index
		  int idx_in =  layerNum * (resLayerX * resLayerY * timeSteps * 3) +
				ch * (resLayerX * resLayerY * timeSteps) +
				t  * (resLayerX * resLayerY) +
				x  *  resLayerY +
				y;
		  
		  //Transfer
		  pixels[idx*3 + ch] = rgb_data[idx_in];
	      }
	  }
      
    return;
}

void LightfieldReconstructionTimeLayerAtt::GetPixel(const LFSamplePixel &s, float* rgb_val)
{   
    //Range sanity
    if (s.x  < 0 || s.x  >= xResolution ||
        s.y  < 0 || s.y  >= yResolution ||
        s.ax < 0 || s.ax >= resAngleX ||
        s.ay < 0 || s.ay >= resAngleY )
	return;

    // Angular direction in radians (v-units)
    // in continuous coordinates (pbrt)
    float vx =  (float(s.ax) + 0.5f) / float(resAngleX) * ( angleMaxXV - angleMinXV ) + angleMinXV;
    float vy =  (float(s.ay) + 0.5f) / float(resAngleY) * ( angleMaxYV - angleMinYV ) + angleMinYV;

    // get current pixel pos
    // in continuous coordinates (pbrt)
    float lightFieldPixelCenterX = (origX + 0.5f*lightFieldPixelSizeX) + float(s.x) * lightFieldPixelSizeX;
    float lightFieldPixelCenterY = (origY + 0.5f*lightFieldPixelSizeY) + float(s.y) * lightFieldPixelSizeY;    

    // shift light field pixels to pixel centers of 1st volume slice
    float rayPositionFirstX = lightFieldPixelCenterX - origZ*vx + layerOrigZ*vx;
    float rayPositionFirstY = lightFieldPixelCenterY - origZ*vy + layerOrigZ*vy;   
    
    //Reset result
    rgb_val[0] = 0.f;
    rgb_val[1] = 0.f;
    rgb_val[2] = 0.f;
    
    //Iterate over layers and add to time slot
    vector<float>r_vals_times( timeSteps, 1.f); //Uniform backlight
    vector<float>g_vals_times( timeSteps, 1.f); //Uniform backlight
    vector<float>b_vals_times( timeSteps, 1.f); //Uniform backlight
    
    //Iterate
    for(int layerNum = 0; layerNum < layerSubSet.size(); layerNum++) 
    {
	//Layer
	int layer = layerSubSet.at( layerNum );

	// shift x position of rays to next slice
	float rayPositionX = rayPositionFirstX - float(layer) * layerDist*vx;
	float rayPositionY = rayPositionFirstY - float(layer) * layerDist*vy;

	//Add to result
	if( basisFunction == 0 ) 
	{
	    // nearest neighbor interpolation (basisfunction 0)

	    // get currently effected layer pixels (convert to discrete by floor)                  
	    int layerPixelIdxForRayX = Floor2Int( lookupScaleX * (rayPositionX - layerOrigX));                        
	    int layerPixelIdxForRayY = Floor2Int( lookupScaleY * (rayPositionY - layerOrigY));   

	    // kick out stuff that's outside
	    if (layerPixelIdxForRayX  < 0 || layerPixelIdxForRayX >= resLayerX ||
		layerPixelIdxForRayY  < 0 || layerPixelIdxForRayY >= resLayerY )
	    {
		//Early exit with overall zero
		return;
	    }
	    
	    //Multiply with timeslot
	    for(int t = 0; t < timeSteps; t++)
	    {
		//Pixel idx (without color)
		int idx = i2px(layerPixelIdxForRayX, layerPixelIdxForRayY, t, layerNum);

		r_vals_times.at(t) *= pixels[idx*3 + 0];
		g_vals_times.at(t) *= pixels[idx*3 + 1];
		b_vals_times.at(t) *= pixels[idx*3 + 2];
	    }
	}
    }
    
    //Average
    if( timeSteps > 0 )
    {
	//Sum
	for(int t = 0; t < timeSteps; t++)
	{
	    rgb_val[0] += r_vals_times.at(t);
	    rgb_val[1] += g_vals_times.at(t);
	    rgb_val[2] += b_vals_times.at(t);
	}
      
	//Inv (NOT IN MATRIX MODEL)
	/*
	float invTimeSteps = 1.f / float(timeSteps);
	rgb_val[0] *= invTimeSteps;
	rgb_val[1] *= invTimeSteps;
	rgb_val[2] *= invTimeSteps;
	*/
    }
    
    //Scale
    rgb_val[0] *= scaleRec;
    rgb_val[1] *= scaleRec;
    rgb_val[2] *= scaleRec;

    return;
}

void LightfieldReconstructionTimeLayerAtt::WriteImage(int numLayer, int time, string filename)
{
    //Range sanity
    if (numLayer < 0 || numLayer >= layerSubSet.size() ||
	time < 0 || time >= timeSteps )
	Severe("Out of numLayer or time in writeimage.\n");

    // Write RGB image
    std::string fname;        
    std::ostringstream convert;   
    
    convert << std::setfill('0') << std::setw(3) << numLayer;
    convert << "_layer_";
    convert << std::setfill('0') << std::setw(3) << time;
    convert << "_time_";
    convert << filename;
    fname = convert.str(); 
    
    int startIdx = numLayer * (resLayerY  *  resLayerX * timeSteps) + time * (resLayerY  *  resLayerX); 
		  
    float* layerRgb = &(pixels[startIdx * 3]);
    ::WriteImage(fname, layerRgb, NULL, resLayerX, resLayerY,
                 resLayerX, resLayerY, 0, 0);  
}

void LightfieldReconstructionTimeLayerAtt::GetView(float* view_rgb, int ax, int ay, int dx, int dy)
{
  //Range sanity
    if (ax < 0  || ax >= resAngleX ||
        ay < 0  || ay >= resAngleY ||
        dx <= 0 || dy <= 0)
    {
	Warning("Out of range ax, ay, dx or dy in getView.\n");
	return;
    }
  
    // Convert image to RGB
    LFSamplePixel queryPixel;
    float rgb_val[3];
    
    //Result dimensions
    int yResResult = Floor2Int( float(yResolution - 1) / float(dy) ) + 1;
    int xResResult = Floor2Int( float(xResolution - 1) / float(dx) ) + 1 ;
    
    //Iterate over view
    for (int y = 0; y < yResolution; y += dy)
    {
        for (int x = 0; x < xResolution; x += dx)
	{
		//Compute pixel
		queryPixel.x = x;
		queryPixel.y = y;
		queryPixel.ax = ax;
		queryPixel.ay = ay;
		
		//Get pixel
		GetPixel(queryPixel, rgb_val);
		
		//Iterate over channels
		for( int ch = 0; ch < 3; ch++ )
		{
		      //Compute output index
		      int idx_out = ch      * (xResResult * yResResult) +
				    (x/dx)  * yResResult +
				    (y/dy);

		      //Output
		      view_rgb[idx_out] = rgb_val[ch];
		}
	}
    }
}
