
/*
    pbrt source code Copyright(c) 1998-2012 Matt Pharr and Greg Humphreys.

    This file is part of pbrt.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

    - Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    - Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
    IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
    TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */


// main/pbrt.cpp*
#include "stdafx.h"
#include "api.h"
#include "probes.h"
#include "parser.h"
#include "parallel.h"

// main program
int main(int argc, char *argv[]) {
    Options options;
    vector<string> filenames;
    
    bool sceneOV = false; //Do no scene overview
    float time = 0.0f;
    
    // Process command-line arguments
    for (int i = 1; i < argc; ++i) {
        if (!strcmp(argv[i], "--ncores")) options.nCores = atoi(argv[++i]);
        else if (!strcmp(argv[i], "--outfile")) options.imageFile = argv[++i];
        else if (!strcmp(argv[i], "--quick")) options.quickRender = true;
        else if (!strcmp(argv[i], "--quiet")) options.quiet = true;
        else if (!strcmp(argv[i], "--verbose")) options.verbose = true;
	else if (!strcmp(argv[i], "--scene_ov")) sceneOV = true;
	else if (!strcmp(argv[i], "--time")) time = (float) atoi(argv[++i]);
        else if (!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h")) {
            printf("usage: pbrt [--ncores n] [--outfile filename] [--quick] [--quiet] [--scene_ov] [--time]"
                   "[--verbose] [--help] <filename.pbrt> ...\n");
            return 0;
        }
        else filenames.push_back(argv[i]);
    }
      
    // Print welcome banner
    if (!options.quiet) {
        printf("pbrt version modified %s of %s at %s [Detected %d core(s)]\n",
               PBRT_VERSION, __DATE__, __TIME__, NumSystemCores());
        printf("Copyright (c)1998-2012 Matt Pharr and Greg Humphreys.\n");
        fflush(stdout);
    }
    
    //Initialize
    pbrtInit(options);
    
    // Process scene description
    PBRT_STARTED_PARSING();
    if (filenames.size() == 0) {
        // Parse scene from standard input
        ParseFile("-");
    } else {
        // Parse scene from input files
        for (u_int i = 0; i < filenames.size(); i++)
            if (!ParseFile(filenames[i]))
                Error("Couldn't open scene file \"%s\"", filenames[i].c_str());
    }
    
    //Setup renderer
    int reconstructionType = 0; // No reconstruction (regular full renderer, no adaptive)
    //int reconstructionType = 1; //Full data (default also for >= 3)
    //int reconstructionType = 2; //Layer reconstruciton
    
    int duplicateCheck = 0; //No duplicate check
    //int duplicateCheck = 1; //Online dense checking
    //int duplicateCheck = 2; //Postprocessing with hashing
    
    //Set the sampling params
    int numChains = 1000;
    int chainLength = 10;
    int maxAcceptedSamplesPerChains = 10;
    float mutation_sigma_spatial = 1.f;
    float mutation_sigma_angular = 1.f;
    float thresh_residual = 0.05;
    int numChainDuplicateChecks = 0; 
    int replaceSamples = 0;
    int sampleTime = 0;
    int sampleFull = 0;
    pbrtSetSamplingParams( numChains, chainLength, maxAcceptedSamplesPerChains,
			  mutation_sigma_spatial, mutation_sigma_angular, thresh_residual, 
			   numChainDuplicateChecks, replaceSamples, sampleTime, sampleFull );
    
    pbrtSetupRenderer(reconstructionType, duplicateCheck, sceneOV); 

    //#### Lightfield renderer
    
    //Get film size
    int resX; int resY;
    float angleMinX; float angleMaxX; int resAngleX;
    float angleMinY; float angleMaxY; int resAngleY;
    pbrtGetFilmSize(&resX, &resY,
		    &angleMinX, &angleMaxX, &resAngleX,
		    &angleMinY, &angleMaxY, &resAngleY);
    
    /*
    //Save state
    APIState* st = new APIState();
    pbrtSaveAPIState(*st);
    pbrtDestroyState(); //Destroy
    
    //Restore state
    pbrtRestoreAPIState(*st);
    */
    
    pbrtSetLayerRecParams( 0.f, 0.f, -4.064f,
			   50.f, 50.f,
			      resX, resY,  3,
			      0.f, 0.f, 0.f,
			      50.f,  50.f, 3.f,
			      0.f,
			      0, NULL, 
			      0 );
    
    /*
    float *rgb_test = new float[resX*resX*3];
    pbrtGetView(rgb_test, 2, 1, 1, 1);
    */
    
    //Render
    pbrtRenderWorld(time);
    
    //Write film
    pbrtWriteFilm();
    
    /*
    //#### Test of sample retrieval

    int numSamples = 0;
    pbrtGetSamples(&numSamples, NULL, NULL);
    printf("Number of samples rendered: %d\n",numSamples);
    
    float *rgb_out_ptr = new float[ 3 * numSamples ];
    int *indices_out_ptr = new int[ 4 * numSamples ];
    
    //Read full samples
    pbrtGetSamples(&numSamples, rgb_out_ptr, indices_out_ptr);
    
    //Cleanup
    delete [] rgb_out_ptr;
    delete [] indices_out_ptr;
    
    //Render
    pbrtRenderWorld();
    
     //#### Test of sample retrieval
    pbrtGetSamples(&numSamples, NULL, NULL);
    printf("Number of samples rendered: %d\n",numSamples);
    
    rgb_out_ptr = new float[ 3 * numSamples ];
    indices_out_ptr = new int[ 4 * numSamples ];
    
    //Read full samples
    pbrtGetSamples(&numSamples, rgb_out_ptr, indices_out_ptr);
    
    //Cleanup
    delete [] rgb_out_ptr;
    delete [] indices_out_ptr;
    */
    
    //#### Test of multiscale version
    /*
    //Write film
    float* rgb = NULL; //new float[resX*resY*resAngleX*resAngleY*3];
    pbrtWriteFilm();
    
    /*
    //Change size and do it again
    pbrtUpdateFilmSize(resX*2, resY*2,
		angleMinX, angleMaxX, 3,
		angleMinY, angleMaxY, 3);
    
    //Render
    pbrtRenderWorld();
    
    //Write film
    pbrtWriteFilm();
    
    //Cleanup
    //delete[] rgb;
    */
    
    //#### Destroy renderer
    pbrtDestroyRenderer();

    pbrtCleanup();
    
    return 0;
}


