
/*
    pbrt mex interface
 */

//Mex
#include "mex.h" 
#include "ObjectHandle.h"
#include <string>

// main/pbrt.cpp*
#include "stdafx.h"
#include "api.h"
#include "probes.h"
#include "parser.h"
#include "parallel.h"

//####################################
//#### MEX persistency class functions
//####################################

APIState* create_api_state (mxArray **out) {
	APIState *new_api_state = new APIState;
	//mexPrintf("Pointer before: %#x, ", *out);
	ObjectHandle<APIState> *handle = new ObjectHandle<APIState>(*new_api_state);
	*out = handle->to_mex_handle();
	//mexPrintf("Pointer after: %#x\n", *out);
	
	//mexPrintf("Api state %#x created.\n", handle->get_object_ptr());
	return handle->get_object_ptr();
}

APIState* get_api_state (const mxArray *in) {
	ObjectHandle<APIState>* handle = ObjectHandle<APIState>::from_mex_handle(in);
	APIState *api_state = handle->get_object_ptr();
	
	//mexPrintf("Api state %#x gotten.\n", handle->get_object_ptr());
	return handle->get_object_ptr();
}

void destroy_api_state (const mxArray *in) {
	ObjectHandle<APIState>* handle = ObjectHandle<APIState>::from_mex_handle(in);
	
	//mexPrintf("Api state %#x deleted.\n", handle->get_object_ptr());
	delete handle;
}


//####################################
//#### PBRT sampling functions
//####################################

void setupPbrt(std::string fn, 
	      int reconstructionType, int duplicateCheck,
	      bool sceneOV, 
	       
	      int numChains, 
	      int chainLength, 
	      int maxAcceptedSamplesPerChains,
	       
	      float mutation_sigma_spatial,
	      float mutation_sigma_angular,
	      float thresh_residual,
	      
	      int numChainDuplicateChecks,
	      int replaceSamples,
	      
	      int sampleTime,
	      int sampleFull) 
{
		
    // Optional arguments
    Options options;
    //options.nCores = ?;
    //options.imageFile = ?;
    options.quickRender = false;
    options.quiet = true;
    options.verbose = false;
	
    //Filenames
    vector<string> filenames;
    filenames.push_back(fn);
   
    // Print welcome banner
    if (!options.quiet) {
        printf("pbrt version modified %s of %s at %s [Detected %d core(s)]\n",
               PBRT_VERSION, __DATE__, __TIME__, NumSystemCores());
        printf("Copyright (c)1998-2012 Matt Pharr and Greg Humphreys.\n");
        printf("The source code to pbrt (but *not* the book contents) is covered by the BSD License.\n");
        printf("See the file LICENSE.txt for the conditions of the license.\n");
        fflush(stdout);
    }
    
    //Initialize
    pbrtInit(options);
    
    // Process scene description
    PBRT_STARTED_PARSING();
    if (filenames.size() == 0) {
        // Parse scene from standard input
        ParseFile("-");
    } else {
        // Parse scene from input files
        for (u_int i = 0; i < filenames.size(); i++)
            if (!ParseFile(filenames[i]))
                Error("Couldn't open scene file \"%s\"", filenames[i].c_str());
    }
    
    //Set the sampling params
    pbrtSetSamplingParams( numChains, chainLength, maxAcceptedSamplesPerChains,
			  mutation_sigma_spatial, mutation_sigma_angular, thresh_residual,
			  numChainDuplicateChecks, replaceSamples, sampleTime, sampleFull );
    
    printf("Used sampling params: %d, %d, %d | %3.3f, %3.3f, %3.3f | %d | repl %d  \n",numChains, 
								    chainLength, 
								    maxAcceptedSamplesPerChains,
								    mutation_sigma_spatial,
								    mutation_sigma_angular,
								    thresh_residual,
								    numChainDuplicateChecks, 
								    replaceSamples);
    
    //Setup renderer
    pbrtSetupRenderer(reconstructionType , duplicateCheck, sceneOV); 
    
    printf("Setup: reconstructionType: %d, duplicateCheck %d, sceneOV: %d, sampletime: %d, samplefull: %d \n", reconstructionType, 
									duplicateCheck, 
									sceneOV, (sampleTime > 0), (sampleFull > 0)
	  );
    
    //Debug: print lightfield size
    /*float sizeX, sizeY;
    pbrtGetLightfieldSize(&sizeX, &sizeY);
    
    printf("Lightfieldsize: %3.3f, %3.3f \n", sizeX, sizeY);
    */
    
    return;
}

void destroyPbrt()
{
    //Destroy renderer
    pbrtDestroyRenderer();
    
    //Cleanup
    pbrtCleanup();
    
    return;
}

//####################################
//#### Main mex entry point
//####################################

void mexFunction(int nlhs, mxArray *plhs[], /* Output variables */
		 int nrhs, const mxArray *prhs[]) /* Input variables */
{
    //Sanity of input and output args
    if(nrhs == 0 || !mxIsClass(prhs[0],"char")) 
	mexErrMsgTxt("First input must be operation string.");
    
    //Read operation string and determine action based on string
    int buflen = mxGetN(prhs[0]) * sizeof(mxChar)+1;
    char *opt = (char*)mxMalloc(buflen);
    mxGetString(prhs[0], opt, buflen);

    //Parse opt
    if( !strcmp(opt, "create") || !strcmp(opt, "create_scene_overview")  ) 
    {
	//Sanity
        if( nlhs != 2 || nrhs <= 4 )
	    mexErrMsgTxt("Wrong input/output parameters.");
	
	//Create api state
	APIState* curr_api_state = create_api_state(&plhs[0]);
	
	//Read scene overview
	bool sceneOV = !strcmp(opt, "create_scene_overview");
	
	//Read filenames    
	int fn_buflen = mxGetN(prhs[1]) * sizeof(mxChar)+1;
	char *fn_buf = (char*)mxMalloc(fn_buflen);
	mxGetString(prhs[1], fn_buf, fn_buflen);
	std::string fn = std::string(fn_buf);
	mexPrintf("File loading: '%s'\n", fn.c_str() );
	
	//#### Get the reconstruction type parameter
	float * params_in = (float*)mxGetData(prhs[2]);
	mwSize rectype_param_size = mxGetNumberOfElements(prhs[2]);
	if( rectype_param_size != 1 )
	    mexErrMsgTxt("Rectype parameters have wrong size.\n");
	
	int reconstructionType = Round2Int(params_in[0]);
	
	//#### Get the douplicate check parameter
	params_in = (float*)mxGetData(prhs[3]);
	mwSize dupchk_param_size = mxGetNumberOfElements(prhs[3]);
	if( dupchk_param_size != 1 )
	    mexErrMsgTxt("Rectype parameters have wrong size.\n");
	
	int duplicateCheck = Round2Int(params_in[0]);
	
	//#### Get the sampling parameter
	params_in = (float*)mxGetData(prhs[4]);
	mwSize sampling_param_size = mxGetNumberOfElements(prhs[4]);
	if( sampling_param_size < 7 )
	    mexErrMsgTxt("Sampling parameters have wrong size.\n");
	
	int numChains = Round2Int(params_in[0]);
	int chainLength = Round2Int(params_in[1]); 
	int maxAcceptedSamplesPerChains = Round2Int(params_in[2]);
	float mutation_sigma_spatial = params_in[3];
	float mutation_sigma_angular = params_in[4];
	float thresh_residual = params_in[5];
	int numChainDuplicateChecks = Round2Int(params_in[6]);
	
	int replaceSamples = 0;
	if( sampling_param_size >= 8 )
	    replaceSamples = Round2Int(params_in[7]);
	
	int sampleTime = 0;
	if( sampling_param_size >= 9 )
	    sampleTime = Round2Int(params_in[8]);
	
	int sampleFull = 0;
	if( sampling_param_size >= 10 )
	    sampleFull = Round2Int(params_in[9]);
	
	//#### Actually start api
	setupPbrt(fn,  
		  reconstructionType, duplicateCheck,
		  sceneOV, 
		  
		  numChains, 
		  chainLength, 
		  maxAcceptedSamplesPerChains,
		  mutation_sigma_spatial,
		  mutation_sigma_angular,
		  thresh_residual,
		  numChainDuplicateChecks,
		  replaceSamples,
		  sampleTime,
		  sampleFull
 		);
	
	//#### Save and return film size
	int ndim = 1;
	int dims[1]; dims[0] = 9;
	
	int resX; int resY;
	float angleMinX; float angleMaxX; int resAngleX;
	float angleMinY; float angleMaxY; int resAngleY;
	pbrtGetFilmSize(&resX, &resY,
			&angleMinX, &angleMaxX, &resAngleX,
			&angleMinY, &angleMaxY, &resAngleY);
	
	int timeRes;
	pbrtGetTimeRes(&timeRes);
	
	plhs[1] = mxCreateNumericArray( ndim, dims, mxSINGLE_CLASS, mxREAL );
	if ( plhs[1] == NULL )
	    mexErrMsgTxt("Could not create output lightfield size mxArray.\n");
	
	//Output
	float *ls_ptr = (float *)mxGetPr(plhs[1]);
	ls_ptr[0] = (float)resX;
	ls_ptr[1] = (float)resY;
	ls_ptr[2] = angleMinX;
	ls_ptr[3] = angleMaxX;
	ls_ptr[4] = (float)resAngleX;
	ls_ptr[5] = angleMinY;
	ls_ptr[6] = angleMaxY;
	ls_ptr[7] = (float)resAngleY;
	ls_ptr[8] = (float)timeRes;
	
	//#### Save state
	pbrtSaveAPIState(*curr_api_state);
    }
    else if( !strcmp(opt, "sample_scene_overview") ) 
    {
	//Sanity
        if( nlhs != 1 || nrhs < 2 )
	    mexErrMsgTxt("Wrong input/output parameter.");
	
	//Restore state
	APIState* curr_api_state = get_api_state(prhs[1]);
	pbrtRestoreAPIState(*curr_api_state);
	
	//#### Get time
	float time = 0.0f;
	if( nrhs >= 3 )
	{
	    float *time_tmp = (float *)mxGetData(prhs[2]);
	    time = *time_tmp;
	}
	
	//#### Render
	printf("Rendering t=%3.3f ...", time );
	pbrtRenderWorld(time);
	printf("done.\n");
	
	//#### Get full rgb film
	//Get film size
	int resX; int resY;
	float angleMinX; float angleMaxX; int resAngleX;
	float angleMinY; float angleMaxY; int resAngleY;
	pbrtGetFilmSize(&resX, &resY,
			&angleMinX, &angleMaxX, &resAngleX,
			&angleMinY, &angleMaxY, &resAngleY);
	
	//Get film
	float* rgb_out_tmp = new float[resX * resY * 3];
	pbrtGetFilmImage(rgb_out_tmp);
	
	int ndim = 3;
	int dims[3];
	dims[0] = resY;
	dims[1] = resX;
	dims[2] = 3;
	
	plhs[0] = mxCreateNumericArray(ndim, dims, mxSINGLE_CLASS, mxREAL);
	if (plhs[0] == NULL)
	    mexErrMsgTxt("Could not create output rgb mxArray.\n");
	
	//RGB out
	float *rgb_out_ptr = (float *)mxGetPr(plhs[0]);
	  
	for(int y = 0; y < resY; y++)
	  for(int x = 0; x < resX; x++)
		for(int ch = 0; ch < 3; ch++)
		{
		      int px_idx =  y * resX + x; 
		      rgb_out_ptr[resY*resX*ch + (y + x*resY)] = rgb_out_tmp[3*px_idx + ch];
		}
	
	//#### Cleanup
	delete [] rgb_out_tmp;
	
	//#### Save state
	pbrtSaveAPIState(*curr_api_state);
    }
    else if( !strcmp(opt, "changeres") ) 
    {
      	//Sanity
        if( nlhs != 1 || nrhs != 3 )
	    mexErrMsgTxt("Wrong input/output parameter.");
	
	//#### Restore state
	APIState* curr_api_state = get_api_state(prhs[1]);
	pbrtRestoreAPIState(*curr_api_state);
	
	//#### Get changed dimensions now
	float *params_in = (float *)mxGetData(prhs[2]);
	mwSize dimension_param_size = mxGetNumberOfElements(prhs[2]);
	if( dimension_param_size != 8 )
	    mexErrMsgTxt("Dimension parameters have wrong size.\n");
	
	int resX = Round2Int(params_in[0]);
	int resY = Round2Int(params_in[1]);
	float angleMinX = params_in[2]; 
	float angleMaxX = params_in[3];
	int resAngleX = Round2Int(params_in[4]);
	float angleMinY = params_in[5]; 
	float angleMaxY = params_in[6]; 
	int resAngleY = Round2Int(params_in[7]);
	
	printf("Dimensions updated to %d, %d, %3.3f, %3.3f, %d, %3.3f, %3.3f, %d\n",
		resX, resY, angleMinX, angleMaxX, resAngleX, angleMinY, angleMaxY, resAngleY);
	
        //#### Change size to defined size
	pbrtUpdateFilmSize(resX, resY,
		    angleMinX, angleMaxX, resAngleX,
		    angleMinY, angleMaxY, resAngleY);
	
	//#### Save and return new film size
	int ndim = 1;
	int dims[1]; dims[0] = 8;
	
	pbrtGetFilmSize(&resX, &resY,
			&angleMinX, &angleMaxX, &resAngleX,
			&angleMinY, &angleMaxY, &resAngleY);
	
	plhs[0] = mxCreateNumericArray( ndim, dims, mxSINGLE_CLASS, mxREAL );
	if ( plhs[0] == NULL )
	    mexErrMsgTxt("Could not create output lightfield size mxArray.\n");
	
	//Output
	float *ls_ptr = (float *)mxGetPr(plhs[0]);
	ls_ptr[0] = (float)resX;
	ls_ptr[1] = (float)resY;
	ls_ptr[2] = angleMinX;
	ls_ptr[3] = angleMaxX;
	ls_ptr[4] = (float)resAngleX;
	ls_ptr[5] = angleMinY;
	ls_ptr[6] = angleMaxY;
	ls_ptr[7] = (float)resAngleY;
	
	//#### Save state
	pbrtSaveAPIState(*curr_api_state);
    }
    else if( !strcmp(opt, "change_layerrec_params") ) 
    {
        //Sanity
        if( nlhs != 0 || nrhs < 4 )
	    mexErrMsgTxt("Wrong input/output parameter.");
	
	//#### Restore state
	APIState* curr_api_state = get_api_state(prhs[1]);
	pbrtRestoreAPIState(*curr_api_state);
	
        //#### Get the layer rec parameter
	float *params_in = (float*)mxGetData(prhs[2]);
	mwSize layer_rec_param_size = mxGetNumberOfElements(prhs[2]);
	if( layer_rec_param_size != 17 )
	    mexErrMsgTxt("Layer rec parameters have wrong size.\n");
	
	float lightfieldOX = params_in[0];
	float lightfieldOY = params_in[1];
	float lightfieldOZ = params_in[2];
	
	float lightfieldSzX = params_in[3];
	float lightfieldSzY = params_in[4];
	
	int layerResX = Round2Int(params_in[5]);
	int layerResY = Round2Int(params_in[6]);
	int layerResZ = Round2Int(params_in[7]);
	
	float layerOX = params_in[8];
	float layerOY = params_in[9];
	float layerOZ = params_in[10];
	
	float layerSzX = params_in[11];
	float layerSzY = params_in[12]; 
	
	float layerDst = params_in[13];
	
	float scaleRec = params_in[14];
	
	int timeSteps = Round2Int( params_in[15] );
	
	int basisFunc = Round2Int( params_in[16] );
	
	// Get the layer subset
	float *layerSubset_in = (float*)mxGetData(prhs[3]);
	int layerSubset_size = mxGetNumberOfElements(prhs[3]);
	
	std::vector<int> layerSubset;
	for(int i = 0; i < layerSubset_size; i++)
	{
	    layerSubset.push_back( Round2Int(layerSubset_in[i]) );
	    //printf("Used layer: %d", layerSubset.back() );
	}
	//printf("\n");
		  
	//Set the parameters
	pbrtSetLayerRecParams( lightfieldOX, lightfieldOY, lightfieldOZ,
			       lightfieldSzX, lightfieldSzY,
			      layerResX, layerResY,  layerResZ,
			      layerOX,  layerOY,  layerOZ,
			      layerSzX,  layerSzY, layerDst,
			      scaleRec,
			      timeSteps, &(layerSubset),
			      basisFunc );
	
	/*
	printf("Used layer rec params: %3.3f, %3.3f, %3.3f | %3.3f, %3.3f |\n "
	      "%d, %d, %d | %3.3f, %3.3f, %3.3f | %3.3f | %3.3f, %3.3f | %3.3f | %d | %d \n",
			      lightfieldOX, lightfieldOY, lightfieldOZ,
			      lightfieldSzX, lightfieldSzY,
			      layerResX, layerResY,  layerResZ,
			      layerOX,  layerOY,  layerOZ,
			      scaleRec,
			      layerSzX,  layerSzY, layerDst,
			      timeSteps,
			      basisFunc );*/
	
	//Optionally set reconstruction
	if( nrhs > 4 )
	{
	      float *rec_in = (float *)mxGetData(prhs[4]);
	      mwSize total_num_of_inelements = mxGetNumberOfElements(prhs[4]);
	      pbrtSetReconstruction(rec_in);
	      
	      printf("New reconstruction data set\n " );
	}
	
	//#### Save state
	pbrtSaveAPIState(*curr_api_state);
    }
    else if( !strcmp(opt, "set1DtestData") ) 
    {
        //Sanity
        if( nlhs != 0 || nrhs != 3 )
	    mexErrMsgTxt("Wrong input/output parameter.");
	
	//#### Restore state
	APIState* curr_api_state = get_api_state(prhs[1]);
	pbrtRestoreAPIState(*curr_api_state);
	
	//Get film size
	int resX; int resY;
	float angleMinX; float angleMaxX; int resAngleX;
	float angleMinY; float angleMaxY; int resAngleY;
	pbrtGetFilmSize(&resX, &resY,
			&angleMinX, &angleMaxX, &resAngleX,
			&angleMinY, &angleMaxY, &resAngleY);
	
        //#### Get the data
	float *data_rgb = (float*)mxGetData(prhs[2]);
	const int  *data_rgb_dims = mxGetDimensions(prhs[2]);
	
	//Check for size
	if( 3 != data_rgb_dims[0] ||
	    resAngleY != data_rgb_dims[1] || 
	    resAngleX != data_rgb_dims[2] || 
	    resY != data_rgb_dims[3] || 
	    resX != data_rgb_dims[4] )
	    printf("Input has wrong size\n");

	//Set data
	pbrtSet1DTestData(data_rgb);
	
	//#### Save state
	pbrtSaveAPIState(*curr_api_state);
    }
    else if( !strcmp(opt, "sample") ) 
    {
	//Sanity
        if( nlhs < 2 || nrhs <= 2 )
	    mexErrMsgTxt("Wrong input/output parameter.");
	
	//Restore state
	APIState* curr_api_state = get_api_state(prhs[1]);
	pbrtRestoreAPIState(*curr_api_state);
	
	//Get film size
	int resX; int resY;
	float angleMinX; float angleMaxX; int resAngleX;
	float angleMinY; float angleMaxY; int resAngleY;
	pbrtGetFilmSize(&resX, &resY,
			&angleMinX, &angleMaxX, &resAngleX,
			&angleMinY, &angleMaxY, &resAngleY);
	
	//#### Pass current reconstruction
	float *rec_in = (float *)mxGetData(prhs[2]);
	mwSize total_num_of_inelements = mxGetNumberOfElements(prhs[2]);
	pbrtSetReconstruction(rec_in);
	
	//#### Update sampling params if necessary
	if( nrhs >= 4 )
	{
	    float *params_in = (float *)mxGetData(prhs[3]);
	    mwSize sampling_param_size = mxGetNumberOfElements(prhs[3]);
	    if( sampling_param_size < 5 )
		mexErrMsgTxt("Sampling parameters have wrong size.\n");
	    
	    int chainLength = Round2Int(params_in[0]); 
	    float mutation_sigma_spatial = params_in[1];
	    float mutation_sigma_angular = params_in[2];
	    float thresh_residual = params_in[3];
	    int numChainDuplicateChecks = Round2Int(params_in[4]);
	    
	    int replaceSamples = 0;
	    if( sampling_param_size > 5 )
		replaceSamples = Round2Int(params_in[5]);
	    
	    pbrtSetSamplingParams( -1, chainLength, -1,
				  mutation_sigma_spatial, mutation_sigma_angular, thresh_residual,
				  numChainDuplicateChecks, replaceSamples, -1, -1);

	    printf("Ch_length %d, sigma_sp %3.3f, sigma_ang %3.3f, thresh %3.3f, dup_chk: %d, repl_smpls %d \n",
		   chainLength, mutation_sigma_spatial, mutation_sigma_angular, thresh_residual,
		   numChainDuplicateChecks, replaceSamples );
	}
	
	//#### Get time
	float time = 0.0f;
	if( nrhs >= 5 )
	{
	    float *time_tmp = (float *)mxGetData(prhs[4]);
	    time = *time_tmp;
	}
	
	//#### Get history samples
	bool getHistory = false;
	if( nrhs >= 6 )
	{
	    float *getHistory_tmp = (float *)mxGetData(prhs[5]);
	    getHistory = ((int)roundf(*getHistory_tmp)) > 0;
	}
	

	//#### Render
	printf("Rendering t=%3.3f ...", time );
	pbrtRenderWorld(time);
	printf("done.\n");
	

	//#### Get the samples and return them
	int numSamples;
	pbrtGetSamples(&numSamples, NULL, NULL, getHistory);
	if( !getHistory )
	  printf("Number of samples rendered: %d\n",numSamples);
	else
	  printf("Number of ALL samples: %d\n",numSamples);
	
	//RGB
	int ndim = 2;
	int dims[2];
	dims[0] = 3;
	dims[1] = numSamples;
	
	plhs[0] = mxCreateNumericArray(ndim, dims, mxSINGLE_CLASS, mxREAL);
	if (plhs[0] == NULL)
	    mexErrMsgTxt("Could not create output rgb mxArray.\n");
	
	float *rgb_out_ptr = (float *)mxGetPr(plhs[0]);
	
	
	//Indices
	dims[0] = 4;
	dims[1] = numSamples;
	
	plhs[1] = mxCreateNumericArray(ndim, dims, mxINT32_CLASS, mxREAL);
	if (plhs[1] == NULL)
	    mexErrMsgTxt("Could not create output rgb mxArray.\n");
	
	int *indices_out_ptr = (int *)mxGetPr(plhs[1]);
	
	//Read full samples
	pbrtGetSamples(&numSamples, rgb_out_ptr, indices_out_ptr, getHistory);
	
	//#### Get the timings and return (optional)
	if( nlhs > 2 )
	{
	    std::vector<float> timings;
	    pbrtGetSamplingTimings( &timings );
	    
	    dims[0] = 1;
	    dims[1] = timings.size(); 
	    
	    plhs[2] = mxCreateNumericArray(ndim, dims, mxSINGLE_CLASS, mxREAL);
	    if (plhs[2] == NULL)
		mexErrMsgTxt("Could not create output timings mxArray.\n");
	    
	    float *timings_out_ptr = (float *)mxGetPr(plhs[2]);
	    
	    for( int i = 0; i < timings.size(); i++ )
		timings_out_ptr[i] = timings.at(i);
	}
	
	//#### Save state
	pbrtSaveAPIState(*curr_api_state);
    }
    else if( !strcmp(opt, "getsamples") ) 
    {
	//Sanity
        if( nlhs < 2 || nrhs < 2 )
	    mexErrMsgTxt("Wrong input/output parameter.");
	
	//Restore state
	APIState* curr_api_state = get_api_state(prhs[1]);
	pbrtRestoreAPIState(*curr_api_state);
	
	//#### Get history samples
	bool getHistory = false;
	if( nrhs >= 3 )
	{
	    float *getHistory_tmp = (float *)mxGetData(prhs[2]);
	    getHistory = ((int)roundf(*getHistory_tmp)) > 0;
	}

	//#### Get the samples and return them
	int numSamples;
	pbrtGetSamples(&numSamples, NULL, NULL, getHistory);
	
	//RGB
	int ndim = 2;
	int dims[2];
	dims[0] = 3;
	dims[1] = numSamples;
	
	plhs[0] = mxCreateNumericArray(ndim, dims, mxSINGLE_CLASS, mxREAL);
	if (plhs[0] == NULL)
	    mexErrMsgTxt("Could not create output rgb mxArray.\n");
	
	float *rgb_out_ptr = (float *)mxGetPr(plhs[0]);
	
	
	//Indices
	dims[0] = 4;
	dims[1] = numSamples;
	
	plhs[1] = mxCreateNumericArray(ndim, dims, mxINT32_CLASS, mxREAL);
	if (plhs[1] == NULL)
	    mexErrMsgTxt("Could not create output rgb mxArray.\n");
	
	int *indices_out_ptr = (int *)mxGetPr(plhs[1]);
	
	//Read full samples
	pbrtGetSamples(&numSamples, rgb_out_ptr, indices_out_ptr, getHistory);
	
	//#### Save state
	pbrtSaveAPIState(*curr_api_state);
    }
    else if( !strcmp(opt, "reset_chains") )
    {
	//Sanity
        if( nlhs != 0 || nrhs <= 1 )
	    mexErrMsgTxt("Wrong input/output parameter.");
	
	//#### Restore state
	APIState* curr_api_state = get_api_state(prhs[1]);
	pbrtRestoreAPIState(*curr_api_state);
	
	//Destroy
	pbrtResetChains();
	
	//#### Save state
	pbrtSaveAPIState(*curr_api_state);
    }
    else if( !strcmp(opt, "get_view_uniform") ) 
    {
	//Sanity
        if( nlhs != 1 || nrhs <= 2 )
	    mexErrMsgTxt("Wrong input/output parameter.");
	
	//Restore state
	APIState* curr_api_state = get_api_state(prhs[1]);
	pbrtRestoreAPIState(*curr_api_state);
	
	//Get film size
	int resX; int resY;
	float angleMinX; float angleMaxX; int resAngleX;
	float angleMinY; float angleMaxY; int resAngleY;
	pbrtGetFilmSize(&resX, &resY,
			&angleMinX, &angleMaxX, &resAngleX,
			&angleMinY, &angleMaxY, &resAngleY);
	
	
	//#### Get view sampling description
	float *params_in = (float *)mxGetData(prhs[2]);
	mwSize view_sampling_param_size = mxGetNumberOfElements(prhs[2]);
	if( view_sampling_param_size != 4 )
	    mexErrMsgTxt("View sampling parameters have wrong size.\n");
	
	int ax = Round2Int(params_in[0]); 
	int ay = Round2Int(params_in[1]); 
	int dx = Round2Int(params_in[2]); 
	int dy = Round2Int(params_in[3]);

	//printf("View sampling ax %d, ay %d | d_x %d, d_y %d \n", ax, ay, dx, dy );


	//#### Allocate RGB data
	int ndim = 3;
	int dims[3];
	dims[0] = (dy > 0) ? ( Floor2Int( float(resY - 1) / float(dy) ) + 1 ) : 0;
	dims[1] = (dx > 0) ? ( Floor2Int( float(resX - 1) / float(dx) ) + 1 ) : 0;
	dims[2] = 3;
	
	plhs[0] = mxCreateNumericArray(ndim, dims, mxSINGLE_CLASS, mxREAL);
	if (plhs[0] == NULL)
	    mexErrMsgTxt("Could not create output rgb mxArray.\n");
	
	float *rgb_out_ptr = (float *)mxGetPr(plhs[0]);
	
	//#### Sample the view from the current reconstruction
	pbrtGetView(rgb_out_ptr, ax, ay, dx, dy);
	
	//#### Save state
	pbrtSaveAPIState(*curr_api_state);
    }
    else if( !strcmp(opt, "destroy") )
    {
	//Sanity
        if( nlhs != 0 || nrhs <= 1 )
	    mexErrMsgTxt("Wrong input/output parameter.");
	
	//Restore state
	APIState* curr_api_state = get_api_state(prhs[1]);
	pbrtRestoreAPIState(*curr_api_state);
	
	//Destroy
	destroyPbrt();
	
	//Destroy api state
	destroy_api_state(prhs[1]);
    }
    else 
	mexErrMsgTxt("Bad input.");
    
    return;
}



