
/*
   Adaptive Lightfield sampler renderer
 */


// renderers/adaptiveLFsamplerrenderer.cpp*
#include "stdafx.h"
#include "renderers/adaptiveLFsamplerrenderer.h"
#include "scene.h"
#include "film.h"
#include "volume.h"
#include "sampler.h"
#include "integrator.h"
#include "progressreporter.h"
#include "camera.h"
#include "intersection.h"
#include "timer.h"

#include "film/lightfield_film.h"

static uint32_t hash(char *key, uint32_t len)
{
    uint32_t hash = 0, i;
    for (hash=0, i=0; i<len; ++i) {
        hash += key[i];
        hash += (hash << 10);
        hash ^= (hash >> 6);
    }
    hash += (hash << 3);
    hash ^= (hash >> 11);
    hash += (hash << 15);
    return hash;
} 

// #### Sampling chain definitions
LFSamplingChain::LFSamplingChain(int maxRecSize)
{
    //Initialize all vars
    sample.x = sample.y = sample.ax = sample.ay = 0;
    
    maxRecordedSize = maxRecSize;
    idx = 0;
    initialized = false;
   
    recorded = new std::vector<LFSamplePixel>(maxRecordedSize);
}

// Copy constructor to handle pass by value.
LFSamplingChain::LFSamplingChain(const LFSamplingChain &copyin)
{     
    //Initialize all vars
    sample.x = copyin.sample.x;
    sample.y = copyin.sample.y;
    sample.ax = copyin.sample.ax;
    sample.ay = copyin.sample.ay;
    
    maxRecordedSize = copyin.maxRecordedSize;
    idx = copyin.idx;
    initialized = copyin.initialized;
   
    recorded = new std::vector<LFSamplePixel>(maxRecordedSize);
    
    //Copy record
    for( int i = 0; i < copyin.recorded->size(); i++ )
	recorded->push_back(copyin.recorded->at(i));
   
    return;
}

LFSamplingChain::~LFSamplingChain()
{
    delete recorded;
    recorded = NULL;
}

bool LFSamplingChain::record(LFSamplePixel& acceptedSample)
{
    if( maxRecordedSize != 0 && recorded->size() >= maxRecordedSize )
      return false;
    
    //Record
    recorded->push_back(acceptedSample);
    
    return true;
}

void LFSamplingChain::reset()
{
    //Reset all
    idx = 0;
    recorded->clear();
}

// #### AdaptiveLFSamplerRendererTask Definitions
AdaptiveLFSamplerRendererTask::AdaptiveLFSamplerRendererTask(const Scene *sc, Renderer *ren, Camera *c,
		    ProgressReporter &pr, Sampler *ms, Sample *sam, 
		    bool visIds, 
		    LightfieldReconstruction * lfRec,
		    std::vector<LFSamplingChain> *schains, int chLength, 
		    float m_sigma_spatial, float m_sigma_angular, float t_residual,
		    int duplicateChk, int duplicateChainChk,
		    int *rsamples, float* lf1DtestData,
		    bool smplTime, bool smplFull, SampleSet* smpleSet,
		    int tn, int tc)
  : reporter(pr)
{
    scene = sc; renderer = ren; camera = c; mainSampler = ms;
    origSample = sam; visualizeObjectIds = visIds; 
    
    lightfieldRec = lfRec;
    chains = schains;
    chainLength = chLength;
    mutation_sigma_spatial = m_sigma_spatial;
    mutation_sigma_angular = m_sigma_angular;
    thresh_residual = t_residual;
    
    duplicateCheck = duplicateChk;
    duplicateChainCheck = duplicateChainChk;
    recordedSamples = rsamples;
    
    sampleTime = smplTime;
    sampleFull = smplFull;
    sampleSet = smpleSet;
    
    totalRecordedSamples = 0;
    totalTestedSamples = 0;
    
    taskNum = tn; taskCount = tc;
    
    //Debug range
    dStartX = 500;
    dEndX = 800;
    dStartY = 500;
    dEndY = 650;
    
    lightfield1DtestData = lf1DtestData;
}

void AdaptiveLFSamplerRendererTask::ComputeSubWindow(int num, int count, int *newXStart,
        int *newXEnd, int *newYStart, int *newYEnd) const {
	  
    //Get film (allways lightfield)
    LightfieldFilm * film = (LightfieldFilm*)camera->film;
	
    // Determine how many tiles to use in each dimension, _nx_ and _ny_
    int dx = film->xResolution, dy = film->yResolution;
    int nx = count, ny = 1;
    while ((nx & 0x1) == 0 && 2 * dx * ny < dy * nx) {
        nx >>= 1;
        ny <<= 1;
    }
    Assert(nx * ny == count);

    // Compute $x$ and $y$ pixel sample range for sub-window
    int xo = num % nx, yo = num / nx;
    float tx0 = float(xo) / float(nx), tx1 = float(xo+1) / float(nx);
    float ty0 = float(yo) / float(ny), ty1 = float(yo+1) / float(ny);
    *newXStart = Floor2Int(Lerp(tx0, 0, film->xResolution));
    *newXEnd   = Floor2Int(Lerp(tx1, 0, film->xResolution));
    *newYStart = Floor2Int(Lerp(ty0, 0, film->yResolution));
    *newYEnd   = Floor2Int(Lerp(ty1, 0, film->yResolution));
}
    
void AdaptiveLFSamplerRendererTask::Run() 
{
    //Start
    PBRT_STARTED_RENDERTASK(taskNum);
    
    //#### Generate sampling data
    Sampler *sampler = mainSampler;
    SamplingData samplingData;
    samplingData.sampler = sampler;
    
    MemoryArena arena;
    samplingData.arena = &arena;
    RNG rng(taskNum);
    samplingData.rng = &rng;
    
    // Allocate space for samples and intersections
    int maxSamples = sampler->MaximumSampleCount();
    samplingData.samples = origSample->Duplicate(maxSamples);
    samplingData.rays = new RayDifferential[maxSamples];
    samplingData.Ls = new Spectrum[maxSamples];
    samplingData.Ts = new Spectrum[maxSamples];
    samplingData.isects = new Intersection[maxSamples];
    
    //Get film (allways lightfield)
    LightfieldFilm * film = (LightfieldFilm*)camera->film;
    samplingData.film = film;
    
    //#### Check the sampler
    if( sampleFull )
    {
	// Get linear sub-_Sampler_ for _AdaptiveLFSamplerRendererTask_
	int xPixelStart, xPixelEnd;
	int yPixelStart, yPixelEnd;
	ComputeSubWindow(taskNum, taskCount, &xPixelStart, &xPixelEnd, &yPixelStart, &yPixelEnd);

	//Total pixels
	int totalPixels = (xPixelEnd - xPixelStart) * (yPixelEnd - yPixelStart) * film->resAngleX * film->resAngleY;
	
	//Current sample
	LFSamplePixel sample;
	sample.x = xPixelStart;
	sample.y = yPixelStart;
	sample.ax = 0;
	sample.ay = 0;
	
	 //Get current chain
	LFSamplingChain * currChain = &( chains->at(taskNum) );

	for( int i = 0; i < totalPixels; i++ )
	{
	    //Sample new sample
	    EvaluateLightfield( &sample, &samplingData );
	    totalTestedSamples += 1;
	    
	    //Record
	    LFSamplePixel newSample = sample;
	    currChain->recorded->push_back(newSample);  
	    
	    totalRecordedSamples += 1;

	    if ( sample.y == yPixelEnd)
		break;
	    
	    //Increment
	    sample.ax += 1;
	    if (sample.ax == film->resAngleX){
		sample.ax = 0;
		sample.ay += 1;
	    }
	    
	    if (sample.ay  == film->resAngleY){
		sample.ay  = 0;
		sample.x += 1;
	    }
	    
	    if (sample.x == xPixelEnd) {
		sample.x = xPixelStart;
		sample.y += 1;
	    }
	}
	
    }
    else
    {
    
	// Get linear sub-_Sampler_ for _AdaptiveLFSamplerRendererTask_
	int cStart, cEnd;
	ComputeSubChains(taskNum, taskCount, &cStart, &cEnd);
	
	//#### Iterate over the chains
	for( int cIdx = cStart; cIdx < cEnd; cIdx++ )
	{
	    //Get current chain
	    LFSamplingChain * currChain = &( chains->at(cIdx) );
	
	    //Initialize chain if necessary
	    if( !currChain->initialized )
	    {
		//Next tested sample
		totalTestedSamples += 1;
	      
		//Unique random number
		currChain->sample.x = Floor2Int( rng.RandomFloatUnique() * ( (float)film->xResolution ) );
		currChain->sample.y = Floor2Int( rng.RandomFloatUnique() * ( (float)film->yResolution ) );
		currChain->sample.ax = Floor2Int( rng.RandomFloatUnique() * ( (float)film->resAngleX ) );
		currChain->sample.ay = Floor2Int( rng.RandomFloatUnique() * ( (float)film->resAngleY ) );
		
		//Debug region
		/*
		if( currChain->sample.x < dStartX || currChain->sample.x >= dEndX ||
		    currChain->sample.y < dStartY || currChain->sample.y >= dEndY )
		{
		    currChain->initialized = true; 
		    continue;
		}
		*/
		
		//Sample new initial sample
		EvaluateLightfield( &(currChain->sample), &samplingData );
		
		//Evaluate function
		float val = getFunctionVal( &(currChain->sample) );
		
		//Initialize if value larger than certain amount
		if( val > 0.f )
		{ 
		    //Disabled for debug
		    currChain->initialized = true; 
		    record(currChain, &(currChain->sample), &samplingData);
		}else
		{
		    //Continue with next chain
		    continue;
		}
	    }
	    
	    //Debug region
	    /*
	    if( currChain->sample.x < dStartX || currChain->sample.x >= dEndX ||
		currChain->sample.y < dStartY || currChain->sample.y >= dEndY )
	    {
		continue;
	    }
	    */
	    
	    //Sample the chain with task's parameters	
	    float ar = sampleChain( currChain, 
				    chainLength,
				    mutation_sigma_spatial,
				    mutation_sigma_angular,
				    thresh_residual,
				    &samplingData);
	    
	    //Reset chain if ar lower than bound
	    if( ar < 0.05 )
		currChain->initialized = false;
	    
	    //Manual monte carlo iteration
	    /*
	    //Iterate over chain
	    for( ; 
		(currChain->idx < chainLength) && (currChain->recorded->size() < currChain->maxRecordedSize); 
		currChain->idx += 1)
	    {
		
		//Initialize chain if necessary
		if( !currChain->initialized )
		{
		    //Next tested sample
		    totalTestedSamples += 1;
		  
		    //Unique random number
		    currChain->sample.x = Floor2Int( rng.RandomFloatUnique() * ( (float)film->xResolution ) );
		    currChain->sample.y = Floor2Int( rng.RandomFloatUnique() * ( (float)film->yResolution ) );
		    currChain->sample.ax = Floor2Int( rng.RandomFloatUnique() * ( (float)film->resAngleX ) );
		    currChain->sample.ay = Floor2Int( rng.RandomFloatUnique() * ( (float)film->resAngleY ) );
		    
		    //Repeatable random number
		    //currChain->sample.x = Floor2Int( rng.RandomFloat() * ( (float)film->xResolution ) );
		    //currChain->sample.y = Floor2Int( rng.RandomFloat() * ( (float)film->yResolution ) );
		    //currChain->sample.ax = Floor2Int( rng.RandomFloat() * ( (float)film->resAngleX ) );
		    //currChain->sample.ay = Floor2Int( rng.RandomFloat() * ( (float)film->resAngleY ) );
		    
		    
		    //Disabled for debug
		    currChain->initialized = true;
		}
		
		//Check if already used (only read here)
		int rec_idx = currChain->sample.y  * ( film->xResolution * film->resAngleY * film->resAngleX ) + 
			      currChain->sample.x  * ( film->resAngleY * film->resAngleX ) + 
			      currChain->sample.ay * ( film->resAngleX ) + 
			      currChain->sample.ax;
			
		if( recordedSamples[rec_idx] )
		    continue;
		
		//Evaluate lightfield pixel
		LFSamplePixel newPixel( currChain->sample );
		EvaluateLightfield( &newPixel, &samplingData );
	      
		//Record in chain with unique addition !
		int origVal = AtomicCompareAndSwap(&(recordedSamples[rec_idx]), 1, 0);
		if( origVal == 0 )
		{
		    currChain->record(newPixel);
		    totalRecordedSamples++;
		}
		
	    }//Continue with next chain element
	    */
	} 
    }
   
    //#### Clean up samplingdata manually after Task is done
    //delete sampler;
    delete[] samplingData.samples;
    delete[] samplingData.rays;
    delete[] samplingData.Ls;
    delete[] samplingData.Ts;
    delete[] samplingData.isects;

    //#### Report results
    reporter.Update();
    PBRT_FINISHED_RENDERTASK(taskNum);
}

//## Adaptive sampling 
float AdaptiveLFSamplerRendererTask::sampleChain(LFSamplingChain *chain, 
						  int   n_samples_chain,
						  float mutation_sigma_spatial,
						  float mutation_sigma_angular,
						  float thresh_residual,
						  SamplingData* sD) 
{
    // Stochastic vars
    float fx = 0; float fy = 0;
    int accepted = 0; float a;
    
    // Evaluate initial function value
    LFSamplePixel sample_x( chain->sample );
    fx = getFunctionVal( &sample_x );

    // Stochastic metropolis-style sampling
    LFSamplePixel sample_y;
    for( int i = 0; i < n_samples_chain; i++ )
    {
	//Increase chain idx here
	chain->idx += 1;
	
        // Mutate
        bool accept = mutate( &sample_x, &sample_y, chain, sD,
			      mutation_sigma_spatial,
			      mutation_sigma_angular );
	/*
	if( !accept )
	  continue;
	*/
	
	//Reinitialize
        if( !accept )
	{
	    //Test next initialization sample
	    totalTestedSamples += 1;
	  
	    //Unique random number
	    sample_x.x = Floor2Int( sD->rng->RandomFloatUnique() * ( (float)sD->film->xResolution ) );
	    sample_x.y = Floor2Int( sD->rng->RandomFloatUnique() * ( (float)sD->film->yResolution ) );
	    sample_x.ax = Floor2Int( sD->rng->RandomFloatUnique() * ( (float)sD->film->resAngleX ) );
	    sample_x.ay = Floor2Int( sD->rng->RandomFloatUnique() * ( (float)sD->film->resAngleY ) );
	    
	    //Sample new initial sample
	    EvaluateLightfield( &sample_x, sD );
	    float val = getFunctionVal( &sample_x );
	    
	    //Initialize if value larger than certain amount
	    if( val > 0.f )
		record(chain, &sample_x, sD);
	    else
		continue;
	}

        //Test next sample
        totalTestedSamples += 1;
	
        // Evaluate on mutated sample
	EvaluateLightfield( &sample_y, sD );
	fy = getFunctionVal( &sample_y );

        a = std::min( 1.f, fy / std::max(fx, 1e-12f) );
        if ( fx <= 0.f || sD->rng->RandomFloatUnique() < a )
	{
	    //Record
            if( fy > 0.f )
	    {
                accepted += 1;
		
		// record in chain
		record( chain, &sample_y, sD );
	    }

            //Copy
            sample_x = sample_y;
            fx = fy;
	    
	}else if( fx > 0.0 )
	{
            //Empty
	}
    }
    
    //Save last sample in chain
    chain->sample = sample_x;
    
    return (float)accepted / (float)n_samples_chain;                                
}

float AdaptiveLFSamplerRendererTask::getFunctionVal( LFSamplePixel* sample )
{
  
    //Compute delta difference function for sample
    float residual[3];
    lightfieldRec->GetPixel(*sample, residual);
    residual[0] = sample->rgb[0] - residual[0];
    residual[1] = sample->rgb[1] - residual[1];
    residual[2] = sample->rgb[2] - residual[2];
    
    //Function that we integrate
    float delta = sqrtf( residual[0] * residual[0] + 
			 residual[1] * residual[1] +
			 residual[2] * residual[2] );

    if( delta < thresh_residual) 
        delta = 0.f;
    
    return delta;
}

bool AdaptiveLFSamplerRendererTask::accepted( LFSamplePixel* sample, LFSamplingChain *chain, SamplingData* sD )
{		  
    //Check bounds
    if( sample->x   < 0 || sample->x   >= sD->film->xResolution ||
	sample->y   < 0 || sample->y   >= sD->film->yResolution ||
	sample->ax  < 0 || sample->ax  >= sD->film->resAngleX ||
	sample->ay  < 0 || sample->ay  >= sD->film->resAngleY ) 
	return false;
    
    //#### Duplicate check
    
    //Check if current sample ...
    if( sample->x == chain->sample.x && 
        sample->y == chain->sample.y && 
        sample->ax == chain->sample.ax && 
        sample->ay == chain->sample.ay)
	return false;
    
    
    //... and check online for further duplicates (optionally)
    if( duplicateCheck == 1 )
    {
	//Full, online checked
	
	int rec_idx = sample->y  * ( sD->film->xResolution * sD->film->resAngleY * sD->film->resAngleX ) + 
		  sample->x  * ( sD->film->resAngleY   * sD->film->resAngleX ) + 
		  sample->ay * ( sD->film->resAngleX ) + 
		  sample->ax;
		  
	//Only check for still possible ones !
	if( recordedSamples[rec_idx] ) //Already accepted before 
	    return false;
    }
    else if( duplicateChainCheck > 0 )
    {
	//Walk through chain if defined
	
	for( int last_sample = std::min((int)chain->recorded->size() - 1, duplicateChainCheck); 
	      last_sample >= 0; 
	      last_sample--)
	{   
	    //Check for last recorded samples ...
	    if( sample->x == chain->recorded->at(last_sample).x && 
		sample->y == chain->recorded->at(last_sample).y && 
		sample->ax == chain->recorded->at(last_sample).ax && 
		sample->ay == chain->recorded->at(last_sample).ay)
		return false;
	}
    }
    else
    {
	//Postprocess or no check at all
    }
    
    //Debug region
    /*
    if( sample->x < dStartX || sample->x >= dEndX ||
	sample->y < dStartY || sample->y >= dEndY )
	return false;
    */
    
    //Only true if all tests failed (not exited before)
    return true;
}

bool AdaptiveLFSamplerRendererTask::mutate( LFSamplePixel* in, LFSamplePixel* out,
					    LFSamplingChain *chain, SamplingData* sD,
					    float mutation_sigma_spatial,
					    float mutation_sigma_angular )
{

    //Maximum number of tests
    int num_test = 0;
    bool accept;
    
    //Repeat until accepted
    do{
      
	//Symmetric normal around last sample
	float x, y, ax, ay;
	Normal2DUnique( x, y, *(sD->rng) );
	Normal2DUnique( ax, ay, *(sD->rng) );
	
	//Normal with sigma mutation_sigma_spatial/mutation_sigma_angular and mean in
	x =  x  * mutation_sigma_spatial + (float)in->x;
	y =  y  * mutation_sigma_spatial + (float)in->y;
	
	ax = ax * mutation_sigma_angular + (float)in->ax;
	ay = ay * mutation_sigma_angular + (float)in->ay;
	
	//Discrete version
	out->x = Round2Int(x);
	out->y = Round2Int(y);
	out->ax = Round2Int(ax);
	out->ay = Round2Int(ay);
    
	//Compute acceptance
	accept = accepted( out, chain, sD );
	
	//Increase number of tests
	num_test += 1;
	
    }while( !accept && num_test <= 20 );
    
    return accept;
}

bool AdaptiveLFSamplerRendererTask::record( LFSamplingChain *chain, LFSamplePixel* sample, SamplingData* sD )
{
    //Sanity check
    if( sample->x   < 0 || sample->x   >= sD->film->xResolution ||
	sample->y   < 0 || sample->y   >= sD->film->yResolution ||
	sample->ax  < 0 || sample->ax  >= sD->film->resAngleX ||
	sample->ay  < 0 || sample->ay  >= sD->film->resAngleY ) 
	return false;
    
    //New pixel
    LFSamplePixel newPixel( *sample );
    
    //Add and check for other duplicates online (optionally)
    bool accept = false;
    if( duplicateCheck == 1 )
    {
	//Full, online check
	int rec_idx = sample->y  * ( sD->film->xResolution * sD->film->resAngleY * sD->film->resAngleX ) + 
		      sample->x  * ( sD->film->resAngleY   * sD->film->resAngleX ) + 
		      sample->ay * ( sD->film->resAngleX ) + 
		      sample->ax;
	
	//Record in chain with unique addition !
	int origVal = AtomicCompareAndSwap(&(recordedSamples[rec_idx]), 1, 0);
	
	//Try to add
	accept = (origVal == 0) && chain->record(newPixel);
	if( accept )
	    totalRecordedSamples++;
    }
    else
    {
	//Postprocess or no check at all
	
	chain->record(newPixel);
	accept = true;
	totalRecordedSamples++;
    }
    
    return accept;
}


void AdaptiveLFSamplerRendererTask::EvaluateLightfield( LFSamplePixel* sample, SamplingData* sD ) 
{

    //Do simple test lookup
    if( lightfield1DtestData != NULL )
    {
      	int rec_idx = sample->x  * ( sD->film->yResolution * sD->film->resAngleY * sD->film->resAngleX ) + 
		      sample->y  * ( sD->film->resAngleY   * sD->film->resAngleX ) + 
		      sample->ax * ( sD->film->resAngleY ) + 
		      sample->ay;
	 
	sample->rgb[0] = lightfield1DtestData[rec_idx * 3 + 0];
	sample->rgb[1] = lightfield1DtestData[rec_idx * 3 + 1];
	sample->rgb[2] = lightfield1DtestData[rec_idx * 3 + 2];
	
	return;
    }

    //Now get the samples 
    int sampleCount = sD->sampler->GetLFPixelSamples( sample->x,
						      sample->y,
						      sample->ax,
						      sample->ay,
						      sD->samples, *(sD->rng) );
    
    // Generate camera rays and compute radiance along rays
    for (int i = 0; i < sampleCount; ++i)
    {
      
	// Find camera ray for _sample[i]_
	PBRT_STARTED_GENERATING_CAMERA_RAY(&(sD->samples[i]));
	
	float rayWeight = camera->GenerateRayDifferential(sD->samples[i], &(sD->rays[i]));
	sD->rays[i].ScaleDifferentials(1.f / sqrtf(sD->sampler->samplesPerPixel));
	
	PBRT_FINISHED_GENERATING_CAMERA_RAY(&(sD->samples[i]), &(sD->rays[i]), rayWeight);

	// Evaluate radiance along camera ray
	PBRT_STARTED_CAMERA_RAY_INTEGRATION(&(sD->rays[i]), &(sD->samples[i]));
	
	if (visualizeObjectIds) {
	    if (rayWeight > 0.f && scene->Intersect(sD->rays[i], &(sD->isects[i]) ) ) {
		// random shading based on shape id...
		uint32_t ids[2] = { sD->isects[i].shapeId, sD->isects[i].primitiveId };
		uint32_t h = hash((char *)ids, sizeof(ids));
		float rgb[3] = { (h & 0xff), (h >> 8) & 0xff, (h >> 16) & 0xff };
		sD->Ls[i] = Spectrum::FromRGB(rgb);
		sD->Ls[i] /= 255.f;
	    }
	    else
		sD->Ls[i] = 0.f;
	}
	else {
	    if (rayWeight > 0.f)
		sD->Ls[i] = rayWeight * renderer->Li(scene, sD->rays[i], &(sD->samples[i]), *(sD->rng),
						    *(sD->arena), &(sD->isects[i]), &(sD->Ts[i]));
	    else {
		sD->Ls[i] = 0.f;
		sD->Ts[i] = 1.f;
	    }

	    // Issue warning if unexpected radiance value returned
	    if (sD->Ls[i].HasNaNs()) {
		Error("Not-a-number radiance value returned "
		      "for image sample.  Setting to black.");
		sD->Ls[i] = Spectrum(0.f);
	    }
	    else if (sD->Ls[i].y() < -1e-5) {
		Error("Negative luminance value, %f, returned"
		      "for image sample.  Setting to black.", sD->Ls[i].y());
		sD->Ls[i] = Spectrum(0.f);
	    }
	    else if (isinf(sD->Ls[i].y())) {
		Error("Infinite luminance value returned"
		      "for image sample.  Setting to black.");
		sD->Ls[i] = Spectrum(0.f);
	    }
	}
	
	PBRT_FINISHED_CAMERA_RAY_INTEGRATION(&(sD->rays[i]), &(sD->samples[i]), &(sD->Ls[i]));
    }
    
    //Integrate the finite area pixel (filter) and compute RGB
    float Lxyz[3];
    Lxyz[0] = 0.f;
    Lxyz[1] = 0.f;
    Lxyz[2] = 0.f;
    float weightSum = 0.f;
    
    for (int i = 0; i < sampleCount; ++i)
    {
	//Start integrating sample
	
	//Check range and add up
	if (sD->Ls[i].HasNaNs()) {
	    Warning("Lightfield sampling renderer ignoring spectrum with NaN values");
	    PBRT_FINISHED_ADDING_IMAGE_SAMPLE();
	    continue;
	}

	// Loop over filter support and add sample to pixel arrays
	float xyz[3];
	sD->Ls[i].ToXYZ(xyz);

	//Add
	Lxyz[0] += xyz[0];
	Lxyz[1] += xyz[1];
	Lxyz[2] += xyz[2];
	weightSum += 1.f;
	
	//Finished integrating samples
    }

    // Convert XYZ color to new samples's RGB
    XYZToRGB(Lxyz, sample->rgb);

    // Normalize pixel with weight sum
    if (weightSum != 0.f) 
    {
	float invWt = 1.f / weightSum;
	sample->rgb[0] = max(0.f, sample->rgb[0] * invWt);
	sample->rgb[1] = max(0.f, sample->rgb[1] * invWt);
	sample->rgb[2] = max(0.f, sample->rgb[2] * invWt);
    }
    
    // Free _MemoryArena_ memory from computing single lightfield sample value
    sD->arena->FreeAll();
}

void AdaptiveLFSamplerRendererTask::ComputeSubChains(int num, int count, 
						     int *newCStart, int *newCEnd) const
{
    // Compute chain range for sub-range
    float tx0 = float(num) / float(count), tx1 = float(num+1) / float(count);
    *newCStart = Floor2Int(Lerp(tx0, 0, chains->size()));
    *newCEnd   = Floor2Int(Lerp(tx1, 0, chains->size()));
}

// SamplerRenderer Method Definitions
AdaptiveLFSamplerRenderer::AdaptiveLFSamplerRenderer(Sampler *s, Camera *c,
                                 SurfaceIntegrator *si, VolumeIntegrator *vi,
                                 bool visIds,
				 LightfieldReconstruction * lfRec,
				 int duplicateChk,
				 int numChains, int chLength, int maxAcceptedSamplesPerChains,
				 float m_sigma_spatial, float m_sigma_angular, float t_residual,  
				 int nChDuplicateChecks, bool rplSamples,
				 bool smplTime, bool smplFull ) 
{
    sampler = s;
    camera = c;
    surfaceIntegrator = si;
    volumeIntegrator = vi;
    visualizeObjectIds = visIds;
    
    //Adaptive sampling stuff
    lightfieldRec = lfRec;
    chains = new std::vector<LFSamplingChain>(numChains, LFSamplingChain(maxAcceptedSamplesPerChains));
    chainLength = chLength;
    mutation_sigma_spatial = m_sigma_spatial;
    mutation_sigma_angular = m_sigma_angular; 
    thresh_residual = t_residual;
    numChainDuplicateChecks = nChDuplicateChecks;
    
    //Sample time
    sampleTime = smplTime;
    
    //Sample full
    sampleFull = smplFull;
    
    //Initialize sampleset which holds all samples
    sampleSet = new SampleSet();
    newSampleBegin = 0;

    //Setup (optional) duplicate checking structures
    duplicateCheck = duplicateChk;
    replaceSamples = rplSamples;
    
    recordedSamples = NULL;
    //recordedKDtree = NULL;
    
    //Reset initialization
    initializedRenderStructures = false;
    
    if( duplicateCheck <= 0 || duplicateCheck > 2 )
    {
	//No duplicate check
    }
    else if( duplicateCheck == 1 )
    {
	//Full, online, but memory expensive record for every sample
	
	//Get film (allways lightfield)
	LightfieldFilm *film = (LightfieldFilm*)camera->film;
	recordedSamples = new int[film->xResolution * film->yResolution *
				  film->resAngleX * film->resAngleY];
				  
	for(int i=0; i < film->xResolution * film->yResolution * film->resAngleX * film->resAngleY; i++)
	    recordedSamples[i] = false;
	
    }else if( duplicateCheck == 2 )
    {
      	recordedSamplesJArray = (Pvoid_t) NULL;
	
	//Test the hash function
	/*
	//Search structures
	Word_t Index;                 // index (or key)
	int    Rc_int;                // boolean values returned (0 or 1)
	Word_t*  PValue;                   // pointer to array element value

	Index = 1000;
	JLI(PValue, recordedSamplesJArray, Index);
	*PValue = (Word_t)1001;
	
	Index = 1500;
	JLI(PValue, recordedSamplesJArray, Index);
	*PValue = (Word_t)1501;
	
	Index = 1300;
	JLI(PValue, recordedSamplesJArray, Index);
	*PValue = (Word_t)1301;

	Index = 1399;
	//Try to get pointer
	JLG(PValue, recordedSamplesJArray, Index);
	if( PValue == NULL )
	  printf("No value found for val[%d]\n", (int)Index);
	else
	  printf("val[%d] = %d\n", (int)Index, (int)(*PValue) );
	
	Index = 1500;
	//Try to get pointer
	JLG(PValue, recordedSamplesJArray, Index);
	if( PValue == NULL )
	  printf("No value found for val[%d]\n", (int)Index);
	else
	  printf("val[%d] = %d\n", (int)Index, (int)(*PValue) );
	  
	Index = 1000;
	//Try to get pointer
	JLG(PValue, recordedSamplesJArray, Index);
	if( PValue == NULL )
	  printf("No value found for val[%d]\n", (int)Index);
	else
	  printf("val[%d] = %d\n", (int)Index, (int)(*PValue) );
	
	//Do a change
	Index = 1000;
	//JLI(PValue, recordedSamplesJArray, Index);
	JLG(PValue, recordedSamplesJArray, Index);
	*PValue = (Word_t)1111;
	
	Index = 1000;
	//Try to get pointer
	JLG(PValue, recordedSamplesJArray, Index);
	if( PValue == NULL )
	  printf("No value found for val[%d]\n", (int)Index);
	else
	  printf("val[%d] = %d\n", (int)Index, (int)(*PValue) );
	*/
    }
    
    //Initialize sampling info
    totalRecordedSamples = 0;
    totalTestedSamples = 0;
    samplingTimings.clear();
    
    //Test data 1D
    lightfield1DtestData = NULL;
}


AdaptiveLFSamplerRenderer::~AdaptiveLFSamplerRenderer() {
    delete sampler;
    sampler = NULL;
    
    delete camera;
    camera = NULL;
    
    delete surfaceIntegrator;
    surfaceIntegrator = NULL;
    
    delete volumeIntegrator;
    volumeIntegrator = NULL;
    
    delete chains;
    chains = NULL;
    
    delete sampleSet;
    sampleSet = NULL;
    
    //Duplicate structures
    if( duplicateCheck <= 0 || duplicateCheck > 2 )
    {
	//Nothing
    }
    else if( duplicateCheck == 1 )
    {
	delete[] recordedSamples;
	recordedSamples = NULL;
    }
    else if( duplicateCheck == 2 )
    {
	//delete recordedKDtree;

	//Simple bit array
	//int Rc_int; 
	//J1FA(Rc_int, recordedSamplesJ1Array);
	
	//Free whole array
	Word_t Rc_word; 
	JLFA(Rc_word, recordedSamplesJArray);
    }
    
    //Test data 1D
    if( lightfield1DtestData != NULL )
	delete[] lightfield1DtestData;
    lightfield1DtestData = NULL;
}

void AdaptiveLFSamplerRenderer::set1DtestData(float *data )
{
    //Set new data
    LightfieldFilm *film = (LightfieldFilm*)camera->film; //Get film (always lightfield)
    if( lightfield1DtestData == NULL )
	lightfield1DtestData = new float[3 * film->xResolution * film->yResolution * 
					film->resAngleX * film->resAngleY];
  
    //Set data in ay, ax, y, x format
    for(int i=0; i <  3 * film->resAngleY * film->resAngleX * film->yResolution * film->xResolution ; i++)
	lightfield1DtestData[i] = data[i];
}

void AdaptiveLFSamplerRenderer::Render(const Scene *scene)
{
    //Fall back to t = 0
    this->Render(scene, 0.f);
}

void AdaptiveLFSamplerRenderer::Render(const Scene *scene, float time)
{
    // Preprocessing
    PBRT_FINISHED_PARSING();
    
    // Allow integrators to do preprocessing for the scene
    if( !initializedRenderStructures )
    {
	PBRT_STARTED_PREPROCESSING();
	surfaceIntegrator->Preprocess(scene, camera, this);
	volumeIntegrator->Preprocess(scene, camera, this);
	PBRT_FINISHED_PREPROCESSING();
	
	//Structures now initialized
	initializedRenderStructures = true;
    }
    
    PBRT_STARTED_RENDERING();
    
    //Set new sampler time 
    float offset = camera->shutterClose - camera->shutterOpen;
    sampler->shutterOpen = time - offset;
    sampler->shutterClose = time + offset;
    
    //Reset online duplicates if samples are replaced !
    if( replaceSamples )
    {
	if( duplicateCheck == 1 )
	{
	    //Get film (always lightfield)
	    LightfieldFilm *film = (LightfieldFilm*)camera->film;
	    
	    for(int i=0; i < film->xResolution * film->yResolution * film->resAngleX * film->resAngleY; i++)
		recordedSamples[i] = false;
	}
    }
    
    //Reset all chains for safety
    for( int cIdx = 0; cIdx < chains->size(); cIdx++ )
    {
	//Get current chain
	LFSamplingChain* currChain = &( chains->at(cIdx) );
	currChain->reset();
    }
    
    //Fully clear sample space if full renderer
    int nFullTasks = 0;
    if( sampleFull )
    {
	//Empty old samples
	sampleSet->pts->clear();
	newSampleBegin = 0;
	
	//Full sampling tasks
	LightfieldFilm *film = (LightfieldFilm*)camera->film;
	int numLFPixels = film->xResolution * film->yResolution;
	nFullTasks = max( 32 * NumSystemCores(), Round2Int( float(numLFPixels) / float(16*16) ) ); //16^2 chains per core
	nFullTasks = RoundUpPow2(nFullTasks);
	
	//Sanity check
	if( chains->size() < nFullTasks )
	{
	    //Error
	    printf("ERROR: Not enought sampling chains !!! EXITING !!!\n");
	    PBRT_FINISHED_RENDERING();
	    return; 
	}
    }
    
    // Allocate and initialize camera _sample_
    Sample *sample = new Sample(sampler, surfaceIntegrator,
                                volumeIntegrator, scene);

    // Compute number of _SamplerRendererTask_s to create for rendering
    int nTasks = max( 32 * NumSystemCores(), Round2Int( float(chains->size()) / float(16*16) ) ); //16^2 chains per core
    nTasks = RoundUpPow2(nTasks);
    
    //Set to full tasks if full rendering
    if( sampleFull )
      nTasks = nFullTasks;
    
    ProgressReporter reporter(nTasks, "Rendering");
    vector<Task *> renderTasks;
    for (int i = 0; i < nTasks; ++i)
        renderTasks.push_back(new AdaptiveLFSamplerRendererTask(scene, this, camera,
                                                      reporter, sampler, sample, 
                                                      visualizeObjectIds, 
						      lightfieldRec,
						      chains, chainLength, 
						      mutation_sigma_spatial, mutation_sigma_angular, thresh_residual,
						      duplicateCheck, numChainDuplicateChecks,
						      recordedSamples, lightfield1DtestData,
						      sampleTime, sampleFull, sampleSet,
                                                      nTasks-1-i, nTasks));
    //Timing
    Timer timer;
    samplingTimings.clear();
	
    //Run tasks
    timer.Start();
    
    EnqueueTasks(renderTasks);
    WaitForAllTasks();
    
    timer.Stop();
    printf("Rendering %3.3f\n", timer.Time() );
    samplingTimings.push_back( timer.Time() );
    
    //Fetch all the numbers of tested/recorded samples
    totalRecordedSamples = 0;
    totalTestedSamples = 0;
    for (uint32_t i = 0; i < renderTasks.size(); ++i)
    {
        totalRecordedSamples += ((AdaptiveLFSamplerRendererTask*)renderTasks[i])->totalRecordedSamples;
	totalTestedSamples += ((AdaptiveLFSamplerRendererTask*)renderTasks[i])->totalTestedSamples;
    }
    
    //Reset new recorded sample (now it has been done a new sequence of records !)
    newSampleBegin = sampleSet->pts->size();
    
    //Fetch and duplicate
    timer.Reset();
    timer.Start();
    
    //Now fetch all the samples from chains
    for( int cIdx = 0; cIdx < chains->size(); cIdx++ )
    {
	//Get current chain
	LFSamplingChain* currChain = &( chains->at(cIdx) );
	
	//Iterate over chain and add current sample
	for( int i = 0; i < currChain->recorded->size(); i++ )
	    sampleSet->pts->push_back( currChain->recorded->at(i) );
	
	//Clear chain
	currChain->reset();
    }
    
    //Skip duplicates for full renderer
    if( ! sampleFull )
    {
	//Remove the duplicates for the newest added samples
	removeDuplicates( );
    }else
    {	
	//Douplicate timings
	samplingTimings.push_back(0.f);
	samplingTimings.push_back(0.f);
	samplingTimings.push_back(0.f);
    }
    
    timer.Stop();
    printf("Duplicates %3.3f\n", timer.Time() );
    samplingTimings.push_back( timer.Time() );
    
    //Cleanup tasks
    for (uint32_t i = 0; i < renderTasks.size(); ++i)
        delete renderTasks[i];
    
    reporter.Done();
    PBRT_FINISHED_RENDERING();
    // Clean up after rendering and store final image
    delete sample;
}

void AdaptiveLFSamplerRenderer::resetChains()
{
    //Reset all chains for safety
    for( int cIdx = 0; cIdx < chains->size(); cIdx++ )
    {
	//Get current chain
	LFSamplingChain* currChain = &( chains->at(cIdx) );
	currChain->reset();
	
	//Restart from next position
	currChain->initialized = false;
    }
    
    //Empty old samples
    sampleSet->pts->clear();
    newSampleBegin = 0;
    
    //Clear duplicates
    if( duplicateCheck <= 0 || duplicateCheck > 2 )
    {
	//Nothing
    }
    else if( duplicateCheck == 1 )
    {
        //Get film (always lightfield)
	LightfieldFilm *film = (LightfieldFilm*)camera->film;
	
	for(int i=0; i < film->xResolution * film->yResolution * film->resAngleX * film->resAngleY; i++)
	    recordedSamples[i] = false;
    }
    else if( duplicateCheck == 2 )
    {
	//Free whole array
	//int Rc_int; 
	//J1FA(Rc_int, recordedSamplesJ1Array);

	//Free whole array
	Word_t Rc_word; 
	JLFA(Rc_word, recordedSamplesJArray);
	
	//Generate new array
	recordedSamplesJArray = (Pvoid_t) NULL;
    }

}

void AdaptiveLFSamplerRenderer::removeDuplicates()
{
    //Do a duplicate check on the samples
    if( duplicateCheck <= 0 || duplicateCheck > 2 )
    {
	//No duplicate check at all
    }
    else if( duplicateCheck == 1 )
    {
	//Duplicates already checked in the rendering tasks
	
    }else if( duplicateCheck == 2 )
    {
	//#### Radiussearch with kd-tree
	/*
	//Create boolean addition flags for new samples
	std::vector<bool> addedNewSample(sampleSet->pts->size() - newSampleBegin);
	for(int i = 0; i < addedNewSample.size(); i++)
	    addedNewSample.at(i) = false;
	
	//Set dataset terminals
	sampleSet->setup_terminal_cond( newSampleBegin, &addedNewSample  );

	//Build new tree index
	Timer timer;
	timer.Start();
	recordedKDtree->buildIndex();
	timer.Stop();
	printf("BuildIndex %3.3f\n", timer.Time() );
	samplingTimings.push_back( timer.Time() );
	
	//Search structures
	int queryPoint[4];
	const int search_radius = static_cast<int>(1);
	std::vector<std::pair<size_t,int> >  ret_matches;
	nanoflann::SearchParams params;
	//params.sorted = false;
	size_t nMatches;
	
	timer.Reset();
	timer.Start();
	
	//Iterate over new samples and add potentially
	int numAcceptedSamples = 0;
	for( int j = newSampleBegin; j < sampleSet->pts->size(); j++)
	{
	    //Query
	    queryPoint[0] = sampleSet->pts->at(j).x;
	    queryPoint[1] = sampleSet->pts->at(j).y;
	    queryPoint[2] = sampleSet->pts->at(j).ax;
	    queryPoint[3] = sampleSet->pts->at(j).ay;

	    //Do radius search
	    bool early_exit;
	    nMatches = recordedKDtree->radiusSearch(queryPoint,search_radius, ret_matches, early_exit, params);
	   
	    //Add potentially
	    bool addSample = true;
	    if( early_exit )
	      addSample = false;
	    
	    //Add potentially
	    addedNewSample.at(j - newSampleBegin) = addSample;
	    
	    //and count if added
	    if( addSample )
	      numAcceptedSamples++;  
	}
	
	timer.Stop();
	printf("Search %3.3f\n", timer.Time() );
	samplingTimings.push_back( timer.Time() );
	
	timer.Reset();
	timer.Start();
	
	//Now copy to temporary array and add the reduced array
	std::vector<LFSamplePixel> * tmpReduced = new std::vector<LFSamplePixel>(numAcceptedSamples);
	
	int offset = 0;
	for( int j = newSampleBegin; j < sampleSet->pts->size(); j++)
	{
	    if( addedNewSample.at(j - newSampleBegin) )
	    {
		tmpReduced->at( offset ) = sampleSet->pts->at(j);
		offset += 1;
	    }
	}
	
	//Debug: get size
	//int tmpsize = tmpReduced->size();
	//int smplsize =  sampleSet->pts->size();
	
	//Overwrite the reduced samples
	offset = newSampleBegin;
	for( int i = 0; i < tmpReduced->size(); i++)
	{
	    sampleSet->pts->at(offset) = tmpReduced->at( i );
	    offset += 1;
	}
	
	//Remove the rest
	sampleSet->pts->erase( sampleSet->pts->begin() + offset, sampleSet->pts->end() );
	
	timer.Stop();
	printf("Reduce and copy %3.3f\n", timer.Time() );
	samplingTimings.push_back( timer.Time() );
	    
	//Remove tmp vector
	delete tmpReduced;
	*/
	    
	//#### Sparse array hashing
	//Create boolean addition flags for new samples
	std::vector<bool> addedNewSample(sampleSet->pts->size() - newSampleBegin,  false);

	//Search structures
	Word_t Index;                 // index (or key)
	int    Rc_int;                // boolean values returned (0 or 1)
	Word_t*  PValue;                   // pointer to array element value

	//Setup timer
	Timer timer;
	timer.Start();
	timer.Stop();
	samplingTimings.push_back( timer.Time() ); //Setup time
	
	timer.Reset();
	timer.Start();
	
	//Iterate over new samples and add potentially
	LightfieldFilm *film = (LightfieldFilm*)camera->film;
	int numAcceptedSamples = 0;
	int currNewSamplePos = newSampleBegin;
	for( int j = newSampleBegin; j < sampleSet->pts->size(); j++)
	{
	    //Compute query index
	    LFSamplePixel* currPixel = &(sampleSet->pts->at(j));
	    Index = (Word_t)(currPixel->y * (film->xResolution * film->resAngleX * film->resAngleY) +
		    currPixel->x * (film->resAngleX * film->resAngleY) +
		    currPixel->ay * (film->resAngleX) +
		    currPixel->ax);

	    //Try to get pointer
	    JLG(PValue, recordedSamplesJArray, Index);
	
	    //Update addedNewSample, hash and count if it should be added
	    if( PValue == NULL )
	    {
		addedNewSample.at(j - newSampleBegin) = true;
		JLI(PValue, recordedSamplesJArray, Index);
		*PValue = (Word_t)currNewSamplePos;
		
		currNewSamplePos++;
		numAcceptedSamples++; 
	    }
	    else if( replaceSamples ) 
	    {
		//Already added and replaced
		Word_t arrayIdx = *PValue;
		
		//Replace only sample vals that are in old array
		if( arrayIdx < newSampleBegin)
		{
		    sampleSet->pts->at(arrayIdx).rgb[0] = currPixel->rgb[0];
		    sampleSet->pts->at(arrayIdx).rgb[1] = currPixel->rgb[1];
		    sampleSet->pts->at(arrayIdx).rgb[2] = currPixel->rgb[2];
		}
	    }
	}
	
	timer.Stop();
	printf("Search %3.3f\n", timer.Time() );
	samplingTimings.push_back( timer.Time() );
	
	timer.Reset();
	timer.Start();
	
	//Now copy to temporary array and add the reduced array
	std::vector<LFSamplePixel> * tmpReduced = new std::vector<LFSamplePixel>(numAcceptedSamples);
	
	int offset = 0;
	for( int j = newSampleBegin; j < sampleSet->pts->size(); j++)
	{
	    if( addedNewSample.at(j - newSampleBegin) )
	    {
		tmpReduced->at( offset ) = sampleSet->pts->at(j);
		offset += 1;
	    }
	}
	
	//Debug: get size
	//int tmpsize = tmpReduced->size();
	//int smplsize =  sampleSet->pts->size();
	
	//Memory
	//J1MU(Rc_word, recordedSamplesJ1Array);       // how much memory was used?
	//printf("%lu Indexes used %lu bytes of memory\n", Rcount, Rc_word);
	
	//Overwrite the reduced samples
	offset = newSampleBegin;
	for( int i = 0; i < tmpReduced->size(); i++)
	{
	    sampleSet->pts->at(offset) = tmpReduced->at( i );
	    offset += 1;
	}
	
	//Remove the rest
	sampleSet->pts->erase( sampleSet->pts->begin() + offset, sampleSet->pts->end() );
	
	timer.Stop();
	printf("Reduce and copy %3.3f\n", timer.Time() );
	samplingTimings.push_back( timer.Time() );
	    
	//Remove tmp vector
	delete tmpReduced;

    }
    
   
}

void AdaptiveLFSamplerRenderer::updateSamplingParams(int chLength,
						     float m_sigma_spatial, 
						     float m_sigma_angular, 
						     float t_residual,
						     int nChDuplicateChecks,
						     bool rplSamples )
{
    //Update chainlength
    chainLength = chLength;
    mutation_sigma_spatial = m_sigma_spatial;
    mutation_sigma_angular = m_sigma_angular;
    thresh_residual = t_residual;
    numChainDuplicateChecks = nChDuplicateChecks;
    replaceSamples = rplSamples;
}

Spectrum AdaptiveLFSamplerRenderer::Li(const Scene *scene,
        const RayDifferential &ray, const Sample *sample, RNG &rng,
        MemoryArena &arena, Intersection *isect, Spectrum *T) const {
    Assert(ray.time == sample->time);
    Assert(!ray.HasNaNs());
    // Allocate local variables for _isect_ and _T_ if needed
    Spectrum localT;
    if (!T) T = &localT;
    Intersection localIsect;
    if (!isect) isect = &localIsect;
    Spectrum Li = 0.f;
    if (scene->Intersect(ray, isect))
        Li = surfaceIntegrator->Li(scene, this, ray, *isect, sample,
                                   rng, arena);
    else {
        // Handle ray that doesn't intersect any geometry
        for (uint32_t i = 0; i < scene->lights.size(); ++i)
           Li += scene->lights[i]->Le(ray);
    }
    Spectrum Lvi = volumeIntegrator->Li(scene, this, ray, sample, rng,
                                        T, arena);
    return *T * Li + Lvi;
}

void AdaptiveLFSamplerRenderer::updateSize()
{
    //Change duplicate check structures
    if( duplicateCheck == 1 )
    {
	//Online duplicate check
	
	//Update recorded samples
	delete[] recordedSamples;
	
	//Get film (allways lightfield)
	LightfieldFilm *film = (LightfieldFilm*)camera->film;
	recordedSamples = new int[film->xResolution * film->yResolution *
				  film->resAngleX * film->resAngleY];
				  
	for(int i=0; i < film->xResolution * film->yResolution * film->resAngleX * film->resAngleY; i++)
	    recordedSamples[i] = false;
    }
    
    //Reset chain positions
    for( int cIdx = 0; cIdx < chains->size(); cIdx++ )
    {
	//Get current chain
	LFSamplingChain* currChain = &( chains->at(cIdx) );
	currChain->reset();
	currChain->initialized = false;
    }
}


Spectrum AdaptiveLFSamplerRenderer::Transmittance(const Scene *scene,
        const RayDifferential &ray, const Sample *sample, RNG &rng,
        MemoryArena &arena) const {
    return volumeIntegrator->Transmittance(scene, this, ray, sample,
                                           rng, arena);
}


