
/*
    Low discrepancy filtering for lightfields.
 */


// samplers/lowdiscrepancy.cpp*
#include "stdafx.h"
#include "samplers/lowdiscrepancy_lightfield.h"
#include "cameras/lightfield.h"
#include "montecarlo.h"

// LDSamplerLightfield Method Definitions
LDSamplerLightfield::LDSamplerLightfield(int xstart, int xend, int ystart, int yend, int xAend, int yAend,
					 int ps,
					float sopen, float sclose)
    : Sampler(xstart, xend, ystart, yend, RoundUpPow2(ps), sopen, sclose) 
    {
          
    //Ranges
    xAnglePixelEnd = xAend;
    yAnglePixelEnd = yAend;
    
    //Position
    xPos = xPixelStart;
    yPos = yPixelStart;
    xAPos = 0;
    yAPos = 0;
    
    if (!IsPowerOf2(ps)) {
        Warning("Pixel samples being rounded up to power of 2");
        nPixelSamples = RoundUpPow2(ps);
    } else
        nPixelSamples = ps;
    sampleBuf = NULL;
}


LDSamplerLightfield::~LDSamplerLightfield() {
    delete[] sampleBuf;
}

void LDSamplerLightfield::updateSize( int xres, int yres,
			    int rAX, int rAY)
{
    //Ranges
    xPixelEnd = xres;
    yPixelEnd = yres;
  
    xAnglePixelEnd = rAX;
    yAnglePixelEnd = rAY;
}


Sampler *LDSamplerLightfield::GetSubSampler(int num, int count) {
    int x0, x1, y0, y1;
    ComputeSubWindow(num, count, &x0, &x1, &y0, &y1);
    if (x0 == x1 || y0 == y1) return NULL;
    return new LDSamplerLightfield(x0, x1, y0, y1, xAnglePixelEnd, yAnglePixelEnd,
			 nPixelSamples, shutterOpen, shutterClose);
}


int LDSamplerLightfield::GetMoreSamples(Sample *samples, RNG &rng) {
    if (yPos == yPixelEnd) return 0;
    if (sampleBuf == NULL)
        sampleBuf = new float[LDPixelSampleFloatsNeeded(samples,
                                                        nPixelSamples)];
    LDPixelAngleSample(xPos, yPos, xAPos, yAPos,
		      shutterOpen, shutterClose,
		      nPixelSamples, samples, sampleBuf, rng);
    
    //Increment
    if (++xAPos == xAnglePixelEnd){
        xAPos = 0;
        ++yAPos;
    }
    
    if (yAPos == yAnglePixelEnd){
        yAPos = 0;
        ++xPos;
    }
    
    if (xPos == xPixelEnd) {
        xPos = xPixelStart;
        ++yPos;
    }
    
    return nPixelSamples;
}

int LDSamplerLightfield::GetLFPixelSamples( int x, int y, int ax, int ay, 
					  Sample *samples, RNG &rng) 
{
    //Range sanity
    if (x  < xPixelStart || x  >= xPixelEnd ||
        y  < yPixelStart || y  >= yPixelEnd ||
        ax < 0 || ax >= xAnglePixelEnd ||
        ay < 0 || ay >= yAnglePixelEnd )
    {
	Warning("Sample in LD lightfield sampler out of range\n");
	return 0;
    }
    
    //Do the sampling
    if (sampleBuf == NULL)
        sampleBuf = new float[LDPixelSampleFloatsNeeded(samples,
                                                        nPixelSamples)];
    LDPixelAngleSample(x, y, ax, ay,
		      shutterOpen, shutterClose, 
		      nPixelSamples, samples, sampleBuf, rng);
    
    //Return the number of samples
    return nPixelSamples;
}

LDSamplerLightfield *CreateLowDiscrepancySamplerLightfield(const ParamSet &params, const LightfieldFilm *film,
							  const Lightfield *lf) 
{
    // Initialize common sampler parameters
    int xend, yend, xAend, yAend;
    xend = film->xResolution;
    yend = film->yResolution;
    xAend = film->resAngleX;
    yAend = film->resAngleY;
    
    int nsamp = params.FindOneInt("pixelsamples", 4);
    if (PbrtOptions.quickRender) nsamp = 1;
    
    return new LDSamplerLightfield(0, xend, 0, yend, xAend, yAend, nsamp,
			 lf->shutterOpen, lf->shutterClose);
}


