% ***Image synthesis for compressive displays***
%
% see our paper: F. Heide, G. Wetzstein, R. Raskar, W. Heidrich.
% ''Adaptive Image Synthesis for Compressive Displays'', SIGGRAPH 2013.
%
% Layered3D naive optimization
%
% Copyright (C) 2013. Felix Heide
% Email: fheide@cs.ubc.ca


function [layersRec lightFieldRec sart_time] = ...
    computeAttenuationLayersFromLightField2D( lightField, lightFieldResolution, layerResolution, maxIters, minTransmission)                                              
                                            
    if nargin == 0
        error('computeAttenuationLayersFromLightField4D only works with all parameters!');    
    end    
                                            
	% declare matrix a global variable to save memory and time!
    global T;                 
                                                    
    % check if system is under-determined
    if prod(layerResolution) > prod(lightFieldResolution)
        error('Optimization is under-determined! This is not supported!');
    end    

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % set desired solution, constraints etc according to the current mode
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % no zeros in the light field - bad things happen in log-space!
    lightField(lightField<eps) = eps;
    
    
    % Sparse matrix for preconditioning.
    % This matrix is a required input for the solver;
    % preconditioning is not really being used in this example
    Jinfo = speye(prod(lightFieldResolution), prod(layerResolution));
        
    % light field to log light field
    lightField = log(lightField);       

    % here, the optimization needs to be constrained to transmission values
    % between minTransmission and 1, so the log values need to be < 0
    lb      = zeros([prod(layerResolution) 1]) + log(minTransmission);
    ub      = zeros([prod(layerResolution) 1]);

    % initial guess - use either zeros or the backprojected solution (zeros
    % seem to work much better)
    layersRec = zeros(layerResolution)-log(eps);

    % check if everything is finite
    if sum(~isfinite(lightField(:))) > 0
        error('(Log) Light Field is not Finite!');
    end        
        
                
    % SART
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % function handle
    Afun = @(x,flag) projectLayersMatrix( x, flag );    
    % run SART
    tic;
    layersRec = SART(Afun,lightField(:),lb,ub,layersRec(:),maxIters);  
    sart_time = toc;
    % compute reconstructed light field
    lightFieldRec = Afun(layersRec(:), 1); 
 
    
%      exitflag
%      output.message     

    % reshape result  
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    lightFieldRec   = reshape(lightFieldRec, lightFieldResolution);
    layersRec       = reshape(layersRec, layerResolution);  
    
    % exponentiate reconstruction
    lightFieldRec   = exp(lightFieldRec);   
    layersRec   = exp(layersRec); 
    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% use explicit matrix (sparse)

function result = projectLayersMatrix( Y, flag )    
    % declare T a global variable
    global T;

    % forward projection
    if flag > 0                  
        result = T*Y;                                       
    % back-projection
    elseif flag < 0 
          result = T'*Y;                                                                                                         
    % forward and then back-projection
    else
      result = T'*(T*Y);        
    end    
end
