% ***Image synthesis for compressive displays***
%
% see our paper: F. Heide, G. Wetzstein, R. Raskar, W. Heidrich.
% ''Adaptive Image Synthesis for Compressive Displays'', SIGGRAPH 2013.
%
% HR3D animation, sampling and optimization
%
% Copyright (C) 2013. Felix Heide
% Email: fheide@cs.ubc.ca

function [layersRec lightFieldRec B_sampled_map rec_time] = ...
    computeAttenuationLayersFromLightField2D_HR3D_anim( lightFieldResolution, ...
                                                         lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lfResolution, lightFieldOrigin,...
                                                         layerResolution, layerSize, layerOrigin, layerDistance, basisFunctionType, ...
                                                         timeSteps, layerSubset, ...
                                                         chainLength, mutation_sigma_spatial, ...
                                                         mutation_sigma_angular, thresh_residual, ...
														 numChainDuplicateChecks, ...
                                                         max_sampling_it, max_it, ...
                                                         time, layersRec_prev, old_fixed_layers, ...
                                                         scaling_factor, ...
                                                         pbrt_ptr, ...
                                                         verbose)                                              
        
    % check for parameters
    if nargin == 0
        error('computeAttenuationLayersFromLightField2D only works with all parameters!');    
    end         
                                                    
    % check if system is under-determined
    if prod(layerResolution) > prod(lightFieldResolution)
        error('Optimization is under-determined! This is not supported!');
    end
    
    % check if two layers
    if length(layerSubset) ~= 2
        error('Currently supports only two layers!');
    end        
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %CUDA
    use_cuda = false;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % set desired resolution, constraints etc according to the current mode
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % resolution
    layerColorSize = [layerResolution(1), layerResolution(2), timeSteps, 3];
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Stochastic observation minimization
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %Initial samples
    samples = [];
    samples_indices = [];
    samples_intersects = [];
    
    %Initial reconstruciton
    lRes= lightFieldResolution; %Tmp
    lrec = zeros(lRes);
    lrec = single(lrec); %% Use single here

    %Figure
    if  strcmp(verbose,'all')
        iter_fig = figure();
        iter_layer_fig = figure();
    end

    %Debug
    if strcmp(verbose, 'brief') || strcmp( verbose, 'all')
        fprintf('\nMinimizing with stochastic observation sampling\n')
    end
    
    %Only used for debug
    if  strcmp(verbose,'all')
        LF_rec = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);
        B_sampled = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);
    end
    B_sampled_map = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);
    
    %{
    %For test now load previous solution
    A_prev = load('Aprev.mat');
    A_prev = A_prev.A;
    
    Bt_prev = load('Bprev.mat');
    Bt_prev = Bt_prev.Bt;
    
    
    %Mimic shift
    A_prev = circshift(A_prev,10);
    Bt_prev = circshift(Bt_prev,10);
    %}
    
    %Get previous solution
    A_prev = reshape( layersRec_prev{1,1}, layerResolution(1) * layerResolution(2), timeSteps, 3 );
    Bt_prev = reshape( layersRec_prev{1,2}, layerResolution(1) * layerResolution(2), timeSteps, 3 );
    ofl = old_fixed_layers;
    
    %Initialize with previous solution
    A = A_prev;
    Bt = Bt_prev;

    %Do sampling iterations
    rec_time = 0;
    for j = 1:max_sampling_it

        %% Sampling step: Sample residual stochastically now
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if strcmp(verbose, 'brief') || strcmp( verbose, 'all')
            fprintf('\n\nSampling process %d.\n', j)
        end

        %% Compute last (dense) reconstruction
        for ch = 1:3
            
            %Compute reconstruction
            curr_rec = reconstructLF_full(A(:,:,ch), (Bt(:,:,ch))', ...
                                        lightFieldAnglesY, lightFieldAnglesX, ... 
                                        lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                        layerResolution, layerSize, layerOrigin, layerDistance, ...
                                        layerSubset(1), layerSubset(2) );
                                    
            curr_rec = curr_rec / scaling_factor;
                                    
            lrec(:,:,:,:,ch) = single( curr_rec );
            
        end

        %% Render data
        render_params = single([chainLength; mutation_sigma_spatial; ...
                                mutation_sigma_angular; thresh_residual; numChainDuplicateChecks]);
                            
        [curr_samples, curr_samples_fullindx] = mex_pbrt('sample', pbrt_ptr, shiftdim( single(lrec), 2) , render_params, single(time));

        curr_samples = double(curr_samples);
        curr_samples_fullindx = int32(curr_samples_fullindx);
        
        %% Generate sample reconstructions
        
        %Compute linear indices
        curr_samples_indx = curr_samples_fullindx(1,:) * ( lRes(1) * lRes(2) * lRes(3) ) + ... %x * (resY *  rAY * rAX)
                            curr_samples_fullindx(2,:) * ( lRes(1) * lRes(2) ) + ... %y * rAY * rAX
                            curr_samples_fullindx(3,:) *   lRes(1) + ...    %ax * rAY 
                            curr_samples_fullindx(4,:) +  1;   %ay + 1 (Matlab matrix offset)
        
        %Transform lightfield
        %{
        % no zeros in the light field - bad things happen in log-space!
        curr_samples( curr_samples < eps ) = eps;
        %}
        
        %HR3D transform
        curr_samples = curr_samples * scaling_factor; 
        
        %Find hit positions
        layerInt = precomputeLayerIntersectionsHR3D_sampled_dense( lightFieldAnglesY, lightFieldAnglesX, ... 
                                                    lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                                    layerResolution, layerSize, layerOrigin, layerDistance, double(curr_samples_fullindx), ...
                                                    layerSubset(1), layerSubset(2), ...
                                                    basisFunctionType, 1 );
        %Remove samples with bad basis             
        validsamples = (layerInt(3,:) > 0) & (layerInt(6,:) > 0);
        curr_samples_indx = curr_samples_indx(:, validsamples );
        curr_samples_fullindx = curr_samples_fullindx(:, validsamples );
        curr_samples = curr_samples(:, validsamples );
        layerInt = layerInt(:, validsamples);

        %Add samples
        samples         = [samples , curr_samples];
        samples_intersects = [samples_intersects , layerInt]; 
        
        %Unique test
        samples_indices = [samples_indices, curr_samples_indx];
        if( ~isequal( unique(samples_indices,'stable'), samples_indices)) 
            error('Samples are not unique !')
        end
        
        %Do gradient iterations
        if strcmp(verbose, 'brief') || strcmp( verbose, 'all')
            fprintf('\n\nGradient decent process %d.\n', j)
        end

        %% Gradient step: Do MATRIX FACTORIZATION iterations
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       
        tic;
        
        %Tolerances
        tol = 1e-3;
        
        %Save prev
        A_old = A;
        Bt_old = Bt;
                   
        %% Run matrix fact
        for ch = 1:3
            
            %Get current matrices
            A_ch = A(:,:,ch);
            B_ch = Bt(:,:,ch)';
            
            %Get prev matrices
            %A_prev_ch = A_prev(:,:,ch);
            %B_prev_ch = Bt_prev(:,:,ch)';
            
            %FULL EIGEN VERSION
            lambdaA_eigen = 0.005; %0.05
            lambdaB_eigen = 0.005; %0.05
            [A_ch, B_ch] = sparse_matrix_factorization_anim(samples, int32(ch), ...
                                                       A_ch, B_ch, int32(samples_intersects), ...
                                                       int32(old_fixed_layers), ... %A_prev_ch, B_prev_ch, int32(old_fixed_layers), ...
                                                       lambdaA_eigen, lambdaB_eigen, tol);
                                                       
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %{
            % small value thresh
            epsilon         = 1e-9;
            epsilonLower    = epsilon;
            
            %Regularizer weights
            lambdaA = 0.005; %0.05
            lambdaB = 0.005; %0.05
            
            %Preallocate and store samples
            L_s = sparse( samples_intersects(3,:), samples_intersects(6,:),samples(ch,:), size(A_ch,1), size(B_ch,2), size(samples_intersects,2) );
            AB = spalloc(size(A_ch,1), size(B_ch,2), size(samples_intersects,2));

            % start multiplicative update rules
            for iter=1:max_it

                %Regularizers 
                Areg = A_ch;% - A_prev(:,:,ch);% - circshift(A,-1);  
                Breg = B_ch;% - Bt_prev(:,:,ch)';% - circshift(A,-1);
                
                %Regularizer exception
                Areg(:,1:ofl) = 0;  
                Breg(1:ofl,:) = 0; 
                
                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                
                %Compute AB once
                %{
                if use_cuda
                    AB_samples = double( AmultB_cuda( single(A_ch), single(B_ch), int32(samples_intersects)) );
                else
                    AB_samples = AmultB(A_ch,B_ch,int32(samples_intersects));
                end
                %}
                
                %Store in sparse  matrix
                AB_samples = sum(A_ch( samples_intersects(3,:), :)' .* B_ch( :,samples_intersects(6,:)), 1);
                AB = sparse(samples_intersects(3,:), samples_intersects(6,:), AB_samples, size(A_ch,1), size(B_ch,2), size(samples_intersects,2));

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                % update A
                %{
                if use_cuda
                    upper = double( LmultBt_cuda( single(samples), int32(ch), single(B_ch), int32(samples_intersects)) );
                else
                    upper = LmultBt(samples, int32(ch), B_ch, int32(samples_intersects));
                end
                %}
                upper = L_s * B_ch';
                upper = max(upper + lambdaA*Areg, epsilon);

                %{
                if use_cuda
                    LBt = double( LmultBt_cuda( single(AB_samples), int32(1), single(B_ch), int32(samples_intersects)) ); %Here allways single ch
                else
                    LBt = LmultBt(AB_samples, int32(1), B_ch, int32(samples_intersects)); %Here allways single ch
                end
                %}
                LBt = AB * B_ch';
                
                A_ch(:,(ofl + 1):end) = A_ch(:,(ofl + 1):end) .* upper(:,(ofl + 1):end) ./ max(LBt(:,(ofl + 1):end), epsilonLower);    
                % clamp to feasible region
                A_ch(A_ch<0)=epsilon; A_ch(A_ch>1)=1; 

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                % update B
                %{
                if use_cuda
                    upper = double( AtmultL_cuda( single(samples), int32(ch), single(A_ch), int32(samples_intersects)) );
                else
                    upper = AtmultL(samples, int32(ch), A_ch, int32(samples_intersects));
                end
                %}
                upper = A_ch' * L_s;
                upper = max(upper + lambdaB*Breg, epsilon);

                %{
                if use_cuda
                    AtL = double( AtmultL_cuda( single(AB_samples), int32(1), single(A_ch), int32(samples_intersects)) ); %Here allways single ch
                else
                    AtL = AtmultL(AB_samples, int32(1), A_ch, int32(samples_intersects)); %Here allways single ch
                end
                %}
                AtL = A_ch' * AB;
                
                B_ch((ofl + 1):end,:) = B_ch((ofl + 1):end,:) .* upper((ofl + 1):end,:) ./ max(AtL((ofl + 1):end,:), epsilonLower);    
                % clamp to feasible region
                B_ch(B_ch<0)=epsilon; B_ch(B_ch>1)=1; 
 
                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                % plot time
                %if strcmp(verbose,'brief') || strcmp(verbose,'all')
                %    disp([num2str(iter) ' | ' num2str(max_it)]);
                %end

            end
            %}
                                                       
            %Write back
            A(:,:,ch) = A_ch;
            Bt(:,:,ch) = B_ch';

        end
        
        tt = toc;
        rec_time = rec_time + tt;
        disp(['Local Matrix factorization took ' num2str(tt) ' secs']);
        
        
        %% Debug
        if  strcmp(verbose,'all')
            %Debug layer
            %figure(), imagesc(reshape(exp(x), [layerResolution(3) layerResolution(2), 3] )), colormap gray, colorbar;

            %1D Generate result and add samples
            for s = 1:size(curr_samples_fullindx,2)

               %Get idx
               x_s =  double(curr_samples_fullindx(1,s));
               y_s =  double(curr_samples_fullindx(2,s));
               ax_s = double(curr_samples_fullindx(3,s));
               ay_s = double(curr_samples_fullindx(4,s));

               %Sanity
               if( B_sampled_map(ax_s + 1, x_s + 1, 1) ~= 0 )
                   warning('Overwrited !!!')
               end

               %Write
               for ch = 1:3
                    B_sampled(ax_s + 1, x_s + 1, ch) = double(curr_samples(ch, s));
                    B_sampled_map(ax_s + 1, x_s + 1, ch) = 1;
               end


            end

            figure(iter_fig);
            subplot(1,3,1), imshow(B_sampled .* B_sampled_map), title( sprintf('Samples It%d',j) );
            subplot(1,3,2), imshow(B_sampled_map), title( sprintf('Sample-Map It%d',j) ); 

            %Compute reconstruction
            for ch = 1:3
                curr_rec = reconstructLF_full(A(:,:,ch), (Bt(:,:,ch))', ...
                                            lightFieldAnglesY, lightFieldAnglesX, ... 
                                            lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                            layerResolution, layerSize, layerOrigin, layerDistance, ...
                                            layerSubset(1), layerSubset(2) );
                                        
                curr_rec = curr_rec / scaling_factor;

                LF_rec(:,:,ch) = reshape(curr_rec, lightFieldResolution(2), lightFieldResolution(4) );

            end
            subplot(1,3,3), imshow(LF_rec ), title( sprintf('Reconstruction It%d',j) );;

            pause(1.0)
            
            %wait
            %w = waitforbuttonpress; 
        end
        
        %Termination
        rel_change = 0;
        diff = A - A_old;
        rel_change = rel_change + norm(diff(:),2)/norm(A_old(:),2);
        diff = Bt - Bt_old;
        rel_change = rel_change + norm(diff(:),2)/norm(Bt_old(:),2);
        
        fprintf('Relative change is %3.3f\n', rel_change)

        if( rel_change < tol )
            break;
        end

    end

    % reshape final result  
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % compute reconstructed light field
    lightFieldRec = zeros(lightFieldResolution);
    for ch = 1:3

        %Compute reconstruction
        curr_rec = reconstructLF_full(A(:,:,ch), Bt(:,:,ch)', ...
                                    lightFieldAnglesY, lightFieldAnglesX, ... 
                                    lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                    layerResolution, layerSize, layerOrigin, layerDistance, ...
                                    layerSubset(1), layerSubset(2) );
        
         curr_rec = curr_rec / scaling_factor;

        lightFieldRec(:,:,:,:,ch) = single( curr_rec );

    end

    lightFieldRec   = reshape(lightFieldRec, lightFieldResolution);
    layerFrontRec       = reshape(A, layerColorSize);  
    layerBackRec       = reshape(Bt, layerColorSize);  
    
    %Pack layers
    layersRec = cell(1,2);
    layersRec{1,1} = layerFrontRec;
    layersRec{1,2} = layerBackRec;
    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%NOTE: THIS FUNCTION IS HERE NOT FORMULATED FOR 4D-LIGHTFIELDS
function rec = reconstructLF_full(A, B, ...
                                lightFieldAnglesY, lightFieldAnglesX, ... 
                                lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                layerResolution, layerSize, layerOrigin, layerDistance, ...
                                layerFrontIdx, layerBackIdx)

    %All single channel
    AB = A * B;

    %Initialize
    rec = zeros( lightFieldResolution(2), lightFieldResolution(4) );
    
    %Compute constants
    lightFieldPixelSize     = lightFieldSize ./ [lightFieldResolution(3) lightFieldResolution(4)];   
    lookupScaleX = (layerResolution(2))/layerSize(2);
    
    layerIndices = [layerFrontIdx, layerBackIdx]; %From front to back
    
    for x = 1:lightFieldResolution(4)
        for ax = 1:lightFieldResolution(2)

            %######## Compute hits
            
            %Angle and center
            vx = lightFieldAnglesX(ax);
            lightFieldPixelCentersX = (lightFieldOrigin(2)+lightFieldPixelSize(2)/2) + (x - 1) * lightFieldPixelSize(2);

            % shift light field pixels to pixel centers of 1st volume slice
            rayPositionsFirstX = lightFieldPixelCentersX - lightFieldOrigin(3)*vx + layerOrigin(3)*vx; 

            % iterate over all volume slices
            layerHitsX = [-1, -1];
            for l = 1:length(layerIndices)

                %Get current layer index and shifts
                layer = layerIndices(l);

                % shift x positions of rays to next slice
                rayPositionsX = rayPositionsFirstX - layer*layerDistance*vx;

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                % set values in propagation matrix

                % nearest interpolation
                % basisFunctionIndicesForRaysX = interp1(basisFunctionsX,basisFunctionsXX,rayPositionsX,'linear',0);

                % get currently effected layer pixels - matrix column indices for each ligh ray                      
                layerPixelIndicesForRaysX = ceil(lookupScaleX * (rayPositionsX-layerOrigin(2)));                        

                % kick out stuff that's outside
                layerPixelIndicesForRaysX(layerPixelIndicesForRaysX>layerResolution(2)) = -1;
                layerPixelIndicesForRaysX(layerPixelIndicesForRaysX<1) = -1;

                if( layerPixelIndicesForRaysX < 0 ) 
                    continue;
                end
                
                % save in vector
                layerHitsX(l) = layerPixelIndicesForRaysX;

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

            end
            
            %Early exit
            if( any( layerHitsX < 0 ) )
                continue;
            end
            
            %Set
            rec(ax, x) = AB(layerHitsX(1), layerHitsX(2));
        end
    end
    
    %Reshape
    rec = reshape(rec, lightFieldResolution(1:end-1) );
end
