% ***Image synthesis for compressive displays***
%
% see our paper: F. Heide, G. Wetzstein, R. Raskar, W. Heidrich.
% ''Adaptive Image Synthesis for Compressive Displays'', SIGGRAPH 2013.
%
% HR3D sampling and optimization
%
% Copyright (C) 2013. Felix Heide
% Email: fheide@cs.ubc.ca

function [layersRec lightFieldRec B_sampled_map rec_time] = ...
    computeAttenuationLayersFromLightField2D_HR3D_stochastic( lightFieldResolution, ...
                                                         lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lfResolution, lightFieldOrigin,...
                                                         layerResolution, layerSize, layerOrigin, layerDistance, basisFunctionType, ...
                                                         timeSteps, layerSubset, ...
                                                         chainLength, mutation_sigma_spatial, ...
                                                         mutation_sigma_angular, thresh_residual, ...
														 numChainDuplicateChecks, ...
                                                         max_sampling_it, max_it, ...
                                                         time, ...
                                                         scaling_factor, ...
                                                         pbrt_ptr, ...
                                                         verbose, lightfield1Dref)      
                                                     
    % check for parameters
    if nargin == 0
        error('computeAttenuationLayersFromLightField2D only works with all parameters!');    
    end         
                                                    
    % check if system is under-determined
    if prod(layerResolution) > prod(lightFieldResolution)
        error('Optimization is under-determined! This is not supported!');
    end
    
    % check if two layers
    if length(layerSubset) ~= 2
        error('Currently supports only two layers!');
    end        
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %CUDA
    use_cuda = true;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % set desired solution, constraints etc according to the current mode
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % initial iterate - random values
    layerColorSize = [layerResolution(1), layerResolution(2), timeSteps, 3];
    
    frontRec = rand( layerColorSize );
    backRec =  rand( layerColorSize );
    
    %frontRec = zeros( layerColorSize ) + 1e-9;
    %backRec =  zeros( layerColorSize ) + 1e-9;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Stochastic observation minimization
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    
    %Initial iterates
    A = reshape(frontRec, [], timeSteps, 3 );
    Bt = reshape(backRec, [], timeSteps, 3 );
    
    %Initial samples
    samples = [];
    samples_indices = [];
    samples_intersects = [];
    
    %Initial reconstruciton
    lRes= lightFieldResolution; %Tmp
    lrec = zeros(lRes);
    lrec = single(lrec); %% Use single here

    %Figure
    if strcmp(verbose,'all')
        iter_fig = figure();
        
        %PSNR plot
        psnr_plot = figure();
        psnrs = [];
        sampleratios = [];
        
        %Compute reconstruction
        for ch = 1:3
            curr_rec = reconstructLF_full(A(:,:,ch), Bt(:,:,ch)', ...
                                        lightFieldAnglesY, lightFieldAnglesX, ... 
                                        lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                        layerResolution, layerSize, layerOrigin, layerDistance, ...
                                        layerSubset(1), layerSubset(2) );

            LF_rec(:,:,ch) = reshape(curr_rec, lightFieldResolution(2), lightFieldResolution(4) );
            LF_rec(:,:,ch) = LF_rec(:,:,ch) / scaling_factor;

        end
        %subplot(1,3,3), imshow(LF_rec);
        
        %Add precomputed file
        %{
        lightfield1Dref = double(imread('./data/LightFieldSlice_50deg.png'));
        lightfield1Dref = imresize( lightfield1Dref, [260, 420]);
        lightfield1Dref = lightfield1Dref./max(lightfield1Dref(:));
        %}

        pad = 5;
        psnr_scale = lightfield1Dref(pad + 1:end - pad,pad + 1:end - pad) ;
        psnr_scale = max(psnr_scale(:));

        diffRec = lightfield1Dref(pad + 1:end - pad,pad + 1:end - pad)./psnr_scale - LF_rec(pad + 1:end - pad,pad + 1:end - pad)./psnr_scale;
        MSE = mean( diffRec(:).^2 );
        PSNR = 10*log10( 1/MSE );
        percentage_samples = size(samples,2) / ( lRes(1) * lRes(2) * lRes(3) * lRes(4) ) * 100;

        fprintf('PSNR = %3.3f, Percent samples = %3.3f \n', PSNR, percentage_samples);
        
        %Save initial PSNR and samples
        psnrs = [psnrs,PSNR];
        sampleratios = [sampleratios,percentage_samples];
        
        %Compute PSNR
        %psnrs = [psnrs,0];
        %sampleratios = [sampleratios,0];
    end

    %Debug
    if strcmp(verbose, 'brief') || strcmp( verbose, 'all')
        fprintf('\nMinimizing with stochastic observation sampling\n')
    end
    
    %Only used for debug
    LF_rec = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);
    B_sampled_single_ch = zeros(lightFieldResolution(2), lightFieldResolution(4));
    B_sampled = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);
    B_sampled_map = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);

    %Do sampling iterations

    rec_time = 0;
    for j = 1:max_sampling_it

        %% Sampling step: Sample residual stochastically now
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if strcmp(verbose, 'brief') || strcmp( verbose, 'all')
            fprintf('\n\nSampling process %d.\n', j)
        end

        %% Compute last (dense) reconstruction
        for ch = 1:3
            
            %Compute reconstruction
            curr_rec = reconstructLF_full(A(:,:,ch), Bt(:,:,ch)', ...
                                        lightFieldAnglesY, lightFieldAnglesX, ... 
                                        lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                        layerResolution, layerSize, layerOrigin, layerDistance, ...
                                        layerSubset(1), layerSubset(2) );
                                    
            curr_rec = curr_rec / scaling_factor;
            
            lrec(:,:,:,:,ch) = single( curr_rec );
            
        end

        %% Render data
        render_params = single([chainLength; mutation_sigma_spatial; ...
                                mutation_sigma_angular; thresh_residual; numChainDuplicateChecks]);
                            
        [curr_samples, curr_samples_fullindx] = mex_pbrt('sample', pbrt_ptr, shiftdim( single(lrec), 2) , render_params, single(time));

        curr_samples = double(curr_samples);
        curr_samples_fullindx = int32(curr_samples_fullindx);
        
        
        %% Generate sample reconstructions
        
        %Compute linear indices
        curr_samples_indx = curr_samples_fullindx(1,:) * ( lRes(1) * lRes(2) * lRes(3) ) + ... %x * (resY *  rAY * rAX)
                            curr_samples_fullindx(2,:) * ( lRes(1) * lRes(2) ) + ... %y * rAY * rAX
                            curr_samples_fullindx(3,:) *   lRes(1) + ...    %ax * rAY 
                            curr_samples_fullindx(4,:) +  1;   %ay + 1 (Matlab matrix offset)
                        
        %Transform lightfield
        %{
        % no zeros in the light field - bad things happen in log-space!
        curr_samples( curr_samples < eps ) = eps;
        %}
        
        %HR3D transform
        curr_samples = curr_samples * scaling_factor;
        
        %Find hit positions
        layerInt = precomputeLayerIntersectionsHR3D_sampled_dense( lightFieldAnglesY, lightFieldAnglesX, ... 
                                                    lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                                    layerResolution, layerSize, layerOrigin, layerDistance, double(curr_samples_fullindx), ...
                                                    layerSubset(1), layerSubset(2), ...
                                                    basisFunctionType, 1 );
        %Remove samples with bad basis             
        validsamples = (layerInt(3,:) > 0) & (layerInt(6,:) > 0);
        curr_samples_indx = curr_samples_indx(:, validsamples );
        curr_samples_fullindx = curr_samples_fullindx(:, validsamples );
        curr_samples = curr_samples(:, validsamples );
        layerInt = layerInt(:, validsamples);

        %Add samples
        samples         = [samples , curr_samples];
        samples_intersects = [samples_intersects , layerInt]; 
        
        %Unique test
        samples_indices = [samples_indices, curr_samples_indx];
        if( ~isequal( unique(samples_indices,'stable'), samples_indices)) 
            error('Samples are not unique !')
        end
        
        %Do gradient iterations
        if strcmp(verbose, 'brief') || strcmp( verbose, 'all')
            fprintf('\n\nGradient decent process %d.\n', j)
        end

        %% Gradient step: Do MATRIX FACTORIZATION iterations
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        tic;
                
        %Tolerances
        tol = 1e-3;
        
        %Save prev
        A_old = A;
        Bt_old = Bt;
      
        %% Run matrix fact
        for ch = 1:3
            
            %Get current matrices
            A_ch = A(:,:,ch);
            B_ch = Bt(:,:,ch)';
            
            %FULL EIGEN VERSION
            [A_ch, B_ch] = sparse_matrix_factorization(samples, int32(ch), ...
                                                       A_ch, B_ch, int32(samples_intersects), tol, int32(max_it));
             
            %{
            %FULL CUDA VERSION
            
            [A_ch, B_ch] = sparse_matrix_factorization_cuda(single(samples), int32(ch), ...
                                                        single(A_ch), single(B_ch), int32(samples_intersects));
                                                    
            A_ch = double( A_ch );
            B_ch = double( B_ch );
            %}

            % initialize variables
            %A_ch = rand(size(A_ch));
            %B_ch = rand(size(B_ch));

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

            %{
            % small value thresh
            epsilon         = 1e-9;
            epsilonLower    = epsilon;
            
            % Regularizers
            Areg = zeros(size(A_ch));
            Breg = zeros(size(B_ch));
            
            %Regularizer weights
            lambdaA = 0.005; %0.05
            lambdaB = 0.005; %0.05
            
            %Preallocate and store samples
            %L_s = sparse( samples_intersects(3,:), samples_intersects(6,:),samples(ch,:), size(A_ch,1), size(B_ch,2), size(samples_intersects,2) );
            %AB = spalloc(size(A_ch,1), size(B_ch,2), size(samples_intersects,2));

            % start multiplicative update rules
            for iter=1:max_it

                %Regularizers 
                Areg = A_ch;% - circshift(A,-1);  
                Breg = B_ch;% - circshift(A,-1);
                
                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                
                %Compute AB once
                if use_cuda
                    AB_samples = double( AmultB_cuda( single(A_ch), single(B_ch), int32(samples_intersects)) );
                else
                    AB_samples = AmultB(A_ch,B_ch,int32(samples_intersects));
                end
                %AB_samples = AmultB(A_ch,B_ch,int32(samples_intersects));
                
                %Store in sparse  matrix
                %AB_samples = sum(A_ch( samples_intersects(3,:), :)' .* B_ch( :,samples_intersects(6,:)), 1);
                %AB = sparse(samples_intersects(3,:), samples_intersects(6,:), AB_samples, size(A_ch,1), size(B_ch,2), size(samples_intersects,2));

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                % update A
                if use_cuda
                    upper = double( LmultBt_cuda( single(samples), int32(ch), single(B_ch), int32(samples_intersects)) );
                else
                    upper = LmultBt(samples, int32(ch), B_ch, int32(samples_intersects));
                end
                %upper = L_s * B_ch';
                upper = max(upper + lambdaA*Areg, epsilon);

                if use_cuda
                    LBt = double( LmultBt_cuda( single(AB_samples), int32(1), single(B_ch), int32(samples_intersects)) ); %Here allways single ch
                else
                    LBt = LmultBt(AB_samples, int32(1), B_ch, int32(samples_intersects)); %Here allways single ch
                end
                %LBt = AB * B_ch';
                
                A_ch = A_ch .* upper ./ (LBt + epsilonLower);    
                % clamp to feasible region
                A_ch(A_ch<0)=epsilon; A_ch(A_ch>1)=1; 

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                % update B
                if use_cuda
                    upper = double( AtmultL_cuda( single(samples), int32(ch), single(A_ch), int32(samples_intersects)) );
                else
                    upper = AtmultL(samples, int32(ch), A_ch, int32(samples_intersects));
                end
                %upper = A_ch' * L_s;
                upper = max(upper + lambdaB*Breg, epsilon);

                if use_cuda
                    AtL = double( AtmultL_cuda( single(AB_samples), int32(1), single(A_ch), int32(samples_intersects)) ); %Here allways single ch
                else
                    AtL = AtmultL(AB_samples, int32(1), A_ch, int32(samples_intersects)); %Here allways single ch
                end
                %AtL = A_ch' * AB;
                
                B_ch = B_ch .* upper ./ (AtL + epsilonLower);    
                % clamp to feasible region
                B_ch(B_ch<0)=epsilon; B_ch(B_ch>1)=1; 

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                % plot time
                %if strcmp(verbose,'brief') || strcmp(verbose,'all')
                %    disp([num2str(iter) ' | ' num2str(max_it)]);
                %end

            end
            %}
            
            %Write back
            A(:,:,ch) = A_ch;
            Bt(:,:,ch) = B_ch';

        end
        
        tt = toc;
        rec_time = rec_time + tt;
        disp(['Local Matrix factorization took ' num2str(tt) ' secs']);
        
        
        %% Debug
        if strcmp(verbose,'all')
            %Debug layer
            %figure(), imagesc(reshape(exp(x), [layerResolution(3) layerResolution(2), 3] )), colormap gray, colorbar;

            %Write
            %if j ~= 1
                for ch = 1:3
                    curr_samples_indx_ch = (ch - 1) * ( lRes(1) * lRes(2) * lRes(3) * lRes(4) ) + ... %x * (resX *resY *  rAY * rAX)
                                           double(curr_samples_indx); 

                    B_sampled(curr_samples_indx_ch(:)) = reshape( curr_samples(ch,:), [], 1 );
                    B_sampled_map(curr_samples_indx_ch(:)) = 1;
                end
            %end


            figure(iter_fig);
            subplot(1,3,1), imshow(B_sampled .* B_sampled_map), title( sprintf('Samples It%d',j) );
            subplot(1,3,2), imshow(B_sampled_map), title( sprintf('Sample-Map It%d',j) ); 

            %Compute reconstruction
            for ch = 1:3
                curr_rec = reconstructLF_full(A(:,:,ch), Bt(:,:,ch)', ...
                                            lightFieldAnglesY, lightFieldAnglesX, ... 
                                            lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                            layerResolution, layerSize, layerOrigin, layerDistance, ...
                                            layerSubset(1), layerSubset(2) );
                                        
                curr_rec = curr_rec / scaling_factor;

                LF_rec(:,:,ch) = reshape(curr_rec, lightFieldResolution(2), lightFieldResolution(4) );

            end
            subplot(1,3,3), imshow(LF_rec), title( sprintf('Reconstruction It%d',j) );

            %Add precomputed file
            %{
            lightfield1Dref = double(imread('./data/LightFieldSlice_50deg.png'));
            lightfield1Dref = imresize( lightfield1Dref, [260, 420]);
            lightfield1Dref = lightfield1Dref./max(lightfield1Dref(:));
            %}
            
            pad = 5;
            psnr_scale = lightfield1Dref(pad + 1:end - pad,pad + 1:end - pad) ;
            psnr_scale = max(psnr_scale(:));
            
            diffRec = lightfield1Dref(pad + 1:end - pad,pad + 1:end - pad)./psnr_scale - LF_rec(pad + 1:end - pad,pad + 1:end - pad)./psnr_scale;
            MSE = mean( diffRec(:).^2 );
            PSNR = 10*log10( 1/MSE );
            percentage_samples = size(samples,2) / ( lRes(1) * lRes(2) * lRes(3) * lRes(4) ) * 100;
            
            fprintf('PSNR = %3.3f, Percent samples = %3.3f \n', PSNR, percentage_samples);
            
            %Plot current PSNR and samples
            psnrs = [psnrs,PSNR];
            sampleratios = [sampleratios,percentage_samples];
            figure(psnr_plot);
            clf(psnr_plot);
            plot(0:length(psnrs)-1, psnrs,'-r',0:length(psnrs)-1, sampleratios,'-b')
            legend('PSNR [dB]','Samples [%]','Location', 'SouthEast');
            xlabel('Iteration');
            title(sprintf('epsilon = %3.3f', thresh_residual));
            
            pause(1.0)

            %wait
            %w = waitforbuttonpress; 
            
            %Write stuff
            %imwrite(B_sampled .* B_sampled_map, sprintf('dump_hr3d_0.05/samples_it%03d.png', j));
            %imwrite(B_sampled_map, sprintf('dump_hr3d_0.05/sampleloc_it%03d.png', j));
            %imwrite(LF_rec, sprintf('dump_hr3d_0.05/rec_it%03d.png', j));
            
            %imwrite(LF_rec, sprintf('rec_it%03d.png', j));
        end
        
        %Termination
        rel_change = 0;
        diff = A - A_old;
        rel_change = rel_change + norm(diff(:),2)/norm(A_old(:),2);
        diff = Bt - Bt_old;
        rel_change = rel_change + norm(diff(:),2)/norm(Bt_old(:),2);

        if( rel_change < tol )
            break;
        end
        
    end
    
    %Write stuff
    imwrite(B_sampled .* B_sampled_map, sprintf('samples_it%03d.png', j));
    imwrite(B_sampled_map, sprintf('sampleloc_it%03d.png', j));
    imwrite(LF_rec, sprintf('rec_it%03d.png', j));

    % reshape final result  
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % compute reconstructed light field
    lightFieldRec = zeros(lightFieldResolution);
    for ch = 1:3

        %Compute reconstruction
        curr_rec = reconstructLF_full(A(:,:,ch), Bt(:,:,ch)', ...
                                    lightFieldAnglesY, lightFieldAnglesX, ... 
                                    lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                    layerResolution, layerSize, layerOrigin, layerDistance, ...
                                    layerSubset(1), layerSubset(2) );

        curr_rec = curr_rec / scaling_factor;
                                    
        lightFieldRec(:,:,:,:,ch) = single( curr_rec );

    end

    lightFieldRec   = reshape(lightFieldRec, lightFieldResolution);
    layerFrontRec       = reshape(A, layerColorSize);  
    layerBackRec       = reshape(Bt, layerColorSize);  
    
    %Pack layers
    layersRec = cell(1,2);
    layersRec{1,1} = layerFrontRec;
    layersRec{1,2} = layerBackRec;
    
    %Save PSNR
    %save('psnrs.mat', 'psnrs');
    %save('sampleratios.mat', 'sampleratios');          
            
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%NOTE: THIS FUNCTION IS HERE NOT FORMULATED FOR 4D-LIGHTFIELDS
function rec = reconstructLF_full(A, B, ...
                                lightFieldAnglesY, lightFieldAnglesX, ... 
                                lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                layerResolution, layerSize, layerOrigin, layerDistance, ...
                                layerFrontIdx, layerBackIdx)

    %All single channel
    AB = A * B;

    %Initialize
    rec = zeros( lightFieldResolution(2), lightFieldResolution(4) );
    
    %Compute constants
    lightFieldPixelSize     = lightFieldSize ./ [lightFieldResolution(3) lightFieldResolution(4)];   
    lookupScaleX = (layerResolution(2))/layerSize(2);
    
    layerIndices = [layerFrontIdx, layerBackIdx]; %From front to back
    
    for x = 1:lightFieldResolution(4)
        for ax = 1:lightFieldResolution(2)

            %######## Compute hits
            
            %Angle and center
            vx = lightFieldAnglesX(ax);
            lightFieldPixelCentersX = (lightFieldOrigin(2)+lightFieldPixelSize(2)/2) + (x - 1) * lightFieldPixelSize(2);

            % shift light field pixels to pixel centers of 1st volume slice
            rayPositionsFirstX = lightFieldPixelCentersX - lightFieldOrigin(3)*vx + layerOrigin(3)*vx; 

            % iterate over all volume slices
            layerHitsX = [-1, -1];
            for l = 1:length(layerIndices)

                %Get current layer index and shifts
                layer = layerIndices(l);

                % shift x positions of rays to next slice
                rayPositionsX = rayPositionsFirstX - layer*layerDistance*vx;

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                % set values in propagation matrix

                % nearest interpolation
                % basisFunctionIndicesForRaysX = interp1(basisFunctionsX,basisFunctionsXX,rayPositionsX,'linear',0);

                % get currently effected layer pixels - matrix column indices for each ligh ray                      
                layerPixelIndicesForRaysX = ceil(lookupScaleX * (rayPositionsX-layerOrigin(2)));                        

                % kick out stuff that's outside
                layerPixelIndicesForRaysX(layerPixelIndicesForRaysX>layerResolution(2)) = -1;
                layerPixelIndicesForRaysX(layerPixelIndicesForRaysX<1) = -1;

                if( layerPixelIndicesForRaysX < 0 ) 
                    continue;
                end
                
                % save in vector
                layerHitsX(l) = layerPixelIndicesForRaysX;

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

            end
            
            %Early exit
            if( any( layerHitsX < 0 ) )
                continue;
            end
            
            %Set
            rec(ax, x) = AB(layerHitsX(1), layerHitsX(2));
        end
    end
    
    %Reshape
    rec = reshape(rec, lightFieldResolution(1:end-1) );
end
