% ***Image synthesis for compressive displays***
%
% see our paper: F. Heide, G. Wetzstein, R. Raskar, W. Heidrich.
% ''Adaptive Image Synthesis for Compressive Displays'', SIGGRAPH 2013.
%
% Layered3D sampling and optimization
%
% Copyright (C) 2013. Felix Heide
% Email: fheide@cs.ubc.ca


function [layersRec lightFieldRec B_sampled_map sart_time] = ...
    computeAttenuationLayersFromLightField2D_stochastic( lightFieldResolution, minTransmission, ...
                                                         lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lfResolution, lightFieldOrigin,...
                                                         layerResolution, layerSize, layerOrigin, layerDistance, basisFunctionType, ...
                                                         chainLength, mutation_sigma_spatial, ...
                                                         mutation_sigma_angular, thresh_residual, ...
														 numChainDuplicateChecks, ...
                                                         max_sampling_it, max_it, ...
                                                         pbrt_ptr, ...
                                                         verbose)                                              
        
    % check for parameters
    if nargin == 0
        error('computeAttenuationLayersFromLightField2D only works with all parameters!');    
    end    
                                            
	% declare matrix a global variable to save memory and time!
    global T;                 
                                                    
    % check if system is under-determined
    if prod(layerResolution) > prod(lightFieldResolution)
        error('Optimization is under-determined! This is not supported!');
    end        

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % set desired solution, constraints etc according to the current mode
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % here, the optimization needs to be constrained to transmission values
    % between minTransmission and 1, so the log values need to be < 0
    lb      = zeros([prod(layerResolution) 1]) + log(minTransmission);
    ub      = zeros([prod(layerResolution) 1]);

    % initial guess - use either zeros or the backprojected solution (zeros
    % seem to work much better)
    layersRec = zeros(layerResolution) + log(eps);
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Stochastic observation minimization
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % function handle
    Afun = @(x,flag) projectLayersMatrix( x, flag );    
    
    %Initial iterate
    layerResolutionColor = [layerResolution, 3];
    x = zeros( layerResolutionColor );
    for ch = 1:3
         x(:,:,:,ch) = layersRec;
    end
    
    %Initial samples
    samples = [];
    samples_indices = [];
    
    %Initial reconstruciton
    lRes= lightFieldResolution; %Tmp
    lrec = zeros(lRes);
    lrec = single(lrec); %% Use single here

    %Figure
    if strcmp(verbose,'brief') || strcmp(verbose,'all')
        iter_fig = figure();
        iter_layer_fig = figure();
    end

    %Debug
    if strcmp(verbose, 'brief') || strcmp( verbose, 'all')
        fprintf('\nMinimizing with stochastic observation sampling\n')
    end
    
    %Only used for debug
    LF_rec = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);
    B_sampled = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);
    B_sampled_map = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);

    %Do sampling iterations

    sart_time = 0;
    T_curr = [];
    for j = 1:max_sampling_it

        %% Sampling step: Sample residual stochastically now
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if strcmp(verbose, 'brief') || strcmp( verbose, 'all')
            fprintf('\n\nSampling process %d.\n', j)
        end

        %% Compute last reconstruction
        for ch = 1:3
            lrec(:,:,:,:,ch) = single( max( exp( ...
                                            reshape( Afun( reshape(x(:,:,:,ch), [], 1 ), 1) ...
                                                    ,lightFieldResolution(1:end-1) ) ) ...
                                           , 0 ) ...
                                     );
        end

        %% Render data
        render_params = single([chainLength; mutation_sigma_spatial; ...
                                mutation_sigma_angular; thresh_residual; numChainDuplicateChecks]);
                            
       [curr_samples, curr_samples_fullindx] = mex_pbrt('sample', pbrt_ptr, shiftdim( single(lrec), 2) , render_params);

        curr_samples = double(curr_samples);
        curr_samples_fullindx = int64(curr_samples_fullindx);
        
        %% Generate sample reconstructions
        
        %Compute linear indices
        curr_samples_indx = curr_samples_fullindx(1,:) * ( lRes(1) * lRes(2) * lRes(3) ) + ... %x * (resY *  rAY * rAX)
                            curr_samples_fullindx(2,:) * ( lRes(1) * lRes(2) ) + ... %y * rAY * rAX
                            curr_samples_fullindx(3,:) *   lRes(1) + ...    %ax * rAY 
                            curr_samples_fullindx(4,:) +  1;   %ay + 1 (Matlab matrix offset)
                        
        if( ~isequal( unique(curr_samples_indx,'stable'), curr_samples_indx)) 
            error('Samples are not unique !')
        end
                        
        %Transform lightfield
        % no zeros in the light field - bad things happen in log-space!
        curr_samples( curr_samples < eps ) = eps;
        % light field to log light field
        curr_samples = log(curr_samples);   

        % check if everything is finite
        if sum(~isfinite(curr_samples(:))) > 0
            error('(Log) Current samples are not Finite!');
        end 

        %Add samples
        samples         = [samples ; curr_samples'];
        samples_indices = [samples_indices ; curr_samples_indx']; 
        
        %Do gradient iterations
        if strcmp(verbose, 'brief') || strcmp( verbose, 'all')
            fprintf('\n\nGradient decent process %d.\n', j)
        end

        %% Gradient step: Do SART iterations
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % run SART
        
        %% Sparse SART matrix preparation (currently the same for all colors)
        %T_curr = T(samples_indices, :);
        
        % single channel light field resolution
        lfResolution = [lightFieldResolution(1) lightFieldResolution(2) lightFieldResolution(3) lightFieldResolution(4)];
        
        T_curr_samples = precomputeSparsePropagationMatrixLayers3D_sampled_dense( lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lfResolution, lightFieldOrigin,...
                                            layerResolution, layerSize, layerOrigin, layerDistance, double(curr_samples_fullindx), ...
                                            basisFunctionType, 1 );
                                        
        T_curr = cat(1, T_curr,  T_curr_samples);
        
        
        %TODO: Debug test
        %matrix_dist = T(samples_indices, :) - T_curr;
        %matrix_dist = full( sum(sum(matrix_dist)) )
        
        
        AfunCurr = @(x,flag) projectLayersMatrixMod( x, flag, T_curr  );
        
         
        %% Run sart
        for ch = 1:3
            tic;
                x(:,:,:,ch) = reshape( ...
                                        SART(AfunCurr, samples(:,ch), lb,ub, reshape( x(:,:,:,ch), [],1) ,max_it), ...
                                        layerResolution );
            tt = toc;
            disp(['Local SART took ' num2str(tt) ' secs']);

            sart_time = sart_time + tt;
        end
        
        
        %% Debug
        
        %Debug layer
        %figure(), imagesc(reshape(exp(x), [layerResolution(3) layerResolution(2), 3] )), colormap gray, colorbar;
        
        %1D Generate result and add samples
        for s = 1:size(curr_samples_fullindx,2)

           %Get idx
           x_s =  double(curr_samples_fullindx(1,s));
           y_s =  double(curr_samples_fullindx(2,s));
           ax_s = double(curr_samples_fullindx(3,s));
           ay_s = double(curr_samples_fullindx(4,s));

           %Sanity
           if( B_sampled_map(ax_s + 1, x_s + 1, 1) ~= 0 )
               warning('Overwrited !!!')
           end

           %Write
           for ch = 1:3
                B_sampled(ax_s + 1, x_s + 1, ch) = double(curr_samples(ch, s));
                B_sampled_map(ax_s + 1, x_s + 1, ch) = 1;
           end
           
           
        end
    
        figure(iter_fig);
        subplot(1,3,1), imshow(exp(B_sampled) .* B_sampled_map), title( sprintf('Samples It%d',j) );
        subplot(1,3,2), imshow(B_sampled_map), title( sprintf('Sample-Map It%d',j) ); 
        
        for ch = 1:3
            LF_rec(:,:,ch) = reshape( Afun( reshape( x(:,:,:,ch),[], 1), 1), ...
                                        lightFieldResolution(2), lightFieldResolution(4) );
        end
        subplot(1,3,3), imshow(exp(LF_rec )), title( sprintf('Reconstruction It%d',j) ); 
        pause(1.0)
        
        %Write stuff
        %imwrite(exp(B_sampled) .* B_sampled_map, sprintf('dump_test/samples_it%03d.png', j));
        %imwrite(B_sampled_map, sprintf('dump_test/sampleloc_it%03d.png', j));
        %imwrite(exp(LF_rec ), sprintf('dump_test/rec_it%03d.png', j));
        
        %wait
        %w = waitforbuttonpress; 
        
    end
    
    imwrite(exp(B_sampled) .* B_sampled_map, sprintf('samples_it%03d.png', j));
    imwrite(B_sampled_map, sprintf('sampleloc_it%03d.png', j));
    imwrite(exp(LF_rec ), sprintf('rec_it%03d.png', j));

    % reshape final result  
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % compute reconstructed light field
    lightFieldRec = zeros(lRes);
    for ch = 1:3
        lightFieldRec(:,:,:,:,ch) = single( reshape( Afun( reshape(x(:,:,:,ch), [], 1 ), 1) ...
                                                ,lightFieldResolution(1:end-1) ) );
    end

    lightFieldRec   = reshape(lightFieldRec, lightFieldResolution);
    layersRec       = x;  
    
    % exponentiate reconstruction
    lightFieldRec   = exp(lightFieldRec);   
    layersRec   = exp(layersRec); 
    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% use explicit matrix (sparse)

function result = projectLayersMatrix( Y, flag )    
    % declare T a global variable
    global T;

    % forward projection
    if flag > 0                  
        result = T*Y;                                       
    % back-projection
    elseif flag < 0 
          result = T'*Y;                                                                                                         
    % forward and then back-projection
    else
      result = T'*(T*Y);        
    end    
end

function result = projectLayersMatrixMod( Y, flag, M )    

    % forward projection
    if flag > 0                  
        result = M*Y;                                       
    % back-projection
    elseif flag < 0 
        result = M'*Y;                                                                                                         
    % forward and then back-projection
    else
        result = M'*(M*Y);        
    end    
end
