%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Given a bunch of parameters, compute the layer intersections
%
%   input: basisFunctionType - 0 discrete light field ray positions and angles
%                               
%   Felix Heide
%   PSM Lab | University of British Columbia
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [layerInt] = precomputeLayerIntersectionsHR3D_sampled_dense( lightFieldAnglesY, lightFieldAnglesX, ... 
                                                    lightFieldSize, lightFieldResolution, lightFieldOrigin, ...
                                                    layerResolution, layerSize, layerOrigin, layerDistance, ...
                                                    sampleid, layerFrontIdx, layerBackIdx, ...
                                                    basisFunctionType, drawMode )
                                                    
    if (basisFunctionType ~= 0) 
        error(['Basis function type ' num2str(basisFunctionType) ' currently not supported!']);
    end
                                                                  
    % absolute x coordinates for light field pixel centers
    lightFieldPixelSize     = lightFieldSize ./ [lightFieldResolution(3) lightFieldResolution(4)];   
    
    % scale to be applied to the float-positions to get a basis function index
    lookupScaleY = (layerResolution(1))/layerSize(1);
    lookupScaleX = (layerResolution(2))/layerSize(2);
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % initialize result (rowvec)
    layerInt = -ones(6, size(sampleid,2)); %Front and back, default -1
    
    % for all samples in vector form
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %Rowvecs
    vyIdx= sampleid(4,:) + 1;
    vxIdx= sampleid(3,:) + 1;  

    % actual angle in v units (rowvecs)
    vy = lightFieldAnglesY(vyIdx);
    vx = lightFieldAnglesX(vxIdx);

     % get current pixel pos (rowvecs)
    lightFieldPixelCentersX = (lightFieldOrigin(2)+lightFieldPixelSize(2)/2) + sampleid(1,:) * lightFieldPixelSize(2);
    lightFieldPixelCentersY = (lightFieldOrigin(1)+lightFieldPixelSize(1)/2) + sampleid(2,:) * lightFieldPixelSize(1);    

    % shift light field pixels to pixel centers of 1st volume slice
    rayPositionsFirstX = lightFieldPixelCentersX - lightFieldOrigin(3)*vx + layerOrigin(3)*vx;
    rayPositionsFirstY = lightFieldPixelCentersY - lightFieldOrigin(3)*vy + layerOrigin(3)*vy;  
    
    % iterate over all volume slices
    layerIndices = [layerFrontIdx, layerBackIdx]; %From front to back
    for l = 1:length(layerIndices)
        
        %Get current layer index and shifts
        layer = layerIndices(l);
        
        % shift x positions of rays to next slice
        rayPositionsX = rayPositionsFirstX - layer * layerDistance*vx;
        rayPositionsY = rayPositionsFirstY - layer * layerDistance*vy;

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % set values in propagation matrix
        
        % nearest interpolation
        % basisFunctionIndicesForRaysX = interp1(basisFunctionsX,basisFunctionsXX,rayPositionsX,'linear',0);

        % get currently effected layer pixels - matrix column indices for each ligh ray                      
        layerPixelIndicesForRaysX = ceil(lookupScaleX * (rayPositionsX-layerOrigin(2)));                        
        layerPixelIndicesForRaysY = ceil(lookupScaleY * (rayPositionsY-layerOrigin(1))); 
        
         %Compute linear indices
        layerLinearIndx = (layerPixelIndicesForRaysX - 1) * layerResolution(1) + ...
                          (layerPixelIndicesForRaysY - 1) + 1; %Matlab offset 

        % kick out stuff that's outside
        layerPixelIndicesForRaysX(layerPixelIndicesForRaysX>layerResolution(2)) = -1;
        layerPixelIndicesForRaysX(layerPixelIndicesForRaysX<1) = -1;
        layerPixelIndicesForRaysY(layerPixelIndicesForRaysY>layerResolution(1)) = -1;
        layerPixelIndicesForRaysY(layerPixelIndicesForRaysY<1) = -1;  
        
        layerLinearIndx( layerPixelIndicesForRaysX < 0 | layerPixelIndicesForRaysY < 0) = -1;
        
        % save in vector
        layerInt( (l - 1)*3 + 1 + 0, : ) = layerPixelIndicesForRaysX; %X
        layerInt( (l - 1)*3 + 1 + 1, : ) = layerPixelIndicesForRaysY; %Y       
        layerInt( (l - 1)*3 + 1 + 2, : ) = layerLinearIndx; %LinIdx  

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    end
    
end


