%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Given a bunch of parameters, compute the sparse propagation matrix for
%   N 1D attenuation layers in log-space given a desired 4D light field
%
%   input:  drawMode -  0 draw nothing
%                       1 show progress bar
%
%           basisFunctionType - 0 discrete light field ray positions and angles
%                               1 discrete light field ray positions, box
%
%   Felix Heide/Gordon Wetzstein [fheide@cs.ubc.ca]
%   PSM Lab | University of British Columbia
%   February 2011
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% http://blogs.mathworks.com/loren/2007/03/01/creating-sparse-finite-element-matrices-in-matlab/
% [row, col, val] = find(accumarray([row, col], val, [], @sum, [], true));
% OR
% [r, c, v] = find(accumarray({row, col}, val))
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TODO:
%   - despite using a single global instance of T, the memory management
%       should be enhanced even more!
%   - linear interpolation for rays (basisFunctionType 1)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function mat = precomputeSparsePropagationMatrixLayers3D_sampled( lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lightFieldResolution, lightFieldOrigin,...
                                                    layerResolution, layerSize, layerOrigin, layerDistance, sampleid, ...
                                                    basisFunctionType, drawMode )
                                                    
    if (basisFunctionType<0) || (basisFunctionType>1)
        error(['Basis function type ' num2str(basisFunctionType) ' currently not supported!']);
    end
                                                                  
    % absolute x coordinates for light field pixel centers
    lightFieldPixelSize     = lightFieldSize ./ [lightFieldResolution(3) lightFieldResolution(4)];  
        
    % assuming that the angles are sampled at equal distances, get that
    % distance
    lightFieldAngleStep = [0 0];
    if lightFieldResolution(1)>1
        lightFieldAngleStep(1) = lightFieldAnglesY(2)-lightFieldAnglesY(1);
    end
    if lightFieldResolution(2)>1
        lightFieldAngleStep(2) = lightFieldAnglesX(2)-lightFieldAnglesX(1);
    end    
    
    if drawMode == 1
        h               = waitbar(0,'Generating propagation matrix ...');
        numWaitbarCalls = size(sampleid,2);
    end 
    
    % scale to be applied to the float-positions to get a basis function index
    lookupScaleY = (layerResolution(1))/layerSize(1);
    lookupScaleX = (layerResolution(2))/layerSize(2);
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % initialize vectors that will store the non-negative matrix indices and values
    % memory doesn't matter
    indexIVector = [];
    indexJVector = [];
    valueSVector = [];
    
    % for all angles in the samples
    for currSampleId = 1:size(sampleid,2)     
        
        vyIdx= sampleid(4,currSampleId) + 1;
        vxIdx= sampleid(3,currSampleId) + 1;  
        

        % update waitbar
        if drawMode == 1
            currentCalls = currSampleId;
            waitbar( currentCalls / numWaitbarCalls );
        end

        % actual angle in v units
        vy = lightFieldAnglesY(vyIdx);
        vx = lightFieldAnglesX(vxIdx);

         % get current pixel pos
        lightFieldPixelCentersX = (lightFieldOrigin(2)+lightFieldPixelSize(2)/2) + sampleid(1,currSampleId) * lightFieldPixelSize(2);
        lightFieldPixelCentersY = (lightFieldOrigin(1)+lightFieldPixelSize(1)/2) + sampleid(2,currSampleId) * lightFieldPixelSize(1);    


        % if we use box-area integration filtering, use a range of angles
        if (basisFunctionType==1)
            % lower boundary of angular box
            vy = lightFieldAnglesY(vyIdx) - lightFieldAngleStep(1)/2;
            vx = lightFieldAnglesX(vxIdx) - lightFieldAngleStep(2)/2;

            % upper boundary of angular box
            vyUpper = lightFieldAnglesY(vyIdx) + lightFieldAngleStep(1)/2;     
            vxUpper = lightFieldAnglesX(vxIdx) + lightFieldAngleStep(2)/2;

        end

        % shift light field pixels to pixel centers of 1st volume slice
        rayPositionsX = lightFieldPixelCentersX - lightFieldOrigin(3)*vx + layerOrigin(3)*vx;
        rayPositionsY = lightFieldPixelCentersY - lightFieldOrigin(3)*vy + layerOrigin(3)*vy;  

        % if we use box-area integration filtering, use a range of angles
        if basisFunctionType > 0
            % shift upper light field pixels to pixel centers of 1st layer
            rayPositionsXUpper = lightFieldPixelCentersX - lightFieldOrigin(3)*vxUpper + layerOrigin(3)*vxUpper;
            rayPositionsYUpper = lightFieldPixelCentersY - lightFieldOrigin(3)*vyUpper + layerOrigin(3)*vyUpper;                            
        end

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        % iterate over all volume slices
        for layer = 1:layerResolution(3)

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % set values in propagation matrix            
            if basisFunctionType == 0       % nearest interpolation

                % basisFunctionIndicesForRaysX = interp1(basisFunctionsX,basisFunctionsXX,rayPositionsX,'linear',0);

                % get currently effected layer pixels - matrix column indices for each ligh ray                      
                layerPixelIndicesForRaysX = ceil(lookupScaleX .* (rayPositionsX-layerOrigin(2)));                        
                layerPixelIndicesForRaysY = ceil(lookupScaleY .* (rayPositionsY-layerOrigin(1)));   

                % kick out stuff that's outside
                layerPixelIndicesForRaysX(layerPixelIndicesForRaysX>layerResolution(2)) = 0;
                layerPixelIndicesForRaysX(layerPixelIndicesForRaysX<1) = 0;
                layerPixelIndicesForRaysY(layerPixelIndicesForRaysY>layerResolution(1)) = 0;
                layerPixelIndicesForRaysY(layerPixelIndicesForRaysY<1) = 0;   

                % early exit
                if layerPixelIndicesForRaysX~=0 && layerPixelIndicesForRaysY ~= 0
                    % convert 4D subscipts to matrix indices
                    matrixRows = currSampleId;

                    % convert 3D subscripts to matrix indices
                    matrixColumns   = sub2ind(layerResolution, layerPixelIndicesForRaysY, layerPixelIndicesForRaysX, layer);                                        

                    indexIVector = cat(1, indexIVector, matrixRows);
                    indexJVector = cat(1, indexJVector, matrixColumns);
                    valueSVector = cat(1, valueSVector, ones(size(matrixRows)));
                end

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % rectangular region around ray in angle

            elseif basisFunctionType == 1   

                % get currently effected layer pixels - matrix column indices for each ligh ray                      
                layerPixelIndicesForRaysXLeft   = ceil(lookupScaleX .* (rayPositionsXUpper-layerOrigin(2)));                        
                layerPixelIndicesForRaysYUp     = ceil(lookupScaleY .* (rayPositionsYUpper-layerOrigin(1)));                    
                layerPixelIndicesForRaysXRight  = ceil(lookupScaleX .* (rayPositionsX-layerOrigin(2)));                        
                layerPixelIndicesForRaysYDown   = ceil(lookupScaleY .* (rayPositionsY-layerOrigin(1)));                                                            

                % this only works if the number of effected pixels is
                % the same everywhere
                numLayerPixelsPerRayPixelX = 1 + layerPixelIndicesForRaysXRight(1)-layerPixelIndicesForRaysXLeft(1);
                numLayerPixelsPerRayPixelY = 1 + layerPixelIndicesForRaysYDown(1)-layerPixelIndicesForRaysYUp(1);

                % check if it really is the same everywhere
                if  (sum( 1 - ((layerPixelIndicesForRaysXRight-layerPixelIndicesForRaysXLeft)==(numLayerPixelsPerRayPixelX-1)) ) ~= 0) || ...
                    (sum( 1 - ((layerPixelIndicesForRaysYDown-layerPixelIndicesForRaysYUp)==(numLayerPixelsPerRayPixelY-1)) ) ~= 0)
                    error('Dimensions are wrong!');
                end                                        

                % matrix row indices are easy: just the current light field indices for each layer pixel
                currentIndicesRows = repmat(currSampleId, [numLayerPixelsPerRayPixelY numLayerPixelsPerRayPixelX]);

                % construct matrix columns
                currentIndicesColumnsX  = zeros([numLayerPixelsPerRayPixelY numLayerPixelsPerRayPixelX]);
                currentIndicesColumnsY  = zeros([numLayerPixelsPerRayPixelY numLayerPixelsPerRayPixelX]);
                currentIndicesColumnsZ  = layer + zeros([numLayerPixelsPerRayPixelY numLayerPixelsPerRayPixelX]);

                for ky=1:numLayerPixelsPerRayPixelY
                    currentIndicesColumnsY(ky,:) = repmat( (layerPixelIndicesForRaysYUp')+ky-1, [1 numLayerPixelsPerRayPixelX]);
                end

                for kx=1:numLayerPixelsPerRayPixelX
                    currentIndicesColumnsX(:,kx) = repmat( layerPixelIndicesForRaysXLeft+kx-1, [numLayerPixelsPerRayPixelY 1]);
                end

                % remember all valid indices
                validIndices = ones([numLayerPixelsPerRayPixelY numLayerPixelsPerRayPixelX]);
                validIndices(currentIndicesColumnsY<1)                  = 0;
                validIndices(currentIndicesColumnsY>layerResolution(1)) = 0;
                validIndices(currentIndicesColumnsX<1)                  = 0;
                validIndices(currentIndicesColumnsX>layerResolution(2)) = 0;

                % set all invalid indices to 1, otherwise it crashes
                currentIndicesColumnsX(validIndices==0) = 1;
                currentIndicesColumnsY(validIndices==0) = 1;

                % get the matrix indices
                currentIndicesColumns   = sub2ind(layerResolution, currentIndicesColumnsY, currentIndicesColumnsX, currentIndicesColumnsZ);
                clear currentIndicesColumnsX currentIndicesColumnsY currentIndicesColumnsZ;

                % matrix entries
                currentIndicesValues    = sum(sum(validIndices,2),1);
                currentIndicesValues(currentIndicesValues~=0) = 1 ./ currentIndicesValues(currentIndicesValues~=0);
                currentIndicesValues    = repmat(currentIndicesValues, [numLayerPixelsPerRayPixelY numLayerPixelsPerRayPixelX]);       

                % vectorize everything
                validIndices            = validIndices(:);                    
                currentIndicesRows      = currentIndicesRows(:);
                currentIndicesColumns   = currentIndicesColumns(:);
                
                indexIVector = cat(1, indexIVector, reshape(currentIndicesRows(validIndices==1),[],1) );
                indexJVector = cat(1, indexJVector, reshape(currentIndicesColumns(validIndices==1),[],1) );
                valueSVector = cat(1, valueSVector, reshape(currentIndicesValues(validIndices==1),[],1) );

            end

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

            % shift x positions of rays to next slice
            rayPositionsX = rayPositionsX - layerDistance*vx;
            rayPositionsY = rayPositionsY - layerDistance*vy;

            if basisFunctionType > 0
                rayPositionsXUpper = rayPositionsXUpper - layerDistance*vxUpper;
                rayPositionsYUpper = rayPositionsYUpper - layerDistance*vyUpper;
            end

        end
    end  
 
    
    % construct the sparse matrix and add all non-zeros entries
	mat = sparse(indexIVector,indexJVector,valueSVector,size(sampleid,2),prod(layerResolution));
	clear indexIVector indexJVector valueSVector;
    
    if drawMode == 1
        close(h);        
    end
    
end


