%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Given a bunch of parameters, compute the sparse propagation matrix for
%   N 1D attenuation layers in log-space given a desired 4D light field
%
%   input:  basisFunctionType - 0 discrete light field ray positions and angles
%
%   Felix Heide/Gordon Wetzstein [fheide@cs.ubc.ca]
%   PSM Lab | University of British Columbia
%   February 2011
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% http://blogs.mathworks.com/loren/2007/03/01/creating-sparse-finite-element-matrices-in-matlab/
% [row, col, val] = find(accumarray([row, col], val, [], @sum, [], true));
% OR
% [r, c, v] = find(accumarray({row, col}, val))
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TODO:
%   - despite using a single global instance of T, the memory management
%       should be enhanced even more!
%   - linear interpolation for rays (basisFunctionType 1)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function mat = precomputeSparsePropagationMatrixLayers3D_sampled_dense( lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lightFieldResolution, lightFieldOrigin,...
                                                    layerResolution, layerSize, layerOrigin, layerDistance, sampleid, ...
                                                    basisFunctionType, drawMode )
                                                    
    if (basisFunctionType ~= 0) 
        error(['Basis function type ' num2str(basisFunctionType) ' currently not supported!']);
    end
                                                                  
    % absolute x coordinates for light field pixel centers
    lightFieldPixelSize     = lightFieldSize ./ [lightFieldResolution(3) lightFieldResolution(4)];  
        
    % assuming that the angles are sampled at equal distances, get that
    % distance
    lightFieldAngleStep = [0 0];
    if lightFieldResolution(1)>1
        lightFieldAngleStep(1) = lightFieldAnglesY(2)-lightFieldAnglesY(1);
    end
    if lightFieldResolution(2)>1
        lightFieldAngleStep(2) = lightFieldAnglesX(2)-lightFieldAnglesX(1);
    end    
    
    % scale to be applied to the float-positions to get a basis function index
    lookupScaleY = (layerResolution(1))/layerSize(1);
    lookupScaleX = (layerResolution(2))/layerSize(2);
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % initialize vectors that will store the non-negative matrix indices and values
    % memory doesn't matter
    indexIVector = [];
    indexJVector = [];
    valueSVector = [];
    
    % for all samples in vector form
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %Rowvecs
    vyIdx= sampleid(4,:) + 1;
    vxIdx= sampleid(3,:) + 1;  

    % actual angle in v units (rowvecs)
    vy = lightFieldAnglesY(vyIdx);
    vx = lightFieldAnglesX(vxIdx);

     % get current pixel pos (rowvecs)
    lightFieldPixelCentersX = (lightFieldOrigin(2)+lightFieldPixelSize(2)/2) + sampleid(1,:) * lightFieldPixelSize(2);
    lightFieldPixelCentersY = (lightFieldOrigin(1)+lightFieldPixelSize(1)/2) + sampleid(2,:) * lightFieldPixelSize(1);    

    % shift light field pixels to pixel centers of 1st volume slice
    rayPositionsX = lightFieldPixelCentersX - lightFieldOrigin(3)*vx + layerOrigin(3)*vx;
    rayPositionsY = lightFieldPixelCentersY - lightFieldOrigin(3)*vy + layerOrigin(3)*vy;  
    
    % iterate over all volume slices
    for layer = 1:layerResolution(3)

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % set values in propagation matrix            
        
		% nearest interpolation
		% basisFunctionIndicesForRaysX = interp1(basisFunctionsX,basisFunctionsXX,rayPositionsX,'linear',0);

		% get currently effected layer pixels - matrix column indices for each ligh ray                      
		layerPixelIndicesForRaysX = ceil(lookupScaleX * (rayPositionsX-layerOrigin(2)));                        
		layerPixelIndicesForRaysY = ceil(lookupScaleY * (rayPositionsY-layerOrigin(1)));   

		% kick out stuff that's outside
		layerPixelIndicesForRaysX(layerPixelIndicesForRaysX>layerResolution(2)) = 0;
		layerPixelIndicesForRaysX(layerPixelIndicesForRaysX<1) = 0;
		layerPixelIndicesForRaysY(layerPixelIndicesForRaysY>layerResolution(1)) = 0;
		layerPixelIndicesForRaysY(layerPixelIndicesForRaysY<1) = 0;  

		% convert 4D subscipts to matrix indices
		matrixRows = 1:size(sampleid,2);
		matrixRows = matrixRows(layerPixelIndicesForRaysX~=0 & layerPixelIndicesForRaysY~=0);
				
		% convert to matrix column indices               
		validXXIndices = layerPixelIndicesForRaysX(layerPixelIndicesForRaysX~=0 & layerPixelIndicesForRaysY~=0);
		validYYIndices = layerPixelIndicesForRaysY(layerPixelIndicesForRaysX~=0 & layerPixelIndicesForRaysY~=0);             
		validZZIndices = layer + zeros(size(validXXIndices));

		% convert 3D subscripts to matrix indices
		matrixColumns   = sub2ind(layerResolution, validYYIndices(:), validXXIndices(:), validZZIndices(:));                                        
			
		%Set matrix
		indexIVector = cat(1, indexIVector, matrixRows(:));
		indexJVector = cat(1, indexJVector, matrixColumns(:));
		valueSVector = cat(1, valueSVector, ones(size(matrixRows)));

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        % shift x positions of rays to next slice
        rayPositionsX = rayPositionsX - layerDistance*vx;
        rayPositionsY = rayPositionsY - layerDistance*vy;

    end
 
    
    % construct the sparse matrix and add all non-zeros entries
	mat = sparse(indexIVector,indexJVector,valueSVector,size(sampleid,2),prod(layerResolution));
	clear indexIVector indexJVector valueSVector;
    
end


