% ***Image synthesis for compressive displays***
%
% see our paper: F. Heide, G. Wetzstein, R. Raskar, W. Heidrich.
% ''Adaptive Image Synthesis for Compressive Displays'', SIGGRAPH 2013.
%
 
% This function is a launcher function for all reconstructon problems on a
% 1D-lightfield slice ! It should demonstrate all of the algorithms
% presented in the paper involving HR3D, Layered3D and animated HR3D. The
% code is easily generalizable to 4D lightfields. 

% Send me an email if you need further code.

% Copyright (C) 2013. Felix Heide
% Email: fheide@cs.ubc.ca


clear all;
close all;
global T;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% user parameters

% number of layers
numLayers  = 3;

% scene to render
scene_fn = '../pbrt_v2/scenes/killeroo-simple-anim.pbrt';
%scene_fn = '../pbrt_v2/scenes/killeroo-simple.pbrt';
%scene_fn = '../pbrt_v2/scenes/killeroo-simple-1D-test.pbrt';

%Reference for PSNR (optional, for debug)
lightfield1Dref = readPFM('./data/lightfield_1d_ref.pfm');

t_start = 0;
t_end = 100;
num_frames = 20;

time_sampling = 0;
full_sampling = 0;

% data path
datapath = 'data/';
resultspath = 'results/';

% Select the method
%method = 0; %Layered 3D
%method = 1; %HR3D
method = 2; %HR3D animation

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if method ~= 2 && strcmp( scene_fn(end - 8:end), 'anim.pbrt' )
    error('Regular filename not found')
end

if method == 2 && ~strcmp( scene_fn(end - 8:end), 'anim.pbrt' )
    error('Animation filename not found')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Sparse matrix functions
addpath('../matlab_helpers/');
%addpath('../matlab_helpers/cuda');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Setup PBRT
addpath('../pbrt_v2/src/mex/');

%Define sampling params
numChains = 300; %10
chainLength = 20; %1000;
maxAcceptedSamplesPerChains = 20; %1000
mutation_sigma_spatial = 2.0; %1.0;
mutation_sigma_angular = 3.0; %1.0;
thresh_residual = 0.075; %0.05;
numChainDuplicateChecks = 0;

replacesamples = 0;

sampling_params = single([numChains; chainLength; maxAcceptedSamplesPerChains; ...
                          mutation_sigma_spatial; mutation_sigma_angular; thresh_residual; ...
						  numChainDuplicateChecks; replacesamples; time_sampling; full_sampling]);

%Reconstruciton type
%0 --> No reconstruction check (no adaptive sampling, but full sampling)
%1 --> Full data reconstruction
%2 --> Layer reconstruction
rec_type = 1; 
rec_type = single(rec_type);

%Duplicate check
%0 --> No duplicate check
%1 --> Online dense checking
%2 --> Postprocessing with kd-tree
duplicate_chk_type = 2; 
duplicate_chk_type = single(duplicate_chk_type);

%Load
[pbrt_ptr, l_sz] = mex_pbrt('create', scene_fn, rec_type, duplicate_chk_type, sampling_params);

%Convert lightfield size (and all in matlab double)
resX =      double(l_sz(1));
resY =      double(l_sz(2));	
angleMinX = double(l_sz(3));
angleMaxX = double(l_sz(4));
resAngleX = double(l_sz(5));
angleMinY = double(l_sz(6));
angleMaxY = double(l_sz(7));
resAngleY = double(l_sz(8));
resTime = double(l_sz(9));

lightFieldResolution = [ resAngleY, resAngleX, resY, resX, 3 ];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Proper version pbrt and matlab same
%Compute angles in v-units (tan) (consistently with pbrt)
angleMinXV = tan( degtorad( angleMinX ) );
angleMaxXV = tan( degtorad( angleMaxX ) );
firstAngleXV = angleMinXV + 0.5/resAngleX*(angleMaxXV - angleMinXV);
lastAngleXV = angleMaxXV - 0.5/resAngleX*(angleMaxXV - angleMinXV);
lightFieldAnglesX = linspace( firstAngleXV, lastAngleXV, resAngleX);

angleMinYV = tan( degtorad( angleMinY ) );
angleMaxYV = tan( degtorad( angleMaxY ) );
firstAngleYV = angleMinYV + 0.5/resAngleY*(angleMaxYV - angleMinYV);
lastAngleYV = angleMaxYV - 0.5/resAngleY*(angleMaxYV - angleMinYV);
lightFieldAnglesY =  linspace( firstAngleYV, lastAngleYV , resAngleY);

%Set resolutions 
% set layer resolution here
layerResolution = [lightFieldResolution(3) lightFieldResolution(4) numLayers];

% layer size is size of light field
lightFieldSize  = [1,50]; %TODO: Change to actual number (Fixed arbitrarily)
layerSize       = lightFieldSize; 

% distance between layers
%layerDistance = 16.7 / (numLayers-1);
layerDistance = 0.1*2.54 * 10;

% origin of the volume in world space [y x z] [mm] - center in layers
layerOrigin         = [0 0 0];
depthRange          = layerDistance * (numLayers-1);                
lightFieldOrigin(3) = -depthRange/2; 

% minimum transmission of each layer
minTransmission = 0.001;

% shift the light field origin to the center of the layers' depth range
bLightFieldOriginInLayerCenter = true;
                                   
% pad layer borders a little bigger
bPadLayerBorders = true;
if bPadLayerBorders    
    % pixel size of the layers and the light field
    lfPixelSize         = lightFieldSize ./ [lightFieldResolution(3) lightFieldResolution(4)];    

    % get maximum angles 
    lfMaxAngle          = [ max(abs(lightFieldAnglesY)) max(abs(lightFieldAnglesX)) ];

    % depth range of layers
    depthRange          = layerDistance * (numLayers-1);
    if bLightFieldOriginInLayerCenter
        depthRange      = depthRange / 2;
    end

    % number of pixels to add on each side of the layers
    numAddPixels        = ceil( lfMaxAngle.*depthRange ./ lfPixelSize );   

    % size of the clamped regions
    addedRegionSize     = numAddPixels .* lfPixelSize;    
    % width of the layers [y x] in mm
    layerSize   = [lightFieldSize(1)+2*addedRegionSize(1) lightFieldSize(2)+2*addedRegionSize(2)];    
    % origin of the volume in world space [y x z] [mm]
    layerOrigin = [lightFieldOrigin(1)-addedRegionSize(1) lightFieldOrigin(2)-addedRegionSize(2) 0];
    % adjust resolution
    layerResolution = [lightFieldResolution(3)+2*numAddPixels(1) lightFieldResolution(4)+2*numAddPixels(2) numLayers];
end


%HR3D params
timeSteps = 10; %Just for tests
layerSubset = [1,3];
layerSubset = single( layerSubset - 1 ); %C-idx

%HR3D anim params
old_fixed_layers = 2;
                    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pre-computation

basisFunctionType = 0;

% single channel light field resolution
lfResolution = [lightFieldResolution(1) lightFieldResolution(2) lightFieldResolution(3) lightFieldResolution(4)];

if method == 0
    
    %Precompute sparse propagation matrix if Layered3D
    
    % compute the sparse propagation matrix - this will internally generate and populate the global variable T
    precomputeSparsePropagationMatrixLayers3D( lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lfResolution, lightFieldOrigin,...
                                                layerResolution, layerSize, layerOrigin, layerDistance, ...
                                                basisFunctionType, 1 );
end                                     
               
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% do reconstruction                       
                
% monochromatic                                                                                     
disp('Reconstructing Layers from Color Light Field');                                           

% reconstruct layers stochastically
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
verbose_reconstruct = 'all';
tic;

if method == 0
    
    %Layered3D
    max_sample_it = 30;
    max_min_it = 100;
    
    if full_sampling
        max_sample_it = 1;
        max_min_it = 500;
    end
    
    [layersRec lightFieldRec B_sampled_map minimization_stochastic_time] = ...
        computeAttenuationLayersFromLightField2D_stochastic(lightFieldResolution, minTransmission, ...
                                                            lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lfResolution, lightFieldOrigin,...
                                                            layerResolution, layerSize, layerOrigin, layerDistance, basisFunctionType, ...
                                                            chainLength, mutation_sigma_spatial, ...
                                                            mutation_sigma_angular, thresh_residual, ...
                                                            numChainDuplicateChecks, ...
                                                            max_sample_it, max_min_it, ...
                                                            pbrt_ptr, ...
                                                            verbose_reconstruct);
                                                        
elseif method == 1
        
    %HR3D
    max_sample_it = 20;
    max_min_it = 150;
    
    if full_sampling
        max_sample_it = 1;
        max_min_it = 2000;
    end
    
    scaling_factor_hr3d = 1;

    [layersRec lightFieldRec B_sampled_map minimization_stochastic_time] = ...
        computeAttenuationLayersFromLightField2D_HR3D_stochastic(lightFieldResolution, ...
                                                        lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lfResolution, lightFieldOrigin,...
                                                        layerResolution, layerSize, layerOrigin, layerDistance, basisFunctionType, ...
                                                        timeSteps, layerSubset, ...
                                                        chainLength, mutation_sigma_spatial, ...
                                                        mutation_sigma_angular, thresh_residual, ...
                                                        numChainDuplicateChecks, ...
                                                        max_sample_it, max_min_it, ...
                                                        0, ...
                                                        scaling_factor_hr3d, ...
                                                        pbrt_ptr, ...
                                                        verbose_reconstruct, lightfield1Dref);
                                                    
elseif method == 2
    
    if full_sampling
        error('Full sampling not supported here\n')
    end

    %HR3D with animation
    scaling_factor = 0.3 * timeSteps;
    max_it = 150;

    %HR3D with animation version 1
    layersRecFrames = cell(num_frames,1);

    %Start with initial frame
    fprintf('\n#### DOING TIME FRAME %d ####\n', t_start)

    verbose_anim_init = 'brief';
    [layersRec lightFieldRec B_sampled_map minimization_stochastic_time] = ...
    computeAttenuationLayersFromLightField2D_HR3D_stochastic(lightFieldResolution, ...
                                                        lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lfResolution, lightFieldOrigin,...
                                                        layerResolution, layerSize, layerOrigin, layerDistance, basisFunctionType, ...
                                                        timeSteps, layerSubset, ...
                                                        chainLength, mutation_sigma_spatial, ...
                                                        mutation_sigma_angular, thresh_residual, ...
                                                        numChainDuplicateChecks, ...
                                                        10, max_it, ...
                                                        t_start, ...
                                                        scaling_factor, ...
                                                        pbrt_ptr, ...
                                                        verbose_anim_init);

    %Save
    layersRecFrames{0 + 1,1} = layersRec;

    %Write reconstruction                                                 
    write_res = true;

    if write_res        
        LF_rec = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);
        for ch = 1:3
        LF_rec(:,:,ch) = reshape( lightFieldRec(:,:,:,:,ch), lightFieldResolution(2), lightFieldResolution(4) );
        end
        imwrite(LF_rec, sprintf('frame%03d.png', t_start));      
    end

    % Now do all the timesteps
    frame_offset = ceil(  (t_end - t_start)/num_frames );
    layersPrev = zeros(layerResolution(1), layerResolution(2), timeSteps, 3, 2);

    for t = 1:num_frames

        time = t * frame_offset;

        %Recompute result with last samples
        layersPrev(:, :, :, :, 1) =  layersRec{1,1};
        layersPrev(:, :, :, :, 2) = layersRec{1,2};            

        verbose_anim = 'all';

        %Initialize with first layers
        layersPrevMat = reshape(layersPrev, [layerResolution(1) *layerResolution(2), timeSteps, 3, 2]);

        ofl = old_fixed_layers;
        layersPrevMat(:,1:ofl,:,1) = layersPrevMat(:,(end - ofl + 1):end,:,1);  
        layersPrevMat(:,1:ofl,:,2) = layersPrevMat(:,(end - ofl + 1):end,:,2); 
        layersPrevMat(:,(ofl+1):end,:,1) = rand(size(layersPrevMat(:,(ofl+1):end,:,1)));  
        layersPrevMat(:,(ofl+1):end,:,2) = rand(size(layersPrevMat(:,(ofl+1):end,:,2)));

        layersPrevMat = reshape(layersPrevMat, [layerResolution(1), layerResolution(2), timeSteps, 3, 2]);


        %Pack layers into cell
        layersPrev_cell = cell(1,2);
        layersPrev_cell{1,1} = layersPrevMat(:,:,:,:,1);
        layersPrev_cell{1,2} = layersPrevMat(:,:,:,:,2);

        %Reset old chains before new sampling (optional)
        mex_pbrt('reset_chains', pbrt_ptr);
        fprintf('Chain full reset done\n')

        fprintf('\n#### DOING TIME FRAME %d ####\n', time)

        [layersRec lightFieldRec B_sampled_map minimization_stochastic_time] = ...
        computeAttenuationLayersFromLightField2D_HR3D_anim(lightFieldResolution, ...
                                                            lightFieldAnglesY, lightFieldAnglesX, lightFieldSize, lfResolution, lightFieldOrigin,...
                                                            layerResolution, layerSize, layerOrigin, layerDistance, basisFunctionType, ...
                                                            timeSteps, layerSubset, ...
                                                            chainLength, mutation_sigma_spatial, ...
                                                            mutation_sigma_angular, thresh_residual, ...
                                                            numChainDuplicateChecks, ...
                                                            10, max_it, ...
                                                            time, layersPrev_cell, old_fixed_layers, ...
                                                            scaling_factor, ...
                                                            pbrt_ptr, ...
                                                            verbose_anim);

        %Save
        layersRecFrames{t + 1,1} = layersRec;

        %Write the reconstruction
        if write_res
            LF_rec = zeros(lightFieldResolution(2), lightFieldResolution(4), 3);
            for ch = 1:3
                LF_rec(:,:,ch) = reshape( lightFieldRec(:,:,:,:,ch), lightFieldResolution(2), lightFieldResolution(4) );
            end
            imwrite(LF_rec, sprintf('frame%03d.png', time));
        end

    end
   
end
                                                    
tt = toc;

disp(['  Reconstruction OVERALL stochastic took ' num2str(tt) ' secs']);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove added border pixels                                  

if bPadLayerBorders    
    
    if method == 0
    
        %Layered3D
        layersRec           = layersRec(numAddPixels(1)+1:end-numAddPixels(1), numAddPixels(2)+1:end-numAddPixels(2), :, :, :); 
        layerResolution(1)  = layerResolution(1) - 2*numAddPixels(1);
        layerResolution(2)  = layerResolution(2) - 2*numAddPixels(2);
    elseif method == 1
        
        %HR3D
        for l =1:size(layersRec,2)
            currHR3DLayer = layersRec{1 , l}; 
            layersRec{1 , l}   = currHR3DLayer(numAddPixels(1)+1:end-numAddPixels(1), numAddPixels(2)+1:end-numAddPixels(2), :, :); 
        end
        
        layerResolution(1)  = layerResolution(1) - 2*numAddPixels(1);
        layerResolution(2)  = layerResolution(2) - 2*numAddPixels(2);
          
    end
    
end

% filename for reconstruction
tomographyMode = 1; %Tomography
solver = 1; %SART
filename = [resultspath 'Reconstruction3D_' num2str(numLayers) 'layers_dist' num2str(layerDistance) ...
            '_tomographyMode' num2str(tomographyMode) '_solver' num2str(solver) '.mat'];

% save data
disp(['Done. Saving data as ' filename]);
save(filename, 'layersRec', 'lightFieldRec', 'lightFieldAnglesY', 'lightFieldAnglesX', ...
                'lightFieldSize', 'lightFieldOrigin', 'minTransmission');    
            
            
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Destroy pbrt
mex_pbrt('destroy', pbrt_ptr);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% plot data
figure();
numSubplots = [1 4];
                
% original light field
%{
subplot(numSubplots(1), numSubplots(2), 1);
img = reshape(lightField, [size(lightField,2) size(lightField,4) size(lightField,5)]);
imshow(img(:,:,1));
title('Original');
%}
                
% reconstruction                
subplot(numSubplots(1), numSubplots(2), 2);
img = reshape(lightFieldRec, [lightFieldResolution(2), lightFieldResolution(4), 3] );
imshow(img);
title('Reconstruction');  

%Sample map
subplot(numSubplots(1), numSubplots(2), 3);
imshow(B_sampled_map);
title('Samples');  

%{
% reference
subplot(numSubplots(1), numSubplots(2), 4);
img = reshape(lightFieldRec_ref, [size(lightFieldRec_ref,2) size(lightFieldRec_ref,4) size(lightFieldRec_ref,5)]);
imshow(img(:,:,1));
title('Reference');  
%}

colormap gray;
drawnow;

